/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more Details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */


/*
 * Initiating Points to AVS Geometry Transform
 *
 */

static char ipoGeom_Id[] = "%W%\t%G%";

#include <stdio.h>

#include <avs/avs.h>
#include <avs/avs_data.h>
#include <avs/geom.h>
#include <avs/flow.h>

#include "Sy_Avs.h"
#include "Sy_Skeleton.h"

extern enum SyRenderMethod  _sy_avs_render_method;
extern enum SyTriangleDecor _sy_avs_triangle_decor;
extern float              _sy_avs_triangle_height;

static GEOMobj*      obj;
static SyIpoDAC*       ipd;

static SyAvsVector*    vertex;
static SyAvsNormal*    normal;
static int           vindex, nindex, 
                     vdim, ndim;

static void Sy_AvsIpo2GeomInit(void);
static void Sy_AvsIpo2GeomPrep(void);
static void Sy_AvsIpoInsert   (void);
static void Sy_AvsIpo2GeomTerm(void);
static void Sy_AvsIpo2GeomExit(void);

void Sy_AvsIpo2Geom (GEOMobj *avs_obj, SyIpoDAC *skel_ipd)
{
  SyWalktask wt;
  
  obj = avs_obj;
  ipd = skel_ipd;

  Sy_ResetWalktask (&wt);

  wt.user_function[WALK] = (void(*)())Sy_AvsIpoInsert;
  
  wt.user_function[INIT] = (void(*)())Sy_AvsIpo2GeomInit;
  wt.user_function[TRAV] = (void(*)())Sy_AvsIpo2GeomPrep;
  wt.user_function[TERM] = (void(*)())Sy_AvsIpo2GeomTerm;
  wt.user_function[ROT ] = (void(*)())NULL;
  wt.user_function[FEXIT] = (void(*)())Sy_AvsIpo2GeomExit;

  Sy_InitIpoTraverse (ipd, &wt);

  Sy_IpoTraverse (ipd, &wt);
}
 
static void Sy_AvsIpo2GeomInit (void)
{
  vertex = (SyAvsVector*) NULL;
  normal = (SyAvsNormal*) NULL;
  vindex = nindex = 0;
  
  switch(_sy_avs_render_method){
  case POLYLINE: 
    vertex = (SyAvsVector*) Sy_Malloc((ipd->ipodim+1)*sizeof(SyAvsVector));
    vdim   = ipd->ipodim+1;
    break;
  case POLYTRIANGLE:
    vertex = (SyAvsVector*) Sy_Malloc(2*(ipd->ipodim+1)*sizeof(SyAvsVector));
    normal = (SyAvsNormal*) Sy_Malloc(2*(ipd->ipodim+1)*sizeof(SyAvsNormal));
    vdim   = ndim = 2*(ipd->ipodim+1);
    break;
  }
}

static void Sy_AvsIpo2GeomPrep (void)
{
  vindex = 0;
  nindex = 0;
}

static void Sy_AvsIpoInsert (void)
{
  switch(_sy_avs_render_method){
  case POLYLINE: 
    vertex[vindex].x = wi.current.pos.x;
    vertex[vindex].y = wi.current.pos.y;
    vertex[vindex].z = 0.0;
    vindex++;
    break;
  case POLYTRIANGLE:
    vertex[vindex].x = wi.current.pos.x;
    vertex[vindex].y = wi.current.pos.y;
    vertex[vindex].z = 0.0;
    vindex++;
    vertex[vindex].x = wi.current.pos.x;
    vertex[vindex].y = wi.current.pos.y;
    vertex[vindex].z = _sy_avs_triangle_height;
    vindex++;
    break;
  }
}

static void Sy_AvsIpo2GeomTerm (void)
{
  int i;

  if (wi.closed_contour){
    switch(_sy_avs_render_method){
    case POLYLINE: 
      vindex = vindex;
      vertex[vindex].x = wi.start.pos.x;
      vertex[vindex].y = wi.start.pos.y;
      vertex[vindex].z = 0.0;
      vindex++;
      break;
    case POLYTRIANGLE:
      vertex[vindex].x = wi.start.pos.x;
      vertex[vindex].y = wi.start.pos.y;
      vertex[vindex].z = 0.0;
      vindex++;
      vertex[vindex].x = wi.start.pos.x;
      vertex[vindex].y = wi.start.pos.y;
      vertex[vindex].z = _sy_avs_triangle_height;
      vindex++;
      break;
    }
  }
  switch(_sy_avs_render_method){
  case POLYLINE: 
    GEOMadd_polyline (obj, vertex, GEOM_NULL, vindex, GEOM_COPY_DATA); 
      break;
  case POLYTRIANGLE:
    for (i = 1; i < vindex; i+=2){
      normal[i-1].nx = normal[i].nx = vertex[i].y   - vertex[i-1].y;
      normal[i-1].ny = normal[i].ny = vertex[i-1].x - vertex[i].x;
      normal[i-1].nz = normal[i].nz = 0.0;
    }
    GEOMadd_polytriangle (obj, vertex, normal, GEOM_NULL, vindex, 
			  GEOM_COPY_DATA); 
  }
}

static void Sy_AvsIpo2GeomExit (void)
{
  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vertex) free (vertex);
    break;
  case POLYTRIANGLE:
    if (vertex) free (vertex);
    if (normal) free (normal);    
    break;
  }
} 

