/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more Details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Voronoi Data Structure to AVS Geometry Transform
 *
 */

static char voroGeom_Id[] = "%W%\t%G%";

#include <stdio.h>

#include <avs/avs.h>
#include <avs/avs_data.h>
#include <avs/geom.h>
#include <avs/flow.h>

#include "Sy_Avs.h"
#include "Sy_Skeleton.h"

extern enum SyRenderMethod  _sy_avs_render_method;
extern enum SyTriangleDecor _sy_avs_triangle_decor;
extern float              _sy_avs_triangle_height;

static GEOMobj*      obj;
static SyVoroDAC*      vd;

static SyAvsVector*    vertex;
static SyAvsNormal*    normal;
static int           vindex, nindex, 
                     vdim, ndim;

static void Sy_AvsVoro2GeomInit(void);
static void Sy_AvsVoro2GeomPrep(void);
static void Sy_AvsVoroInsert   (void);
static void Sy_AvsVoro2GeomTerm(void);
static void Sy_AvsVoro2GeomExit(void);

void Sy_AvsVoro2Geom (GEOMobj *avs_obj, SyVoroDAC *skel_vd)
{
  SyWalktask wt;
  
  obj = avs_obj;
  vd  = skel_vd;

  Sy_ResetWalktask (&wt);

  wt.user_function[WALK] = (void(*)())Sy_AvsVoroInsert;

  wt.user_function[INIT] = (void(*)())Sy_AvsVoro2GeomInit;
  wt.user_function[TRAV] = (void(*)())Sy_AvsVoro2GeomPrep;
  wt.user_function[TERM] = (void(*)())Sy_AvsVoro2GeomTerm;
  wt.user_function[ROT ] = (void(*)())Sy_AvsVoroInsert;
  wt.user_function[FEXIT] = (void(*)())Sy_AvsVoro2GeomExit;

  Sy_InitVoronoiTraverse (vd, &wt);

  Sy_VoronoiTraverse (vd, &wt);
}
 
static void Sy_AvsVoro2GeomInit (void)
{
  vertex = (SyAvsVector*) NULL;
  normal = (SyAvsNormal*) NULL;
  vindex = nindex = 0;
  
  switch(_sy_avs_render_method){
  case POLYLINE: 
    vertex = (SyAvsVector*) Sy_Malloc(4*(vd->edim+1)*sizeof(SyAvsVector));
    vdim   = 4*(vd->edim+1);
    break;
  case POLYTRIANGLE:
    vertex = (SyAvsVector*) Sy_Malloc(8*(vd->edim+1)*sizeof(SyAvsVector));
    normal = (SyAvsNormal*) Sy_Malloc(8*(vd->edim+1)*sizeof(SyAvsNormal));
    vdim   = ndim = 8*(vd->edim+1);
    break;
  }
}

static void Sy_AvsVoro2GeomPrep (void)
{
  vindex = 0;
  nindex = 0;
}

static void Sy_AvsVoroInsert (void)
{

  SyVertex vtmp;
  float x, y;


  vtmp  = syWstate.vd->vertex[syWstate.current.head_vertex];
  x = vtmp.xden/vtmp.num;
  y = vtmp.yden/vtmp.num;



  switch(_sy_avs_render_method){
  case POLYLINE: 
    vertex[vindex].x = x;
    vertex[vindex].y = y;
    vertex[vindex].z = 0.0;
    vindex++;
    break;
  case POLYTRIANGLE:
    vertex[vindex].x = x;
    vertex[vindex].y = y;
    vertex[vindex].z = 0.0;
    vindex++;
    vertex[vindex].x = x;
    vertex[vindex].y = y;
    vertex[vindex].z = _sy_avs_triangle_height;
    vindex++;
    break;
  }
}

static void Sy_AvsVoro2GeomTerm (void)
{
  static void Sy_AvsVoroInsert (void);
  
  int i;

  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vindex <= 1) {
      vindex = nindex = 0;
      return;
    }
    GEOMadd_polyline (obj, vertex, GEOM_NULL, vindex, GEOM_COPY_DATA); 
      break;
  case POLYTRIANGLE:
    if (vindex <= 2) {
      vindex = nindex = 0;
      return;
    }
    for (i = 1; i < vindex; i+=2){
      normal[i-1].nx = normal[i].nx = vertex[i].y   - vertex[i-1].y;
      normal[i-1].ny = normal[i].ny = vertex[i-1].x - vertex[i].x;
      normal[i-1].nz = normal[i].nz = 0.0;
    }
    GEOMadd_polytriangle (obj, vertex, normal, GEOM_NULL, vindex, 
			  GEOM_COPY_DATA); 
  }
  vindex = nindex = 0;
  Sy_AvsVoroInsert ();
}

static void Sy_AvsVoro2GeomExit (void)
{
  switch(_sy_avs_render_method){
  case POLYLINE: 
    if (vertex) free (vertex);
    break;
  case POLYTRIANGLE:
    if (vertex) free (vertex);
    if (normal) free (normal);    
    break;
  }
} 


GEOMobj* Sy_AvsPoint2Geom (SyPoint *point, int n)
{

  GEOMobj   *obj;
  SyAvsVector *vertex;
  float     *radius;
  int       i;
  
  vertex = (SyAvsVector*) Sy_Malloc (n*sizeof(SyAvsVector));
  radius = (float*)     Sy_Malloc (n*sizeof(float));
  
  for (i = 1; i <= n; i++){
    vertex[i-1].x = point[i].x;
    vertex[i-1].y = point[i].y;
    vertex[i-1].z = 0.0;
    radius[i-1]   = 0.3;
  }

  obj =  GEOMcreate_sphere (GEOM_NULL, vertex, radius, GEOM_NULL, 
			    GEOM_NULL, n, GEOM_DONT_COPY_DATA);

  return (obj);
}

