/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts. 
	    (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993,1994,1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char ctranId[] = "@(#)ctran.c	1.7\t4/19/91"; 

#include <stdio.h>
#include <sys/types.h>
#include "Sy_Skeleton.h"
#include "Sy_Uparse.h"
#include "Sy_Graphics.h"

#define SY_MAXNOPT     2

static char* usageStrings[] = {
"[options...] {infile,-} [outfile,-]",
"-ascii          : output in ascii format",
"-verbose        : verbose mode",
"-check          : run consistency check after crack extraction",
"-timer          : show resource usage",
"-cleanup        : do cleanup preprocessing",
"-clthreshold val: cleanup threshold",
"-pixinfo        : list pixel values in image (gray-level 8bit only !)",
"-pixval  val    : set foreground pixel value (gray-level 8bit only !)",
"-invert         : reverse video",
"-contourset     : output format = set of contour strings",
"-xcenter coord  : ",
"-ycenter coord  : 'focus of interest'",
"-scale factor   : scale factor",
"-xupleft coord  : ",
"-yupleft coord  : upper left corner of display 'window'",
"-display screen : X Windows display identifier",
"infile          : image (1/8/24-bit), GIF/TIFF/JPEG etc.",
"outfile         : initiating points (raster 'cracks')", NULL};

static SyOption longopt[] = {
{"xcenter", SY_ARG_REQ},
{"ycenter", SY_ARG_REQ},
{"scale",   SY_ARG_REQ},
{"xupleft", SY_ARG_REQ},
{"yupleft", SY_ARG_REQ},
{"pixval",  SY_ARG_REQ},
{"clthreshold", SY_ARG_REQ},
{"ascii",   SY_ARG_NONE},
{"verbose", SY_ARG_NONE},
{"check",   SY_ARG_NONE},
{"pixinfo", SY_ARG_NONE},
{"timer",   SY_ARG_NONE},
{"cleanup", SY_ARG_NONE},
{"invert",  SY_ARG_NONE},
{"contourset", SY_ARG_NONE},
{"display", SY_ARG_REQ},SY_NULL_OPTION};

main(int argc, char **argv)
{
  SyIpoDAC         ipd;
  SyImage_t        *ima = NULL;
  SyIOhandle       *iohandle = NULL;
  SyRegpot         pot;
  
  SyCratranControl cntl;
  
  char*          nopt[SY_MAXNOPT];
  char*          display;
  
  int            is_verbose, is_ascii, is_timer, 
                 is_check, is_inpipe, io_result,
                 is_cleanup, is_contourset, is_pixinfo, is_pixval,
                 pixval = 0;
  int            nrnopt;
 
  float          uscale, clthreshold;
  int            xPos, yPos, xZen, yZen, pid;

  char           hostname    [SY_HOSTNAME_LEN], 
                 win_title   [SY_TITLE_LEN], 
                 icon_title  [SY_TITLE_LEN],
                 display_info[SY_DISPLAY_STRING_LEN],
                 charbuf     [SY_CHARBUF_SIZE];
   
  if (!uparse_long(argv, argc, NULL, longopt, nopt, SY_MAXNOPT, &nrnopt)
      || (nrnopt < 1)){
    Sy_Usage (argv[0], usageStrings); exit(1);
  }

  gethostname (hostname, SY_HOSTNAME_LEN);     
  sprintf (win_title, "%s @ %s <%s>", argv[0], hostname, nopt[0]);
  sprintf (icon_title, "%s", nopt[0]);
  sprintf (display_info, "%s", long_option("display") ? 
	   long_argopt("display"): "default X server");
  
  display    = long_option("display") ? long_argopt("display") : (char*)NULL;
  
  is_ascii      = long_option("ascii");
  is_check      = long_option("check");
  is_verbose    = long_option("verbose");
  is_timer      = long_option("timer");
  is_cleanup    = long_option("cleanup");
  is_contourset = long_option("contourset");
  is_pixinfo    = long_option("pixinfo");

  sscanf (long_option("xcenter") ? long_argopt("xcenter"): "0",  "%d", &xZen);
  sscanf (long_option("ycenter") ? long_argopt("ycenter"): "0", "%d", &yZen);
  sscanf (long_option("xupleft") ? long_argopt("xupleft"): SY_DEFXSTR,"%d",
	  &xPos);
  sscanf (long_option("yupleft") ? long_argopt("yupleft"): SY_DEFYSTR,"%d",
	  &yPos);
  sscanf (long_option("scale") ? long_argopt("scale")  : "0", "%f", &uscale);
  
  sscanf (long_option("clthreshold") ? long_argopt("clthreshold") : "0.10", 
	  "%f", &clthreshold);

  sscanf ((is_pixval = long_option("pixval")) ? 
	  long_argopt ("pixval"): "0", "%d", &pixval);
  
  Sy_InitIpo(&ipd);

  ipd.info              = (SyIpoInfo*) Sy_Malloc(sizeof(SyIpoInfo));
  ipd.info->clthreshold = clthreshold;
  ipd.is_contour_index  = SY_FALSE;
  
  Sy_SetVerbose (is_verbose);
  
  /* read image */
  
  ima = Sy_LoadImageFromFile (nopt[0], SY_GREY_OR_BITMAP, NULL, is_verbose);
  
  if (ima == NULL){
    Sy_SimpleError ("Could not load image from %s\n", argv[0]); 
  }

  if (is_pixinfo){
   Sy_CheckPixvalInImage (ima);
   exit(0);
  }

  if (is_timer) Sy_InitTimer();
  
  /* run crack extraction */
  cntl.reverse_bits = 0;
  cntl.invert_image = long_option("invert");
  ima->is_pixval    = is_pixval;
  ima->greyval      = pixval;

  ipd.label.p_ = NULL;
  ipd.label.c_ = NULL;

  Sy_Cratran (ima, &ipd, &cntl);
  
  Sy_FreeImage (ima);
  
  /* build initiating contours */
  
  Sy_Message ("\n%s ", "building boundary chains ");
  
  Sy_CratranBuildIcn (&ipd);
  
  Sy_Message ("%s\n", "done");
  
  if (is_timer) Sy_ShowTimer();
  
  sprintf (charbuf, "Boundary points (cracklines): %d\n", ipd.ipodim);
  Sy_Message ("%s", charbuf);
  
  if (is_cleanup){
    Sy_CleanupIpo (&ipd, &pot);
  }
  
  if (is_check){
    Sy_Message ("%s ", "Consistency check");
    if (Sy_CratranIpoCheck(&ipd) == 0) Sy_Message ("%s\n", "passed");
    else                              Sy_Message ("%s\n", "failed");
  }
  
  if (nrnopt > 1){
    iohandle = Sy_CreateIOhandle ();
    Sy_SetIOproperty (iohandle, FORMAT, 
		      (SY_VOID*)((is_ascii) ? SY_ASCII:SY_BINARY));
    Sy_SetIOproperty (iohandle, TYPE, (SY_VOID*)(is_contourset ? 
		      SY_CONTOUR_UNORDERED_STRIP : SY_IPD_STRIP));
    Sy_SetIOproperty (iohandle, DATA, (SY_VOID*) &ipd);
    Sy_SetIOproperty (iohandle, IO_MODE, (SY_VOID*)SY_FREE_AFTER_WRITE);
    
    Sy_PutIOhandle (iohandle);
    Sy_ManageOutput (nopt[1]);
    
    sprintf (charbuf, "Writing points to %s \n", nopt[1]);
    Sy_Message ("%s", charbuf);

    Sy_DestroyIpo(&ipd);
    Sy_Free(iohandle);
  }
  else {
    sprintf (charbuf, "%s: preparing display on %s\n", argv[0], 
	     display_info);
    Sy_Message ("%s", charbuf);
    
    if ((pid = fork()) == 0){
      Sy_XIpoDisp (&ipd, win_title, icon_title, xPos, yPos, xZen, yZen,
		   uscale, display);
      Sy_DestroyIpo (&ipd);
      exit(0);
    }

    Sy_DestroyIpo (&ipd);
  }
}






