/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/* This program was adapted from software written by Paul Vixie in 1985/86
 * using examples from the book: "Algorithms & Data Structures," Niklaus 
 * Wirth, Prentice-Hall, 1986, ISBN 0-13-022005-1.
 */

/* #define DEBUG */

#include <stdio.h>
#include <memory.h>
#include "Sy_AvlTree.h"

#ifdef DEBUG
#define		MSG(msg)	printf("DEBUG: '%s'\n", msg);
#else
#define		MSG(msg)
#endif

static int sprout(SyAvlTree **ppr, void *pc_data, void *key, int *pi_balance, 
		  int (*pfi_compare) (void*, void*), 
		  int (*pfi_delete) (void*));

static int delete(SyAvlTree **ppr_p, int (*pfi_compare) (void*, void*), 
		  void *key, int (*pfi_uar) (void*, void*), 
		  int *pi_balance, int *pi_uar_called);

static void del(SyAvlTree **ppr_r, int *pi_balance, SyAvlTree **ppr_q, 
		int (*pfi_uar) (void*, void*), int *pi_uar_called);

static void balanceL(SyAvlTree **ppr_p, int *pi_balance);
static void balanceR(SyAvlTree **ppr_p, int *pi_balance);


void Sy_AvlTreeInit(SyAvlTree **ppr_tree)
{
  ENTER("tree_init");
  *ppr_tree = NULL;
  EXITV;
}

void *Sy_AvlTreeSrch(SyAvlTree **ppr_tree, int (*pfi_compare) (void*, void*), 
		     void *key)
{
  int	    i_comp;
  SyAvlTree *pr_new;

  ENTER("tree_srch");

  if (*ppr_tree){
    i_comp = (*pfi_compare)(key, (**ppr_tree).key);
    if (i_comp > 0)
      EXITNV(Sy_AvlTreeSrch( &(**ppr_tree).tree_r, pfi_compare, key));
    if (i_comp < 0) 
      EXITNV(Sy_AvlTreeSrch(&(**ppr_tree).tree_l, pfi_compare, key));
    /* not higher, not lower... this must be the one. */
    EXITNV((**ppr_tree).tree_p);
  }
  /* grounded. NOT found. */
  EXITNV(NULL);
}


int Sy_AvlTreeAdd(SyAvlTree **ppr_tree, int (*pfi_compare) (void*, void*), 
		  void *pc_user, void *key, int (*pfi_delete) (void*))
{
  int	i_balance = FALSE;
  int   new_node;

  ENTER("tree_add");
  new_node = 
    sprout(ppr_tree, pc_user, key, &i_balance, pfi_compare, pfi_delete);
  EXITNV(new_node);
}


static int sprout(SyAvlTree **ppr, void *pc_data, void *key, 
		   int *pi_balance, 
		   int (*pfi_compare) (void*, void*), 
		   int (*pfi_delete) (void*))
{
  SyAvlTree *p1, *p2;
  int	    cmp, new_node = 0;
  
  ENTER("sprout");

  /* are we grounded?  if so, add the node "here" and set the rebalance
     flag, then exit. */
  if (!*ppr) {
    MSG("grounded. adding new node, setting h=true");
    *ppr = (SyAvlTree *) malloc(sizeof(SyAvlTree));
    (*ppr)->tree_l = NULL;
    (*ppr)->tree_r = NULL;
    (*ppr)->tree_b = 0;
    (*ppr)->tree_p = pc_data;
    (*ppr)->key    = key;
    *pi_balance = TRUE;
    EXITNV(1);
  }
  /* compare the data using routine passed by caller. */
  cmp = (*pfi_compare)(key, (*ppr)->key);
  
  /* if LESS, prepare to move to the left. */
  if (cmp < 0) {
    MSG("LESS. sprouting left.");
    new_node = 
      sprout(&(*ppr)->tree_l, pc_data, key, pi_balance, pfi_compare, 
	     pfi_delete);
    if (*pi_balance) {	/* left branch has grown longer */
      MSG("LESS: left branch has grown");
      switch ((*ppr)->tree_b){
      case 1:	/* right branch WAS longer; balance is ok now */
	MSG("LESS: case 1.. balnce restored implicitly");
	(*ppr)->tree_b = 0;
	*pi_balance = FALSE;
	break;
      case 0:	/* balance WAS okay; now left branch longer */
	MSG("LESS: case 0.. balnce bad but still ok");
	(*ppr)->tree_b = -1;
	break;
      case -1:
	/* left branch was already too long. rebalance */
	MSG("LESS: case -1: rebalancing");
	p1 = (*ppr)->tree_l;
	if (p1->tree_b == -1) {	/* LL */
	  MSG("LESS: single LL");
	  (*ppr)->tree_l = p1->tree_r;
	  p1->tree_r = *ppr;
	  (*ppr)->tree_b = 0;
	  *ppr = p1;
	}
	else {			/* double LR */
	  MSG("LESS: double LR");  
	  p2 = p1->tree_r;
	  p1->tree_r = p2->tree_l;
	  p2->tree_l = p1;
	  
	  (*ppr)->tree_l = p2->tree_r;
	  p2->tree_r = *ppr;
	  
	  if (p2->tree_b == -1)
	    (*ppr)->tree_b = 1;
	  else
	    (*ppr)->tree_b = 0;
	    
	  if (p2->tree_b == 1)
	    p1->tree_b = -1;
	  else
	    p1->tree_b = 0;
	  *ppr = p2;
	} /*else*/
	(*ppr)->tree_b = 0;
	*pi_balance = FALSE;
      } /*switch*/
    } /*if*/
    EXITNV(new_node);
  } /*if*/
  
  /* if MORE, prepare to move to the right.
   */
  if (cmp > 0) {
    MSG("MORE: sprouting to the right");
    new_node = sprout(&(*ppr)->tree_r, pc_data, key, pi_balance,
		      pfi_compare, pfi_delete);
    if (*pi_balance) {	/* right branch has grown longer */
      MSG("MORE: right branch has grown");
      
      switch ((*ppr)->tree_b){
      case -1: MSG("MORE: balance was off, fixed implicitly");
	(*ppr)->tree_b = 0;
	*pi_balance = FALSE;
	break;
      case 0:  MSG("MORE: balance was okay, now off but ok");
	(*ppr)->tree_b = 1;
	break;
      case 1:  MSG("MORE: balance was off, need to rebalance");
	p1 = (*ppr)->tree_r;
	if (p1->tree_b == 1) {	/* RR */
	  MSG("MORE: single RR");
	  (*ppr)->tree_r = p1->tree_l;
	  p1->tree_l = *ppr;
	  (*ppr)->tree_b = 0;
	  *ppr = p1;
	}
	else {			/* double RL */
	  MSG("MORE: double RL");
	  p2 = p1->tree_l;
	  p1->tree_l = p2->tree_r;
	  p2->tree_r = p1;
	  
	  (*ppr)->tree_r = p2->tree_l;
	  p2->tree_l = *ppr;
	  
	  if (p2->tree_b == 1)
	    (*ppr)->tree_b = -1;
	  else
	    (*ppr)->tree_b = 0;
	  
	  if (p2->tree_b == -1)
	    p1->tree_b = 1;
	  else
	    p1->tree_b = 0;
	  
	  *ppr = p2;
	} /*else*/
	(*ppr)->tree_b = 0;
	*pi_balance = FALSE;
      } /*switch*/
    } /*if*/
    EXITNV(new_node);
  } /*if*/
  
  /* not less, not more: this is the same key!  replace... */
  MSG("I found it!  Replacing data value");
  *pi_balance = FALSE;
  if (pfi_delete) (*pfi_delete)((*ppr)->tree_p);
  (*ppr)->tree_p = pc_data;
  EXITNV(new_node);
}

int Sy_AvlTreeDelete (SyAvlTree **ppr_p, int (*pfi_compare) (void*, void*), 
		      void *key, int (*pfi_uar) (void*, void*))
{
  int i_balance = FALSE, i_uar_called = FALSE;

  ENTER("tree_delete");
  EXITNV(delete(ppr_p, pfi_compare, key, pfi_uar,
	      &i_balance, &i_uar_called));
}


static int delete(SyAvlTree **ppr_p, int (*pfi_compare) (void*, void*), 
		  void *key, int (*pfi_uar) (void*, void*), 
		  int *pi_balance, int *pi_uar_called)
{
  SyAvlTree *pr_q;
  int	i_comp, i_ret;
  
  ENTER("delete");
  
  if (*ppr_p == NULL) {
    MSG("key not in tree");
    EXITNV(FALSE);
  }
  
  i_comp = (*pfi_compare)((*ppr_p)->key, key);
  if (i_comp > 0) {
    MSG("too high - scan left");
    i_ret = delete(&(*ppr_p)->tree_l, pfi_compare, key, pfi_uar,
		   pi_balance, pi_uar_called);
    if (*pi_balance)
      balanceL(ppr_p, pi_balance);
  }
  else if (i_comp < 0) {
    MSG("too low - scan right");
    i_ret = delete(&(*ppr_p)->tree_r, pfi_compare, key, pfi_uar,
		   pi_balance, pi_uar_called);
    if (*pi_balance)
      balanceR(ppr_p, pi_balance);
  }
  else {
    MSG("equal");
    pr_q = *ppr_p;
    if (pr_q->tree_r == NULL) {
      MSG("right subtree null");
      *ppr_p = pr_q->tree_l;
      *pi_balance = TRUE;
    }
    else if (pr_q->tree_l == NULL) {
      MSG("right subtree non-null, left subtree null");
      *ppr_p = pr_q->tree_r;
      *pi_balance = TRUE;
    }
    else {
      MSG("neither subtree null");
      del(&pr_q->tree_l, pi_balance, &pr_q, pfi_uar,
	  pi_uar_called);
      if (*pi_balance)
	balanceL(ppr_p, pi_balance);
    }
    free(pr_q);
    if (!*pi_uar_called && *pfi_uar)
      (*pfi_uar)(pr_q->tree_p, pr_q->key);
    i_ret = TRUE;
  }
  EXITNV(i_ret);
}


static void del(SyAvlTree **ppr_r, int *pi_balance, SyAvlTree **ppr_q, 
		int (*pfi_uar) (void*, void*), int *pi_uar_called)
{
  ENTER("del");

  if ((*ppr_r)->tree_r != NULL) {
    del(&(*ppr_r)->tree_r, pi_balance, ppr_q, pfi_uar,
	pi_uar_called);
    if (*pi_balance)
      balanceR(ppr_r, pi_balance);
  } else {
    if (pfi_uar) (*pfi_uar)((*ppr_q)->tree_p, (*ppr_q)->key);
    *pi_uar_called = TRUE;
    (*ppr_q)->tree_p = (*ppr_r)->tree_p;
    *ppr_q = *ppr_r;
    *ppr_r = (*ppr_r)->tree_l;
    *pi_balance = TRUE;
  }
  
  EXITV;
}


static void balanceL(SyAvlTree **ppr_p, int *pi_balance)
{
  SyAvlTree *p1, *p2;
  int	b1, b2;
  
  ENTER("balanceL");
  MSG("left branch has shrunk");
  
  switch ((*ppr_p)->tree_b){
  case -1: MSG("was imbalanced, fixed implicitly");
    (*ppr_p)->tree_b = 0;
    break;
  case 0: MSG("was okay, is now one off");
    (*ppr_p)->tree_b = 1;
    *pi_balance = FALSE;
    break;
  case 1: MSG("was already off, this is too much");
    p1 = (*ppr_p)->tree_r;
    b1 = p1->tree_b;
    if (b1 >= 0) {
      MSG("single RR");
      (*ppr_p)->tree_r = p1->tree_l;
      p1->tree_l = *ppr_p;
      if (b1 == 0) {
	MSG("b1 == 0");
	(*ppr_p)->tree_b = 1;
	p1->tree_b = -1;
	*pi_balance = FALSE;
      } else {
	MSG("b1 != 0");
	(*ppr_p)->tree_b = 0;
	p1->tree_b = 0;
      }
      *ppr_p = p1;
    } else {
      MSG("double RL");
      p2 = p1->tree_l;
      b2 = p2->tree_b;
      p1->tree_l = p2->tree_r;
      p2->tree_r = p1;
      (*ppr_p)->tree_r = p2->tree_l;
      p2->tree_l = *ppr_p;
      if (b2 == 1)
	(*ppr_p)->tree_b = -1;
      else
	(*ppr_p)->tree_b = 0;
      if (b2 == -1)
	p1->tree_b = 1;
      else
	p1->tree_b = 0;
      *ppr_p = p2;
      p2->tree_b = 0;
    }
  }
  EXITV;
}


static void balanceR(SyAvlTree **ppr_p, int *pi_balance)
{
  SyAvlTree *p1, *p2;
  int	    b1, b2;
  
  ENTER("balanceR");
  MSG("right branch has shrunk");
  switch ((*ppr_p)->tree_b)
    {
    case 1: MSG("was imbalanced, fixed implicitly");
      (*ppr_p)->tree_b = 0;
      break;
    case 0: MSG("was okay, is now one off");
      (*ppr_p)->tree_b = -1;
      *pi_balance = FALSE;
      break;
    case -1: MSG("was already off, this is too much");
      p1 = (*ppr_p)->tree_l;
      b1 = p1->tree_b;
      if (b1 <= 0) {
	MSG("single LL");
	(*ppr_p)->tree_l = p1->tree_r;
	p1->tree_r = *ppr_p;
	if (b1 == 0) {
	  MSG("b1 == 0");
	  (*ppr_p)->tree_b = -1;
	  p1->tree_b = 1;
	  *pi_balance = FALSE;
	} else {
	  MSG("b1 != 0");
	  (*ppr_p)->tree_b = 0;
	  p1->tree_b = 0;
	}
	*ppr_p = p1;
      } else {
	MSG("double LR");
	p2 = p1->tree_r;
	b2 = p2->tree_b;
	p1->tree_r = p2->tree_l;
	p2->tree_l = p1;
	(*ppr_p)->tree_l = p2->tree_r;
	p2->tree_r = *ppr_p;
	if (b2 == -1)
	  (*ppr_p)->tree_b = 1;
	else
	  (*ppr_p)->tree_b = 0;
	if (b2 == 1)
	  p1->tree_b = -1;
	else
	  p1->tree_b = 0;
	*ppr_p = p2;
	p2->tree_b = 0;
      }
    }
  EXITV;
}


int Sy_AvlTreeTrav (SyAvlTree **ppr_tree, int (*pfi_uar) (void*))
{
  ENTER("tree_trav")
    
    if (!*ppr_tree)
      EXITNV(TRUE);
  
  if (!Sy_AvlTreeTrav(&(**ppr_tree).tree_l, pfi_uar))
    EXITNV(FALSE);
  if (!(*pfi_uar)((**ppr_tree).tree_p))
    EXITNV(FALSE);
  if (!Sy_AvlTreeTrav(&(**ppr_tree).tree_r, pfi_uar))
    EXITNV(FALSE);
  EXITNV(TRUE);
}


void Sy_AvlTreeMung (SyAvlTree **ppr_tree, int (*pfi_uar) (void*, void*))
{
  ENTER("tree_mung");
  if (*ppr_tree)
    {
      Sy_AvlTreeMung(&(**ppr_tree).tree_l, pfi_uar);
      Sy_AvlTreeMung(&(**ppr_tree).tree_r, pfi_uar);
      if (pfi_uar)
	(*pfi_uar)((**ppr_tree).tree_p, (**ppr_tree).key);
      free(*ppr_tree);
      *ppr_tree = NULL;
    }
  EXITV;
}
