/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_DrawVoroId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"

#ifndef NOX
#include "voronoi.xbm"
#include <X11/Xatom.h>
static Atom proto_atom = None, delete_atom = None;

typedef struct {
  Window  winId;
  char    *tag;
  void   (*eventHandler)();
} SyLtEventAction;

static XEvent  LtXevent;

static GC      gc;
static Pixmap  icon;

static XPoint   *pointPtr;
static XSegment *segmentPtr;
static int      pointDim, segmentDim;

static Display *display;
static char    *displayName = NULL;
static int     screenNr;

static void Sy_CatchEvent(XEvent *event, Window winId, unsigned int winNr, 
			  char *tag);
static void Sy_ExposeEvent(XExposeEvent *event, Window winId, 
			   unsigned int winNr);
static void Sy_Voro2X(SyVoroDAC *vd, XPoint *pointPtr, int *pointDim, 
		      XSegment *segmentPtr, int *segmentDim, int *winWidth, 
		      int *winHeight, int xZen, int yZen, float uscale);
static void Sy_Voro2XRaw(SyVoroDAC *vd, XPoint *pointPtr, int *pointDim, 
			 XSegment *segmentPtr, int *segmentDim, 
			 SyDisplayControl control);
#endif
/**********************************************************************
 * function Sy_XVoroDisp (vd, windowName, iconName, xPos, yPos, 
 *                       xZen, yZen, uscale, userDisplay, control)
 * 
 * purpose: displays a Voronoi-diagram
 *
 * param:   [input] vd   -  Voronoi-diagram data
 *********************************************************************/
int Sy_XVoroDisp (SyVoroDAC *vd, char *windowName, char *iconName, 
		  int xPos, int yPos, int xZen, int yZen, float uscale, 
		  char *userDisplay, SyDisplayControl control)
{
#ifndef NOX
  SyLtEventAction  eventBuff;
  int            winWidth, winHeight;
  int            borderWidth = SY_DEF_BORDERWIDTH;

  pointPtr   = (XPoint*)   Sy_Malloc ((unsigned)vd->pdim * sizeof(XPoint));
  segmentPtr = (XSegment*) Sy_Malloc ((unsigned)vd->edim * sizeof(XSegment));

  if (control.raw_mode == SKE_ON){
    Sy_Voro2XRaw (vd, pointPtr, &pointDim, segmentPtr, &segmentDim, control); 
    winWidth  = control.window_dimX;
    winHeight = control.window_dimY;
  }
  else{
    Sy_Voro2X (vd, pointPtr, &pointDim, segmentPtr, &segmentDim, 
	       &winWidth, &winHeight, xZen, yZen, uscale); 
  }

  Sy_DestroyVoronoi (vd);

  displayName  = userDisplay != NULL ? userDisplay : NULL;
  if ((display = XOpenDisplay (displayName)) == NULL) return (-1);  
  screenNr     = XDefaultScreen (display);

  eventBuff.winId = XCreateSimpleWindow (display, 
					 RootWindow (display, screenNr), 
					 xPos, yPos, winWidth, 
					 winHeight, borderWidth,
					 BlackPixel (display, screenNr),
					 WhitePixel (display, screenNr));
  
  gc   = XCreateGC (display, eventBuff.winId, 0, 0);
  icon = XCreateBitmapFromData (display, eventBuff.winId, voro_icon_bits, 
                                voro_icon_width, voro_icon_height);
  XSetForeground (display, gc, BlackPixel (display, screenNr));
  XSetBackground (display, gc, WhitePixel (display, screenNr));
  XSetStandardProperties (display, eventBuff.winId, windowName, 
			  iconName , icon, NULL, 0, NULL);

  XSelectInput (display, eventBuff.winId, 
		ExposureMask | StructureNotifyMask);

  proto_atom = XInternAtom(display, "WM_PROTOCOLS", False);
  delete_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
  if ((proto_atom != None) && (delete_atom != None))
    XChangeProperty(display, eventBuff.winId, proto_atom, XA_ATOM, 32,
		    PropModeReplace, (unsigned char*) &delete_atom, 1);
  
  eventBuff.eventHandler = Sy_CatchEvent;
   
  XMapWindow (display, eventBuff.winId);
  while (LtXevent.type != Expose && LtXevent.xany.window != eventBuff.winId) 
    XNextEvent(display, &LtXevent);

  while (1){
    if (LtXevent.xany.window == eventBuff.winId)
      eventBuff.eventHandler (&LtXevent, LtXevent.xany.window, 1, NULL);
    XNextEvent (display, &LtXevent);
  }
#else
  Sy_Message("%s", "no X11 graphics support available\n");
#endif
}

#ifndef NOX
static void Sy_CatchEvent (XEvent *event, Window winId, 
			   unsigned int winNr, char *tag)
{
  XClientMessageEvent *me;
  switch (event->type){
  case Expose :
    Sy_ExposeEvent ((XExposeEvent*)event, winId, winNr); 
    break;
  case DestroyNotify:
    exit(2); break;
  case ClientMessage:
    me = (XClientMessageEvent*)event;
    if ((me->window == winId) && (me->data.l[0] == delete_atom)){
      exit(2);
    }
    break;
  default : break;
  }
  return;
}

static void Sy_ExposeEvent (XExposeEvent *event, Window winId, 
			    unsigned int winNr)
{
  XPoint   *cur_point_block;
  XSegment *cur_segment_block;
  int      count;

  for (cur_point_block = pointPtr, count = 0; count < pointDim/SY_XBLOCK;
       cur_point_block += SY_XBLOCK, count++){
    XDrawPoints  (display, winId, gc, cur_point_block, SY_XBLOCK, 
		  CoordModeOrigin);
  };
  if (pointDim%SY_XBLOCK != 0)
    XDrawPoints  (display, winId, gc, cur_point_block, 
		  pointDim%SY_XBLOCK, CoordModeOrigin);

  for (cur_segment_block = segmentPtr, count = 0; count < segmentDim/SY_XBLOCK;
       cur_segment_block += SY_XBLOCK, count++){
    XDrawSegments (display, winId, gc, cur_segment_block, SY_XBLOCK);
  };
  if (segmentDim%SY_XBLOCK != 0)
    XDrawSegments (display, winId, gc, cur_segment_block, 
		   segmentDim%SY_XBLOCK);
  return;
}

static void Sy_Voro2X (SyVoroDAC *vd, XPoint *pointPtr, 
		       int *pointDim, XSegment *segmentPtr, 
		       int *segmentDim, int *winWidth, int *winHeight, 
		       int xZen, int yZen, float uscale)
{
  SyVertex   vhead, vtail;
  int	   i;
  float    x1, y1, x2, y2;
  float    scaleX, scaleY, scaleXY;
  float    xOff, yOff, xShift, yShift;
  float    rubber;
  float    dirX, dirY;
  int      maxX, minX, minY;
  int      dimX, dimY, sDim; 
  int      effDimX, effDimY;
  int      add_segment;

  maxX = minX = vd->point[1].x;
  minY = max(vd->point[1].y,SY_DEF_MINY);
  
  for (i = 1; i <= vd->pdim; i++){
    maxX = max(maxX, vd->point[i].x);
    minX = min(minX, vd->point[i].x);
  }

  dimX = maxX-minX;
  dimY = vd->point[vd->pdim].y - minY;
  
  if (uscale > 0){
    effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  xShift  = effDimX/2 + SY_MIN_BORDER;
  yShift  = effDimY/2 + SY_MIN_BORDER;

  for (i = 1; i <= vd->pdim; i++){
    pointPtr[i-1].x = scaleXY * (vd->point[i].x - xOff) + xShift;
    pointPtr[i-1].y = scaleXY * (vd->point[i].y - yOff) + yShift;
  }

  for (i = 1, sDim = 0; i <= vd->edim; i++) {
    vtail = vd->vertex[vd->edge[i].tail_vertex];
    vhead = vd->vertex[vd->edge[i].head_vertex];

    x1 = scaleXY * (vtail.xden / vtail.num - xOff) + xShift;
    y1 = scaleXY * (vtail.yden / vtail.num - yOff) + yShift;
    x2 = scaleXY * (vhead.xden / vhead.num - xOff) + xShift;
    y2 = scaleXY * (vhead.yden / vhead.num - yOff) + yShift;

    switch (vd->edge[i].rank) {
    case 0:
      dirX = x2 - x1;
      dirY = y2 - y1;
      rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
      x2 += rubber * dirX;
      y2 += rubber * dirY;
      x1 -= rubber * dirX;
      y1 -= rubber * dirY;
      add_segment = 1;
      break;
    case 1:
      if (!(add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)))) break;
      dirX = x2 - x1;
      dirY = y2 - y1;
      rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
      x2 = x1 + rubber * dirX;
      y2 = y1 + rubber * dirY;
      break;
    default:
      if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)
			 && (SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) break;
      if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE))) {
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	x2 = x1 + rubber * dirX;
	y2 = y1 + rubber * dirY;
	break;
      }
      if (add_segment = ((SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) {
	dirX = x1 - x2;
	dirY = y1 - y2;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	x1 = x2 + rubber * dirX;
	y1 = y2 + rubber * dirY;
        break;
      }
    }
    if (add_segment) {
      segmentPtr[sDim].x1 = x1;
      segmentPtr[sDim].y1 = y1;
      segmentPtr[sDim].x2 = x2;
      segmentPtr[sDim].y2 = y2;
      sDim++;
    }
  }
  *winWidth   = effDimX + 2*SY_MIN_BORDER;
  *winHeight  = effDimY + 2*SY_MIN_BORDER;
  *segmentDim = sDim;
  *pointDim   = vd->pdim;
}


static void Sy_Voro2XRaw (SyVoroDAC *vd, XPoint *pointPtr, 
			  int *pointDim, XSegment *segmentPtr, 
			  int *segmentDim, SyDisplayControl control)
{
  SyVertex   vhead, vtail;
  float    x1, y1, x2, y2;
  float    rubber;
  float    dirX, dirY;
  int      i, sDim, add_segment;

  if (control.show_boundary == SKE_ON){
    for (i = 1; i <= vd->pdim; i++){
      pointPtr[i-1].x = vd->point[i].x;
      pointPtr[i-1].y = vd->point[i].y;
    }
  }

  for (i = 1, sDim = 0; i <= vd->edim; i++) {
    vtail = vd->vertex[vd->edge[i].tail_vertex];
    vhead = vd->vertex[vd->edge[i].head_vertex];

    x1 = vtail.xden / vtail.num;
    y1 = vtail.yden / vtail.num;
    x2 = vhead.xden / vhead.num;
    y2 = vhead.yden / vhead.num;

    switch (vd->edge[i].rank) {
    case 0:
      dirX = x2 - x1;
      dirY = y2 - y1;
      rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
      x2 += rubber * dirX;
      y2 += rubber * dirY;
      x1 -= rubber * dirX;
      y1 -= rubber * dirY;
      add_segment = 1;
      break;
    case 1:
      if (!(add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)))) break;
      dirX = x2 - x1;
      dirY = y2 - y1;
      rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
      x2 = x1 + rubber * dirX;
      y2 = y1 + rubber * dirY;
      break;
    default:
      if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)
			 && (SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) break;
      if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE))) {
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	x2 = x1 + rubber * dirX;
	y2 = y1 + rubber * dirY;
	break;
      }
      if (add_segment = ((SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) {
	dirX = x1 - x2;
	dirY = y1 - y2;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	x1 = x2 + rubber * dirX;
	y1 = y2 + rubber * dirY;
        break;
      }
    }
    if (add_segment) {
      segmentPtr[sDim].x1 = x1;
      segmentPtr[sDim].y1 = y1;
      segmentPtr[sDim].x2 = x2;
      segmentPtr[sDim].y2 = y2;
      sDim++;
    }
  }
  *segmentDim = sDim;
  *pointDim   = (control.show_boundary == SKE_ON)? vd->pdim : 0;
}
#endif /* NOX */
