/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * simple I/O for initiating points
 *
 */

static char sy_IpoIOId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Lang.h"
#include "Sy_Yacc.h"

/*********************************************************************
 * function int Sy_Ipo_fprintf (fp, ipd, iomode)
 * 
 * purpose: sends SyIpoDAC data to a file referenced through fp
 *          (ASCII format)
 *
 * param:   [input] SyIpoDAC* ipd    -  
 *          [input] FILE*     fp   - file pointer  
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Ipo_fprintf (FILE *fp, char *indata, unsigned int iomode)
{
  SyIpoDAC* ipd;
  int i, io_result;
	
  ipd = (SyIpoDAC*) indata;

  io_result = fprintf (fp, 
		       "/* contour info: */\npointdim = %d; edgedim = %d;\n", 
		       (int) ipd->ipodim, (int) ipd->icndim);

  if (io_result == EOF) return (io_result); 

  io_result = fprintf (fp, "has_contour_index = %s;\n", 
		       ipd->is_contour_index ? "true" : "false");

  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_TRUE);
  fprintf (fp, "\n");

  for (i = 1; i <= ipd->ipodim; i++){

    io_result = fprintf (fp, "%d %d\n", 
			 (int) ipd->ipo[i].x, 
			 (int) ipd->ipo[i].y);

    if (io_result == EOF) return (io_result); 
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)ipd->ipo);
    ipd->ipo    = (SyIpo*) NULL;
  }

  if (ipd->icndim == 0){ return(1); }

  for (i = 1; i <= ipd->ipodim; i++){
    
    io_result = fprintf (fp, "%6d %6d %f %f\n",
                         (int) ipd->ipl[i].tail_icn,   
                         (int) ipd->ipl[i].head_icn,
                               ipd->ipl[i].grad.x, 
                               ipd->ipl[i].grad.y);
    
    if (io_result == EOF) return (io_result); 
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)ipd->ipl);
    ipd->ipl = (SyIpl*) NULL;
  }

  for (i = 1; i <= ipd->icndim; i++) {
    io_result = fprintf (fp, "%6d %6d\n", 
			 ipd->icn[i].tail_ipo, 
			 ipd->icn[i].head_ipo);
    
    if (io_result == EOF) return (io_result); 
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
   Sy_Free ((SY_VOID*)ipd->icn);
   ipd->icn    = (SyIcn*) NULL;
   ipd->icndim = 0;
 }

  if (ipd->is_contour_index){

    for (i = 1; i <= ipd->ipodim; i++){
      io_result = fprintf (fp, "%6d %6d\n",
			   (int) ipd->contour_index[i].contour_id,   
			   (int) ipd->contour_index[i].point_id);

      if (io_result == EOF) return (io_result); 
    }

    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)ipd->contour_index);
      ipd->contour_index    = NULL;
    }
  } /* if is_contour_index */

  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    ipd->ipodim = 0;
  }

  return (1);
}


/**********************************************************************
 * function int Sy_Ipo_fscanf (fp, ipd)
 * 
 * purpose: gets SyIpoDAC data from a file referenced through fp
 *          (ASCII format)
 *
 * param:   [inout] SyIpoDAC ipd   -  SyIpoDAC data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Ipo_fscanf (FILE *fp, SyIpoDAC **ret_ipd, unsigned int iomode)
{
  SyIpoDAC  *ipd;
  int	i, io_result;
  int   buff [SY_NBUFF];
  float fbuf [SY_NBUFF];
  int   bytes = 0;

  SyDataObject *dobj;

  ipd = (SyIpoDAC*) Sy_Malloc (sizeof(SyIpoDAC));
  Sy_InitIpo(ipd);
	
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
	      "missing pointdim info...\n", "");  
  }
  ipd->ipodim = dobj->t.i_val; 
	
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
	      "missing edgedim info...\n", "");  
  }
  ipd->icndim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "has_contour_index");
  if (dobj == NULL) ipd->is_contour_index = SY_FALSE;
  else ipd->is_contour_index = strcmp(dobj->t.s_val, "true") ? 
	 SY_FALSE : SY_TRUE; 
  
  ipd->ipo = Sy_AllocIpo (ipd->ipodim);
  ipd->ipl = Sy_AllocIpl (ipd->ipodim);
  ipd->icn = Sy_AllocIcn (ipd->icndim);

  if (ipd->is_contour_index){
    ipd->contour_index = Sy_AllocContourIndex(ipd->ipodim);
  }
	
  for (i = 1; i <= ipd->ipodim; i++){
    io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]);  
    ipd->ipo[i].x = buff[0]; 
    ipd->ipo[i].y = buff[1];
    
    if (io_result != 2) return (EOF); 
  }

  if (ipd->icndim == 0) { *ret_ipd = ipd; return (bytes); }
  
  for (i = 1; i <= ipd->ipodim; i++){ 
    io_result = fscanf (fp, "%d %d %f %f",
			&buff[0], &buff[1], &fbuf[0], &fbuf[1]);
    
    ipd->ipl[i].tail_icn  = buff[0];   
    ipd->ipl[i].head_icn  = buff[1];
    ipd->ipl[i].grad.x    = fbuf[0]; 
    ipd->ipl[i].grad.y    = fbuf[1];
    
    if (io_result != 4) return (EOF); 
  }
  
  for (i = 1; i <= ipd->icndim; i++) {  
    io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]); 
    ipd->icn[i].tail_ipo = buff[0];
    ipd->icn[i].head_ipo = buff[1];
   
    if (io_result != 2) return (EOF); 
  }

  if (ipd->is_contour_index){
    for (i = 1; i <= ipd->ipodim; i++){ 
      io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]);
      
      ipd->contour_index[i].contour_id = buff[0];   
      ipd->contour_index[i].point_id   = buff[1];
      
      if (io_result != 2) return (EOF); 
    } 
  }

  *ret_ipd = ipd;
  return (bytes);
}


/*********************************************************************
 * function int Sy_Ipo_fwrite (fp, ipd)
 * 
 * purpose: sends SyIpoDAC data to a file referenced through fp
 *          (BINARY format)
 *
 * param:   [input] SyIpoDAC* ipd  -  SyIpoDAC data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Ipo_fwrite (FILE *fp, char *indata, unsigned int iomode)
{
  SyIpoDAC* ipd;
  int i, 
      io_result;
	
  ipd = (SyIpoDAC*) indata;

  io_result = fprintf (fp, 
		       "/* contour info: */\npointdim = %d; edgedim = %d;\n", 
		       (int) ipd->ipodim, (int) ipd->icndim);

  if (io_result == EOF) return (EOF); 

  io_result = fprintf (fp, "has_contour_index = %s;\n", 
		       ipd->is_contour_index ? "true" : "false");

  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_FALSE);

  io_result = fwrite ((SY_VOID*)ipd->ipo, sizeof(SyIpo), ipd->ipodim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)ipd->ipo);
    ipd->ipo    = (SyIpo*) NULL;
  }

  if (ipd->icndim == 0){ return(1); }

  io_result = fwrite ((SY_VOID*)ipd->ipl, sizeof(SyIpl), ipd->ipodim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)ipd->ipl);
    ipd->ipl = (SyIpl*) NULL;
  }

  io_result = fwrite ((SY_VOID*)ipd->icn, sizeof(SyIcn), ipd->icndim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
   Sy_Free ((SY_VOID*)ipd->icn);
   ipd->icn    = (SyIcn*) NULL;
 }

  if (ipd->is_contour_index){
    io_result = fwrite ((SY_VOID*)ipd->contour_index, sizeof(SySkContourIndex), 
			ipd->ipodim+1, fp);

    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)ipd->contour_index);
      ipd->contour_index = NULL;
    }
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    ipd->ipodim = 0;
    ipd->icndim = 0;
  }

  return (1);
}

/**********************************************************************
 * function int Sy_Ipo_fread (fp, ipd)
 * 
 * purpose: gets SyIpoDAC data from a file referenced through fp
 *          (BINARY format)
 *
 * param:   [inout] SyIpoDAC* ipd  -  SyIpoDAC data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Ipo_fread (FILE *fp, SyIpoDAC **ret_ipd, unsigned int iomode)
{
  SyIpoDAC  *ipd;
  int	io_result, bytes = 0;
  SyDataObject *dobj;
  
  ipd = (SyIpoDAC*) Sy_Malloc (sizeof(SyIpoDAC));
  Sy_InitIpo(ipd);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fread", 
	      "missing pointdim info...\n", "");  
  }
  ipd->ipodim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fread", 
	      "missing edgedim info...\n", "");  
  }
  ipd->icndim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "has_contour_index");
  if (dobj == NULL) ipd->is_contour_index = SY_FALSE;
  else ipd->is_contour_index = strcmp(dobj->t.s_val, "true") ? 
	                       SY_FALSE : SY_TRUE; 
  
  ipd->ipo = Sy_AllocIpo (ipd->ipodim);
  ipd->ipl = Sy_AllocIpl (ipd->ipodim);
  ipd->icn = Sy_AllocIcn (ipd->icndim);
  
  if (ipd->is_contour_index){
    ipd->contour_index = Sy_AllocContourIndex(ipd->ipodim);
  }
  
  io_result = fread ((SY_VOID*)ipd->ipo, sizeof(SyIpo), ipd->ipodim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  if (ipd->icndim == 0){ *ret_ipd = ipd; return (bytes); }
  
  io_result = fread ((SY_VOID*)ipd->ipl, sizeof(SyIpl), ipd->ipodim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)ipd->icn, sizeof(SyIcn), ipd->icndim+1, fp);
  
  if (io_result == 0) return (EOF);

  if (ipd->is_contour_index){
    io_result = fread ((SY_VOID*)ipd->contour_index, sizeof(SySkContourIndex), 
		       ipd->ipodim+1, fp);
    if (io_result == 0) return (EOF);
  }

  *ret_ipd = ipd; 
  return (bytes);
}

/**********************************************************************
 * function int Sy_IpoRaw_fscanf (fp, ipd)
 * 
 * purpose: gets SyIpo data from a file referenced through fp
 *          (ASCII format)
 *
 * param:   [inout] SyIpoDAC ipd   -  SyIpoDAC data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_IpoRaw_fscanf (FILE *fp, SyIpoDAC *ipd)
{
  int i, io_result;
  int buff [SY_NBUFF];
  int bytes = 0;
  
  SyDataObject *dobj;

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
	      "missing pointdim info...\n", "");  
  }
  ipd->ipodim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
	      "missing edgedim info...\n", "");  
  }
  ipd->icndim = dobj->t.i_val; 
  
  if (io_result == 0) return (EOF); 
  
  ipd->ipo = Sy_AllocIpo(ipd->ipodim);
  
  for (i = 1; i <= ipd->ipodim; i++){  
    io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]);  
    ipd->ipo[i].x = buff[0]; 
    ipd->ipo[i].y = buff[1];
    
    if (io_result != 2) return (EOF); 
  }
  
  return (bytes);
}

/**********************************************************************
 * function int Sy_IpoRaw_fread (fp, ipd)
 * 
 * purpose: gets SyIpo data only from a file referenced through fp
 *          (BINARY format)
 *
 * param:   [inout] SyIpoDAC* ipd  -  SyIpoDAC data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_IpoRaw_fread (FILE *fp, SyIpoDAC *ipd)
{
	int	io_result, bytes = 0;

	SyDataObject *dobj;

	dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
	if (dobj == NULL){
	  Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
		    "missing pointdim info...\n", "");  
	}
        ipd->ipodim = dobj->t.i_val; 
	
	dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
	if (dobj == NULL){
	  Sy_Error ("Sy_IpoIO", "Sy_Ipo_fscanf", 
		    "missing edgedim info...\n", "");  
	}
	ipd->icndim = dobj->t.i_val; 

        if (io_result == 0) return (EOF); 

	ipd->ipo = Sy_AllocIpo(ipd->ipodim);

	io_result = fread ((SY_VOID*)ipd->ipo, sizeof(SyIpo), ipd->ipodim+1, fp);

        if (io_result == 0) return (EOF); 

        return (bytes);
}

#define SY_CONTOURSET_DIM 512

int Sy_Contour_fscanf (FILE *fp, SySkContourSet **ret_cs, unsigned int iomode)
{
  SySkContourSet *cs;
  int	  i, io_result, pointdim;
  float   buff [SY_NBUFF];
  int     bytes = 0;

  SyDataObject *dobj;
  
  if (*ret_cs == NULL){
    cs = (SySkContourSet*) Sy_Malloc (sizeof(SySkContourSet));
    cs->dim = 0;
    cs->maxdim = SY_CONTOURSET_DIM;
    cs->contour = (SySkContour*) Sy_Malloc (sizeof(SySkContour)*cs->maxdim);
    cs->is_fg_info    = SY_FALSE;
    cs->is_all_closed = SY_TRUE;
    cs->is_polygon    = SY_FALSE;
    *ret_cs = cs;
  }
  else{
    cs = *ret_cs;
  }

  if (cs->dim >= cs->maxdim){
    cs->maxdim += SY_CONTOURSET_DIM;
    cs->contour = (SySkContour*) Sy_Realloc ((SY_VOID*)cs->contour, 
					     sizeof(SySkContour)*cs->maxdim);
  }

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_IpoIO", "Sy_Contour_fscanf", 
	      "missing pointdim info...\n", "");  
  }	
  pointdim = cs->contour[cs->dim].len = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "closed");
  if (dobj == NULL){
    cs->contour[cs->dim].is_closed = SY_TRUE;
    /* assume contour is closed */
  }	
  else{
    if (strcmp(dobj->t.s_val, "true")==0){ 
      cs->contour[cs->dim].is_closed = SY_TRUE;}
    else{        
      cs->contour[cs->dim].is_closed = SY_FALSE;
      cs->is_all_closed = SY_FALSE;
    }
  }

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "polygon");
  if (dobj == NULL){
    cs->contour[cs->dim].is_polygon = SY_FALSE;
    /* assume contour is not described by line segments */
  }	
  else{
    if (strcmp(dobj->t.s_val, "true")==0){ 
      cs->contour[cs->dim].is_polygon = SY_TRUE;
      cs->is_polygon                  = SY_TRUE;
    }
    else{        
      cs->contour[cs->dim].is_polygon = SY_FALSE;
    }
  }
  
  cs->contour[cs->dim].point = (SyIpoFloat*) Sy_Malloc(sizeof(SyIpoFloat)
						       *(pointdim+2));
  for (i = 1; i <= pointdim; i++){
    io_result = fscanf (fp, "%f %f", &buff[0], &buff[1]);  
    cs->contour[cs->dim].point[i].x = buff[0]; 
    cs->contour[cs->dim].point[i].y = buff[1];
    
    if (io_result != 2){
      Sy_Free(cs->contour[cs->dim].point);
      return (EOF); 
    }
  }

  if (syIsVerbose){
    fprintf (stderr, "Read contour no %d.      \r", cs->dim);
  }
  cs->dim++;
 
  return 2*pointdim*sizeof(float);
}

static int Sy_XCompare(SY_CONST_VOID *ipo1, SY_CONST_VOID *ipo2)
{
  return (((SyTmpIpo*)ipo1)->x - ((SyTmpIpo*)ipo2)->x);
}
 
static int Sy_YCompare(SY_CONST_VOID *ipo1, SY_CONST_VOID *ipo2)
{
  return (((SyTmpIpo*)ipo1)->y - ((SyTmpIpo*)ipo2)->y);
}

SyIpoDAC* Sy_ContourSet2Ipo(SySkContourSet *cs, SyControl *in_ctrl)
{
  SyIpoDAC  *ipd;
  float     x, y, x0, y0, x1, y1, 
            level_factor, ori_factor, norm;
  int       i, xint, yint, xint0, yint0, xint1, yint1, 
            xintold, yintold, oldi, oldy;
  int       ptr = 0, cptr, startipoptr, endipoptr,  
            ipoptr = 1, ipodim = 0, is_closed = SY_TRUE, 
            ipomaxdim = 1024, rot;
  SyTmpIpo  *ipo = (SyTmpIpo*)NULL;
  int       *ipotab = (int*)NULL;
  SyVector  bndry0, bndry1;
  SyControl ctrl0, *ctrl;
  SyBoolean err_flag = SY_FALSE;
  
  if (syIsVerbose){
    fprintf (stderr, "Converting contour set with %d elements.\n",
	     cs->dim);
  }

  if (in_ctrl == NULL){
    ctrl = &ctrl0;
    ctrl->is_multiply = SY_FALSE;
    ctrl->mult_f = 1.0;
  }
  else ctrl = in_ctrl;

  if (cs->dim <= 0) return NULL;
  
  ipd = (SyIpoDAC*) Sy_Malloc (sizeof(SyIpoDAC));
  Sy_InitIpo(ipd);

  ipo = (SyTmpIpo*) Sy_Malloc (sizeof(SyTmpIpo)*(ipomaxdim+2));

  for (cptr = 0; cptr < cs->dim; cptr++){

    startipoptr = ipoptr;
    is_closed   = cs->contour[cptr].is_closed;

    x0 = cs->contour[cptr].point[1].x;
    y0 = cs->contour[cptr].point[1].y;
    x1 = cs->contour[cptr].point[cs->contour[cptr].len].x;
    y1 = cs->contour[cptr].point[cs->contour[cptr].len].y;

    xint0 = floor(ctrl->is_multiply ? x0 * ctrl->mult_f : x0 + 0.5); 
    yint0 = floor(ctrl->is_multiply ? y0 * ctrl->mult_f : y0 + 0.5);
    xint1 = floor(ctrl->is_multiply ? x1 * ctrl->mult_f : x1 + 0.5); 
    yint1 = floor(ctrl->is_multiply ? y1 * ctrl->mult_f : y1 + 0.5);

    if (xint1 == xint0 && yint1 == yint0){
      fprintf (stderr, "Removing (last) point no %d from contour no %d\n",
	       cs->contour[cptr].len, cptr);
      fprintf (stderr, "in order to avoid self-intersection.\n");
      cs->contour[cptr].len--;
    }

    for (i = 1; i <= cs->contour[cptr].len; i++){

      if (ipoptr >= ipomaxdim){
	ipomaxdim += 1024;
	ipo = (SyTmpIpo*) Sy_Realloc (ipo, sizeof(SyTmpIpo)*(ipomaxdim+2));
      }

      x    = cs->contour[cptr].point[i].x;
      y    = cs->contour[cptr].point[i].y;
      xint = floor(ctrl->is_multiply ? x * ctrl->mult_f : x + 0.5); 
      yint = floor(ctrl->is_multiply ? y * ctrl->mult_f : y + 0.5);
    
      if ( (ipoptr > 1 && (xint != xintold || yint != yintold)) || ipoptr<=1){
	ipo[ipoptr].x             = xint;
	ipo[ipoptr].y             = yint;
	ipo[ipoptr].ci.contour_id = cptr;
	ipo[ipoptr].ci.point_id   = i;
	ipo[ipoptr].tail_ipo      = ipoptr - 1;
	ipo[ipoptr].head_ipo      = ipoptr + 1;
	ipo[ipoptr].tail_icn      = ipoptr - 1;
	ipo[ipoptr].head_icn      = ipoptr;
	ipo[ipoptr].index         = ipoptr++;
	xintold = xint; yintold = yint;
#ifdef DEBUG
	printf ("input : %d %d\n", xint, yint);
#endif DEBUG
      }
    } /* single contour */
    endipoptr = ipoptr-1;
    ipo[startipoptr].tail_ipo = is_closed ? endipoptr : 0;
    ipo[endipoptr].head_ipo   = is_closed ? startipoptr : 0; 

    if (!is_closed) ipo[endipoptr].head_icn = 0;
    ipo[startipoptr].tail_icn = ipo[endipoptr].head_icn;

    if (cs->is_fg_info && cs->contour[cptr].is_closed){
      
      ipo[endipoptr+1] = ipo[startipoptr];

      level_factor = (cs->contour[cptr].level % 2) ? -1 : 1;
      ori_factor   = cs->contour[cptr].fg_left_side ? 1 : -1;
		      
      rot = level_factor * ori_factor;

      bndry0.dirx = ipo[startipoptr+1].x - ipo[startipoptr].x;
      bndry0.diry = ipo[startipoptr+1].y - ipo[startipoptr].y;
      bndry1.dirx = ipo[startipoptr].x - ipo[endipoptr].x;
      bndry1.diry = ipo[startipoptr].y - ipo[endipoptr].y;

      Sy_NormVector (&bndry0); Sy_NormVector (&bndry1);
      
      switch(rot){
      case 1:
	ipo[startipoptr].grad.dirx = -(bndry0.diry + bndry1.diry); 
	ipo[startipoptr].grad.diry = bndry0.dirx + bndry1.dirx;
	break;
      case -1:
	ipo[startipoptr].grad.dirx = bndry0.diry + bndry1.diry; 
	ipo[startipoptr].grad.diry = -(bndry0.dirx + bndry1.dirx);
	break;
      } /* switch */

      Sy_NormVector (&ipo[startipoptr].grad);

      switch(rot){
      case 1:
	for (i = startipoptr+1; i <= endipoptr; i++){
	  bndry0.dirx = ipo[i+1].x - ipo[i].x;
	  bndry0.diry = ipo[i+1].y - ipo[i].y;
	  bndry1.dirx = ipo[i].x - ipo[i-1].x;
	  bndry1.diry = ipo[i].y - ipo[i-1].y;

	  Sy_NormVector (&bndry0); Sy_NormVector (&bndry1);

	  ipo[i].grad.dirx = -(bndry0.diry + bndry1.diry); 
	  ipo[i].grad.diry = bndry0.dirx + bndry1.dirx;

	  Sy_NormVector (&ipo[i].grad);

#ifdef DEBUG
	  fprintf (stderr, "boundary dir. (x,y) = (%f,%f)\n", 
		   ipo[i].grad.dirx,
		   ipo[i].grad.diry);
#endif
	}
	break;
      case -1:
	for (i = startipoptr+1; i <= endipoptr; i++){
	  bndry0.dirx = ipo[i+1].x - ipo[i].x;
	  bndry0.diry = ipo[i+1].y - ipo[i].y;
	  bndry1.dirx = ipo[i].x - ipo[i-1].x;
	  bndry1.diry = ipo[i].y - ipo[i-1].y;

	  Sy_NormVector (&bndry0); Sy_NormVector (&bndry1);

	  ipo[i].grad.dirx = bndry0.diry + bndry1.diry; 
	  ipo[i].grad.diry = -(bndry0.dirx + bndry1.dirx);

	  Sy_NormVector (&ipo[i].grad);

#ifdef DEBUG
	  fprintf (stderr, "boundary dir. (x,y) = (%f,%f)\n", 
		   ipo[i].grad.dirx,
		   ipo[i].grad.diry);
#endif

	}
	break;
      } /* switch */
    } /* if fg_info */

  } /* all contours */

  ipodim = ipoptr-1;
  ipotab = (int*) Sy_Malloc (sizeof(int)*(ipodim+1));
  
  qsort ((SY_VOID*)(ipo + 1), (int)ipodim, sizeof(SyTmpIpo), Sy_YCompare);
 
  oldy = ipo[1].y; oldi = 1;
  for (i = 2; i <= ipodim; i++){
    if (ipo[i].y != oldy){
      qsort ((SY_VOID*)(ipo+oldi), i-oldi, sizeof(SyTmpIpo), Sy_XCompare);
      oldi = i; oldy = ipo[i].y;
    }
  }

  if (ipodim - oldi >= 1){
    qsort ((SY_VOID*)(ipo+oldi), ipodim-oldi+1, sizeof(SyTmpIpo), Sy_XCompare);
  }

  /* check for self-intersection */

  err_flag = SY_FALSE;
  for (i = 2; i <= ipodim; i++){
    if (ipo[i].x == ipo[i-1].x && ipo[i].y == ipo[i-1].y){
      fprintf (stderr, "ERROR: self-intersection of contour[s] at (%d,%d)\n",
	       ipo[i].x, ipo[i].y);
      err_flag = SY_TRUE;
    }
  }
    
  if (err_flag){
    Sy_SimpleError (SY_FMTNL, "ABORTING program.");
  }
    
  /* end of check */

  for (i = 1; i <= ipodim; i++){ipotab[ipo[i].index] = i;}
  
  for (i = 1; i <= ipodim; i++){
    ipo[i].tail_ipo = ipotab[ipo[i].tail_ipo];
    ipo[i].head_ipo = ipotab[ipo[i].head_ipo];
  }

  ipd->ipodim = ipodim;
  ipd->icndim = ipodim;

  ipd->ipo           = Sy_AllocIpo (ipd->ipodim);
  ipd->ipl           = Sy_AllocIpl (ipd->ipodim);
  ipd->contour_index = Sy_AllocContourIndex (ipd->ipodim);

  ipd->is_contour_index = SY_TRUE;

  for (i = 1; i <= ipodim; i++){
    ipd->ipo[i].x = ipo[i].x;
    ipd->ipo[i].y = ipo[i].y;
    ipd->contour_index[i] = ipo[i].ci;
    ipd->ipl[i].tail_icn = ipo[i].tail_icn;
    ipd->ipl[i].head_icn = ipo[i].head_icn;
  }

  if (cs->is_fg_info){
    for (i = 1; i <= ipodim; i++){
      ipd->ipl[i].grad.x   = ipo[i].grad.dirx;
      ipd->ipl[i].grad.y   = ipo[i].grad.diry;
    }
  }
  
  Sy_Free (ipotab);
  Sy_Free (ipo);

#ifdef DEBUG
  for (i = 1; i <= ipodim; i++){
    fprintf (stderr, "output: %d %d\n", ipd->ipo[i].x, ipd->ipo[i].y);
  }
#endif DEBUG
  return ipd;
}

#define SY_IPOBUFFDIM 1024
static SyWalktask wt;
static int contourdim = 0, ipomaxdim = 0, ipocount = 0;
static SyIpo *ipo = NULL;

static void Sy_CSetOutputWalk(void)
{
  if (ipocount >= ipomaxdim){
    ipomaxdim += SY_IPOBUFFDIM;
    ipo = Sy_ReallocIpo (ipo, ipomaxdim);
  }
  ipo[ipocount++] = wi.ipd->ipo[wi.current.head_ipo];
}

static void Sy_CSetOutputTrav(void)
{
  ipocount = 0; contourdim++;
  if (ipo == NULL){ 
    ipomaxdim = SY_IPOBUFFDIM;
    ipo = Sy_AllocIpo(ipomaxdim);
  }
}

static void Sy_CSetOutputTerm(void)
{
  FILE *fp;
  int i;

  fp = wt.user_file[0];

  fprintf (fp, "object contour_%d {\n", wi.contour_id);
  fprintf (fp, "type = contour_unord; format = ascii;\n");
  fprintf (fp, "contour_id = %d;\n", wi.contour_id);
  fprintf (fp, "closed = true;\n");
  fprintf (fp, "pointdim = %d;\n", ipocount);
  fprintf (fp, "x_extent = 0; y_extent = 0;\n");
  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_TRUE);
  fprintf (fp, "\n");

  for (i = 0; i < ipocount; i++){
    fprintf (fp, "%d %d\n", ipo[i].x, ipo[i].y);  
  }

  Sy_Wrendload (fp, SY_FALSE);
  fprintf (fp, "}\n"); 
}

int Sy_Ipo2ContourSet_fprintf (FILE *fp, char *indata, unsigned int iomode)
{
  SyIpoDAC* ipd;
  int i, io_result;
  FILE *farr[1];

  ipd = (SyIpoDAC*) indata;

  wt.user_file    = farr;
  wt.user_file[0] = fp;
  wt.fildim       = 1;

  wt.user_function[WALK] = (void(*)())Sy_CSetOutputWalk;
  wt.user_function[TRAV] = (void(*)())Sy_CSetOutputTrav;
  wt.user_function[TERM] = (void(*)())Sy_CSetOutputTerm;

  Sy_InitIpoTraverse (ipd, &wt);
  Sy_IpoTraverse (ipd, &wt);  

  Sy_Free(ipo);

  return SY_TRUE;
}

static int Sy_XCompareFPoint(SY_CONST_VOID *ipo1, SY_CONST_VOID *ipo2)
{
  float val;
  val = ((SyTmpFPoint*)ipo1)->x - ((SyTmpFPoint*)ipo2)->x;
  if (val > 0.0) return 1;
  if (val < 0.0) return -1;
  return 0;
}
 
static int Sy_YCompareFPoint(SY_CONST_VOID *ipo1, SY_CONST_VOID *ipo2)
{
  float val;
  val = ((SyTmpFPoint*)ipo1)->y - ((SyTmpFPoint*)ipo2)->y;
  if (val > 0.0) return 1;
  if (val < 0.0) return -1;
  return 0;
}

void Sy_AddFgInfo2ContourSet (SySkContourSet *cs)
{
  float x, y;
  int i, xint, yint, xintold, yintold, oldi, oldy, newptr;
  int ptr = 0, cptr, maxpointdim = 0, pcount = 0, pointdim;
  int curlabel;
  SyTmpFPoint *point = NULL;
  SySkContourTree *tree, *root;

  double angle, arg, det, scale, winding_number_left, winding_number_right;
  SyVector bndry0, ray_left, ray_right, oldray_left, oldray_right;
  SyLocvec leftv, rightv;
  
  if (syIsVerbose){
    fprintf (stderr, "\nComputing foreground/background info.\n");
  }
  if (cs->dim <= 0) return;

  tree = (SySkContourTree*) Sy_Malloc(sizeof(SySkContourTree));
  
  Sy_AddInsideOutsideInfo2ContourSet (cs);

  for (cptr = 0; cptr < cs->dim; cptr++) maxpointdim += cs->contour[cptr].len;
  point = (SyTmpFPoint*) Sy_Malloc (sizeof(SyTmpFPoint)*(maxpointdim+2));

  for (cptr = pcount = 0; cptr < cs->dim; cptr++){
    for (i = 1; i <= cs->contour[cptr].len; i++){
      point[pcount].x = cs->contour[cptr].point[i].x;
      point[pcount].y = cs->contour[cptr].point[i].y;
      point[pcount].contour_id = cptr;
      point[pcount].point_id   = i;
      point[pcount].set        = SY_FALSE;
      pcount++;
    } /* single contour */
  } /* all contours */

  pointdim = pcount;

  qsort ((SY_VOID*)point, pointdim, sizeof(SyTmpFPoint), Sy_YCompareFPoint);
 
  ptr = 0;
  tree->contour_id = point[ptr].contour_id;
  tree->next = NULL; tree->down = NULL;
  root = tree;

  while (ptr < pointdim){
    curlabel = point[ptr].contour_id;
    for (i = ptr; i < pointdim; i++){
      if (point[i].contour_id == curlabel) point[i].set = SY_TRUE;}

    for (i = ptr+1; i < pointdim && (point[i].contour_id == curlabel ||
		      point[i].set); i++);
    newptr = i;
    if (newptr >= pointdim) break;
    
    /* new point is point[newptr] */
    /* compare with tree */

    Sy_InsertSkContourTree (cs, root, &point[newptr], 0);   
    ptr = newptr;
  }/* while */

  cs->is_fg_info = SY_TRUE;
  Sy_Free (point);
  Sy_DeleteSkContourTree(root);
}

void Sy_DeleteSkContourTree(SySkContourTree *t)
{
  if (t->down){
    Sy_DeleteSkContourTree(t->down);
    t->down = NULL;
  }
  if (t->next) {
    Sy_DeleteSkContourTree(t->next);
    t->next = NULL;
  }
  Sy_Free(t);
}

SySkContourTree* Sy_InsertSkContourTree (SySkContourSet *cs, 
					 SySkContourTree *t, 
					 SyTmpFPoint *p, int level)
{
  SySkContourTree *child, *sibling;

  if (Sy_IsWithinSkContour (cs, t, p)){
    if (t->down) Sy_InsertSkContourTree (cs, t->down, p, level+1);
    else {
      t->down = child = 
	(SySkContourTree*) Sy_Malloc(sizeof(SySkContourTree));
      child->next = child->down = NULL;
      child->contour_id = p->contour_id;
      child->point_id = p->point_id;
      child->level = level+1;
      cs->contour[p->contour_id].level = level+1;
      return child;
    }
  }
  else{
    if (t->next) Sy_InsertSkContourTree (cs, t->next, p, level);
    else {
      t->next = sibling = 
	(SySkContourTree*) Sy_Malloc(sizeof(SySkContourTree));
      sibling->next = sibling->down = NULL;
      sibling->contour_id = p->contour_id;
      sibling->point_id = p->point_id;
      sibling->level = level;
      cs->contour[p->contour_id].level = level;
      return sibling;
    }
  }
}

void Sy_DeleteSkContourSet (SySkContourSet *cs)
{
  int cptr;

  if (cs == NULL) return;
  if (cs->dim <= 0) return;
  
  for (cptr = 0; cptr < cs->dim; cptr++){
    if (cs->contour[cptr].point)
      Sy_Free(cs->contour[cptr].point);
  }
  Sy_Free(cs->contour);
}


int Sy_IsWithinSkContour (SySkContourSet *cs, SySkContourTree *t, 
			  SyTmpFPoint *hot)
{
  double angle, arg, det, scale, winding_number;
  SyVector ray, oldray;
  SyLocvec v0;
  SyIpoFloat *p;
  
  int cptr, i, len;
  
  cptr = t->contour_id;
  if (cs->contour[cptr].is_closed == SY_FALSE) return SY_FALSE;

  p   = cs->contour[cptr].point;
  len = cs->contour[cptr].len;
  p[len+1] = p[1];

  /* computing the winding number */
#ifdef DEBUG
  fprintf(stderr, "computing winding number for contour %d\n",
	  cptr);
#endif

  v0.xoff = hot->x;
  v0.yoff = hot->y;

  oldray.dirx  = p[1].x - v0.xoff;
  oldray.diry  = p[1].y - v0.yoff;
    
  winding_number = 0.0;
    
  for (i = 2; i <= len+1; i++){
    
    ray.dirx  = p[i].x - v0.xoff;
    ray.diry  = p[i].y - v0.yoff;

    scale = 1.0/(SY_NORM(ray.dirx, ray.diry)*
		 SY_NORM(oldray.dirx, oldray.diry));
    
    det = SY_DET(oldray.dirx, oldray.diry, ray.dirx, ray.diry);
    
    arg = ( ray.dirx * oldray.dirx + ray.diry * oldray.diry) * scale;
    
    if (arg > 1.0) arg = 1.0; else if (arg < -1.0) arg = -1.0;
    angle = acos(arg) * SY_SGN(det); 
    
    winding_number += angle;
      
    oldray = ray;
  }
    
#ifdef DEBUG
  fprintf(stderr, "WN: %g\n", winding_number);
#endif
  
  /* end computing winding number */    
  /* computing local foreground/background info */
  
  return ( SY_ALV(winding_number) > 3.1416 );
}

void Sy_AddInsideOutsideInfo2ContourSet (SySkContourSet *cs)
{
  int i, cptr, len;
  SyIpoFloat *p;
  SySkContourTree tree;

  double angle, arg, det, scale, winding_number_left, winding_number_right;
  SyVector bndry0, ray_left, ray_right, oldray_left, oldray_right;
  SyLocvec leftv, rightv;
  
  if (syIsVerbose){
    fprintf (stderr, "Computing inside/outside info.\n");
  }
  if (cs->dim <= 0) return;
  
  for (cptr = 0; cptr < cs->dim; cptr++){
    if (cs->contour[cptr].is_closed == SY_FALSE) continue;

    p   = cs->contour[cptr].point;
    len = cs->contour[cptr].len;
    p[len+1] = p[1];

    /* computing the winding number */
#ifdef DEBUG
    fprintf(stderr, "computing winding number\n");
#endif

    bndry0.dirx = p[2].x - p[1].x;
    bndry0.diry = p[2].y - p[1].y;
    leftv.dirx  = -1.0*bndry0.diry; leftv.diry  = bndry0.dirx;
    rightv.dirx = bndry0.diry;      rightv.diry = -1.0*bndry0.dirx;
    scale = 1.0/SY_NORM(leftv.dirx, leftv.diry); 
    leftv.xoff  = 0.5*(p[2].x + p[1].x)+SY_SMALL_OFFSET*scale*leftv.dirx;
    leftv.yoff  = 0.5*(p[2].y + p[1].y)+SY_SMALL_OFFSET*scale*leftv.diry;
    rightv.xoff = 0.5*(p[2].x + p[1].x)+SY_SMALL_OFFSET*scale*rightv.dirx;
    rightv.yoff = 0.5*(p[2].y + p[1].y)+SY_SMALL_OFFSET*scale*rightv.diry;

    oldray_left.dirx  = p[1].x - leftv.xoff;
    oldray_left.diry  = p[1].y - leftv.yoff;
    oldray_right.dirx = p[1].x - rightv.xoff;
    oldray_right.diry = p[1].y - rightv.yoff;
    
    winding_number_left = winding_number_right = 0.0;
    
    for (i = 2; i <= len+1; i++){
    
      ray_left.dirx  = p[i].x - leftv.xoff;
      ray_left.diry  = p[i].y - leftv.yoff;
      ray_right.dirx = p[i].x - rightv.xoff;
      ray_right.diry = p[i].y - rightv.yoff;

      /* winding number left */

      scale = 1.0/(SY_NORM(ray_left.dirx, ray_left.diry)*
		   SY_NORM(oldray_left.dirx, oldray_left.diry));
      
      det = SY_DET(oldray_left.dirx, oldray_left.diry,
		   ray_left.dirx, ray_left.diry);
      
      arg = ( ray_left.dirx * oldray_left.dirx + ray_left.diry *
	      oldray_left.diry) * scale;

      if (arg > 1.0) arg = 1.0; else if (arg < -1.0) arg = -1.0;
      angle = acos(arg) * SY_SGN(det); 
      
      winding_number_left += angle;
      
      /* winding number right */
      
      scale = 1.0/(SY_NORM(ray_right.dirx, ray_right.diry)*
		   SY_NORM(oldray_right.dirx, oldray_right.diry));
      
      det = SY_DET(oldray_right.dirx, oldray_right.diry,
		   ray_right.dirx, ray_right.diry);
      
      arg = ( ray_right.dirx * oldray_right.dirx + ray_right.diry *
	      oldray_right.diry) * scale;
      
      if (arg > 1.0) arg = 1.0; else if (arg < -1.0) arg = -1.0;
      angle = acos(arg) * SY_SGN(det); 

      winding_number_right += angle;

      oldray_left  = ray_left;
      oldray_right = ray_right;
    }
#ifdef DEBUG    
    fprintf(stderr, "WNL: %g, WNR: %g\n", winding_number_left,
	    winding_number_right);
#endif
  
    /* end computing winding number */    
    /* computing local foreground/background info */

    cs->contour[cptr].fg_left_side = 
      SY_ALV(winding_number_left) > SY_ALV(winding_number_right);
    
  }
}

#define SY_FIPOBUFDIM 16

static SyIpoFloat *fipo = NULL;
static int s_maxbufdim  = 0,
           s_bufdim     = 0,
           s_cptr       = 0,
           s_res        = 1.0,
           s_p          = 0;

static int s_oldx, s_oldy;
static int s_xa, s_ya, s_xb, s_yb;

static void Sy_FillPointBuf (int x, int y)
{
  if (s_bufdim >= s_maxbufdim){
    s_maxbufdim += SY_FIPOBUFDIM;
    fipo = (SyIpoFloat*) Sy_Realloc(fipo, sizeof(SyIpoFloat)*s_maxbufdim);
  }

  if (x == s_xa && y == s_ya){
    s_oldx = x; s_oldy = y;
    fipo[s_bufdim].x = x;
    fipo[s_bufdim].y = y;
    s_bufdim++;
    return;
  }

  if (x == s_xb && y == s_yb){
    fipo[s_bufdim].x = x;
    fipo[s_bufdim].y = y;
    s_bufdim++;
    return;
  } 

  if ( (sq(x-s_oldx)+sq(y-s_oldy)) >= s_res || s_res == 1.0){
    fipo[s_bufdim].x = x;
    fipo[s_bufdim].y = y;
    s_bufdim++;
    s_oldx = x; s_oldy = y;
  }
    
}

SySkContourSet *Sy_Polygon2Points (SySkContourSet *cs, SyControl *in_ctrl)
{
  SySkContourSet *rcs = NULL;
  SyControl      ctrl0, *ctrl;

  float          x, y, x0, y0, x1, y1;

  int            xint, yint, xint0, yint0, xint1, yint1, 
                 xintstart, yintstart;

  int cptr, i;

  if (syIsVerbose){
    fprintf (stderr, "Conversion: Polygon -> Dense Point Array...");
  }

  if (in_ctrl == NULL){
    ctrl = &ctrl0;
    ctrl->is_multiply = SY_FALSE;
    ctrl->mult_f      = 1.0;
    ctrl->resolution  = 1.0;
  }
  else ctrl = in_ctrl;

  s_res = sq(ctrl->resolution);

  rcs = (SySkContourSet*) Sy_Malloc (sizeof(SySkContourSet));
  rcs->dim    = cs->dim;
  rcs->maxdim = cs->dim;
  rcs->contour = (SySkContour*) Sy_Malloc (sizeof(SySkContour)*(cs->dim));
  rcs->is_fg_info    = cs->is_fg_info;
  rcs->is_all_closed = cs->is_all_closed;
  rcs->is_polygon    = SY_FALSE;

  s_maxbufdim = SY_FIPOBUFDIM;
  fipo = (SyIpoFloat*) Sy_Malloc(sizeof(SyIpoFloat)*s_maxbufdim);
  
  for (cptr = 0; cptr < cs->dim; cptr++){
    rcs->contour[cptr]            = cs->contour[cptr];
    rcs->contour[cptr].point      = NULL;
    rcs->contour[cptr].is_polygon = SY_FALSE;

    if (cs->contour[cptr].is_polygon == SY_FALSE){
      rcs->contour[cptr].point = 
	(SyIpoFloat*) Sy_Malloc (sizeof(SyIpoFloat)*(cs->contour[cptr].len+2));
      memcpy (rcs->contour[cptr].point, cs->contour[cptr].point, 
	      sizeof(SyIpoFloat)*(cs->contour[cptr].len+1));
      continue;
    }

    x0 = cs->contour[cptr].point[1].x;
    y0 = cs->contour[cptr].point[1].y;
    x1 = cs->contour[cptr].point[cs->contour[cptr].len].x;
    y1 = cs->contour[cptr].point[cs->contour[cptr].len].y;

    xint0 = floor(ctrl->is_multiply ? x0 * ctrl->mult_f : x0 + 0.5); 
    yint0 = floor(ctrl->is_multiply ? y0 * ctrl->mult_f : y0 + 0.5);
    xint1 = floor(ctrl->is_multiply ? x1 * ctrl->mult_f : x1 + 0.5); 
    yint1 = floor(ctrl->is_multiply ? y1 * ctrl->mult_f : y1 + 0.5);

    xintstart = xint0;
    yintstart = yint0;

    if (xint1 == xint0 && yint1 == yint0){
      fprintf (stderr, "Removing (last) point no %d from contour no %d\n",
	       cs->contour[cptr].len, cptr);
      fprintf (stderr, "in order to avoid self-intersection.\n");
      cs->contour[cptr].len--;
    }

    xint1 = xint0;
    yint1 = yint0;

    s_bufdim  = 0;
    s_cptr    = cptr;

    for (i = 2; i <= cs->contour[cptr].len; i++){
      s_p = i;

      xint0 = xint1;
      yint0 = yint1;

      x1 = cs->contour[cptr].point[i].x;
      y1 = cs->contour[cptr].point[i].y;
      xint1 = floor(ctrl->is_multiply ? x1 * ctrl->mult_f : x1 + 0.5); 
      yint1 = floor(ctrl->is_multiply ? y1 * ctrl->mult_f : y1 + 0.5);

      s_oldx = s_xa = xint0; s_oldy = s_ya = yint0;
      s_xb = xint1; s_yb = yint1;

      Sy_DrawDigitalLine (xint0, yint0, xint1, yint1, Sy_FillPointBuf);
 
    } /* for all points */

    s_oldx = s_xa = xint1; s_oldy = s_ya = yint1;
    s_xb = xintstart; s_yb = yintstart;

    Sy_DrawDigitalLine (xint1, yint1, xintstart, yintstart, Sy_FillPointBuf);

    s_bufdim--; /* preventing self-intersection */

    rcs->contour[cptr].len   = s_bufdim;
    rcs->contour[cptr].point = 
      (SyIpoFloat*) Sy_Malloc (sizeof(SyIpoFloat)*(s_bufdim+2));
    memcpy (rcs->contour[cptr].point+1, fipo, sizeof(SyIpoFloat)*(s_bufdim+1));
    
  } /* for all contours */

  return rcs;
  Sy_Free(fipo);
}

