/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */


static char sy_SkelFixId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

static void Sy_InfixTrav(void);
static void Sy_InfixTerm(void);
static void Sy_InfixWalk(void);

static SySkelDAC *vsk;
static SyIpoDAC   *ipd;
static float    maxbirth0;
static int      skelid,
                maxvertex,
                maxedge;

void Sy_Infix (SySkelDAC *vsk_in, SyIpoDAC *ipd_in)
{
  SyWalktask wt;

  vsk = vsk_in;
  ipd = ipd_in;

  Sy_ResetWalktask (&wt);
 
  wt.user_function[(int)TRAV] = (void(*)())Sy_InfixTrav;
  wt.user_function[(int)TERM] = (void(*)())Sy_InfixTerm; 
  wt.user_function[(int)WALK] = (void(*)())Sy_InfixWalk;
  
  Sy_SkeletonTraverse (vsk, ipd, SY_FIX_THRES, 0, &wt);
}

static void Sy_InfixTrav(void)
{
  sprintf (syStrBuf, "traverse starting, skel id = %d ", syWstate.id);
  Sy_Message (SY_FMTNL, syStrBuf);
  maxbirth0 = SY_NEG_INFTY;
  maxedge   = 0;
  maxvertex = 0;
  skelid    = syWstate.id;
}


static void Sy_InfixWalk(void)
{
  float birth0;

  birth0 = syWstate.vsk->regul[syWstate.current.head_vertex].birth[0]; 
  if ((birth0 > maxbirth0) && (birth0 < SY_POS_INFTY)){
    maxbirth0 = birth0;
    maxvertex = syWstate.current.head_vertex;
    maxedge   = syWstate.current.next_edge;
  }
}

static void Sy_InfixTerm(void)
{
  int i, 
      ref_vertex, ref_edge,
      next_edge,
      tail_vertex, head_vertex;

  float p_birth, *birth_ptr;

  if (maxedge == 0) return;

  ref_vertex = maxvertex;
  ref_edge   = maxedge;
  p_birth    = SY_NEG_INFTY;

  if (vsk->edge[ref_edge].tail_vertex == ref_vertex){
    next_edge = vsk->edge[ref_edge].tail_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].head_vertex].birth[0]);

  }
  else{
    next_edge = vsk->edge[ref_edge].head_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].tail_vertex].birth[0]);
  }

  if (next_edge == 0) return;

  if ((birth_ptr[2] > p_birth) && (birth_ptr[2] < SY_FIX_THRES)) p_birth = birth_ptr[2];
  else
  if ((birth_ptr[1] > p_birth) && (birth_ptr[1] < SY_FIX_THRES)) p_birth = birth_ptr[1];
  else
  if ((birth_ptr[0] > p_birth) && (birth_ptr[0] < SY_FIX_THRES)) p_birth = birth_ptr[0];


  while (next_edge != ref_edge){

    if (ref_vertex == vsk->edge[next_edge].tail_vertex){
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].head_vertex].birth[0]);
      next_edge = vsk->edge[next_edge].tail_edge;
    }
    else{
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].tail_vertex].birth[0]);
      next_edge = vsk->edge[next_edge].head_edge;
    }

    if ((birth_ptr[2]>p_birth) && (birth_ptr[2]< SY_FIX_THRES)) p_birth = birth_ptr[2];
    else
    if ((birth_ptr[1]>p_birth) && (birth_ptr[1]< SY_FIX_THRES)) p_birth = birth_ptr[1];
    else
    if ((birth_ptr[0]>p_birth) && (birth_ptr[0]< SY_FIX_THRES)) p_birth = birth_ptr[0];
  }

  if (p_birth > maxbirth0){
    sprintf (syStrBuf, "fixing skeleton nr %d... ", skelid);
    Sy_Message (SY_FMTNL, syStrBuf);

    for (i = 1; i <= vsk->vdim; i++){

      if (vsk->label.v_[i] == skelid){

	if (vsk->regul[i].birth[0] > SY_FIX_THRES)
	  vsk->regul[i].birth[0] = maxbirth0;

	if (vsk->regul[i].birth[1] > SY_FIX_THRES)
	  vsk->regul[i].birth[1] = maxbirth0;

	if (vsk->regul[i].birth[2] > SY_FIX_THRES)
	  vsk->regul[i].birth[2] = maxbirth0;
      }
    }

    if (vsk->erdim  > 0){
      for (i = 1; i <= vsk->edim; i++){

	if (vsk->label.e_[i] == skelid){
	  vsk->eresid[i].tail_residual = 
          vsk->eresid[i].head_residual = maxbirth0;
	}      
      }
    } /* if */
  }
}


static void Sy_MaxfixTrav(void);
static void Sy_MaxfixTerm(void);
static void Sy_MaxfixWalk(void);

void Sy_Maxfix (SySkelDAC *vsk_in, SyIpoDAC *ipd_in)
{
  SyWalktask wt;

  vsk = vsk_in;
  ipd = ipd_in;

  Sy_ResetWalktask (&wt);
 
  wt.user_function[(int)TRAV] = (void(*)())Sy_MaxfixTrav;
  wt.user_function[(int)TERM] = (void(*)())Sy_MaxfixTerm; 
  wt.user_function[(int)WALK] = (void(*)())Sy_MaxfixWalk;
  
  Sy_SkeletonTraverse (vsk, ipd, 0.0, 0, &wt);
}


static float maxbirth2;


static void Sy_MaxfixTrav(void)
{
  sprintf (syStrBuf, "starting, skel id = %d ", syWstate.id);
  Sy_Message (SY_FMTNL, syStrBuf);
  maxbirth2 = SY_NEG_INFTY;
  maxedge   = 0;
  maxvertex = 0;
  skelid    = syWstate.id;
}


static void Sy_MaxfixWalk(void)
{
  float birth2;

  birth2 = syWstate.vsk->regul[syWstate.current.head_vertex].birth[2]; 
  if ((birth2 > maxbirth2) && (birth2 < SY_POS_INFTY)){
    maxbirth2 = birth2;
    maxvertex = syWstate.current.head_vertex;
    maxedge   = syWstate.current.next_edge;
  }
}

static void Sy_MaxfixTerm(void)
{
  int i, 
      ref_vertex, ref_edge,
      next_edge, p_edge;

  float p_birth, *birth_ptr;

  if (maxedge == 0) return;

  ref_vertex = maxvertex;
  ref_edge   = maxedge;
  p_birth    = SY_NEG_INFTY;
  p_edge     = maxedge;

  if (vsk->edge[ref_edge].tail_vertex == ref_vertex){
    next_edge = vsk->edge[ref_edge].tail_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].head_vertex].birth[0]);
  }
  else{
    next_edge = vsk->edge[ref_edge].head_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].tail_vertex].birth[0]);
  }

  if (birth_ptr[2] > p_birth) p_birth = birth_ptr[2];
  
  while ((next_edge != ref_edge) && (next_edge != 0)){

    if (ref_vertex == vsk->edge[next_edge].tail_vertex){
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].head_vertex].birth[0]);
      next_edge = vsk->edge[next_edge].tail_edge;
    }
    else{
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].tail_vertex].birth[0]);
      next_edge = vsk->edge[next_edge].head_edge;
    }

    if (birth_ptr[2]>p_birth){
      p_birth = birth_ptr[2]; p_edge = next_edge;
    }
  }

  if (p_birth < SY_POS_INFTY){
    vsk->regul[vsk->edge[p_edge].tail_vertex].birth[2] = SY_POS_INFTY;
    vsk->regul[vsk->edge[p_edge].head_vertex].birth[2] = SY_POS_INFTY;
    if (vsk->erdim > 0){
      vsk->eresid[p_edge].tail_residual = SY_POS_INFTY;
      vsk->eresid[p_edge].head_residual = SY_POS_INFTY;
    }
  }
}
