/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * 
 * Skeleton Pyramid 
 *
 */ 

static char sy_SkelPyramidId[] = "$Id$";

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

static SySkelDAC *vsk;
static SyIpoDAC   *ipd;

static 
SySkelTable      skeltab;

static int      *pyrlab;
static int      *vlabel;
static float    pdelta, pth;
static
SyRegulControl  *rctmp;

static float    maxbirth0, maxbirth2, maxbirth2_aux;
static int      skelid,
                maxvertex, maxvertex_aux,
                maxedge, maxedge_aux,
                verbose_local = 0;

static int      depth, max_depth=0, nrec;

static void Sy_PyramidTrav(void);
static void Sy_PyramidTerm(void);
static void Sy_PyramidWalk(void);
static void Sy_PyramidGenerate(int *next_edge, int *current_head_vertex, 
			       int skelid, int rank, SyRegulControl *rc);

static int  Sy_ResBufCompare(SY_CONST_VOID *a, SY_CONST_VOID *b);

void Sy_Pyramid (SySkelDAC *vsk_in, SyIpoDAC *ipd_in, SyRegulControl *rc)
{
  SyWalktask wt;
  int      start_edge, tail_vertex, head_vertex, skelid;
  float    max_residual_in_table, median_residual;
  float    *res_buf = NULL;
  int      rank, res_buf_count, i;

  if (syIsVerbose){
    Sy_Message (SY_FMTNL, 
		"Adding hierarchical information (Skeleton Pyramid):");
  }

  rctmp  = rc;
  vsk    = vsk_in;
  ipd    = ipd_in;

  if (rc->is_pyr_cluster_hint){
    pdelta = rc->pyr_cluster_hint/100.0;}
  else{
    pdelta = rc->pyr_cluster_delta/100.0;}

  pth    = rc->pyr_trigger_threshold/100.0;

  pyrlab = vsk->hier = (int*) Sy_Malloc ((vsk->edim+1)*sizeof(int));
  if (vsk->hier == NULL){
    Sy_SimpleError (SY_FMTNL, "Allocation of pyramid labels failed... ");
    exit(2);
  }
  vsk->is_hierarchy = SY_TRUE;
  memset ((char*)vsk->hier, 0, (vsk->edim+1)*sizeof(int)); 

  vlabel = (int*) Sy_Malloc((vsk->vdim+1)*sizeof(int));
  memset((char*) vlabel, 0, (vsk->vdim+1)*sizeof(int));

  skeltab.size  = SY_INIT_SKEL_TAB_SIZE;
  skeltab.count = 0;
  skeltab.elem  = 
    (SySkelTableElem*) Sy_Malloc (SY_INIT_SKEL_TAB_SIZE*sizeof(SySkelTableElem));

  Sy_ResetWalktask (&wt);
 
  wt.user_function[(int)TRAV] = (void(*)())Sy_PyramidTrav;
  wt.user_function[(int)TERM] = (void(*)())Sy_PyramidTerm; 
  wt.user_function[(int)WALK] = (void(*)())Sy_PyramidWalk;
  
  wt.is_search_threshold      = rc->is_search_threshold;
  wt.search_thres = (rc->is_search_threshold) ? rc->search_threshold :
                                                rc->term_threshold;

  if (wt.search_thres < rc->term_threshold){
    wt.search_thres = rc->term_threshold;
  }

  Sy_InitSkelTraverse (vsk, ipd, 0, &wt);
  Sy_SkeletonTraverse (vsk, ipd, rc->term_threshold, 0, &wt);

  if (skeltab.count == 0){
    Sy_SimpleError (SY_FMTNL, "No skeletons found ! ");
    exit(2);
  }

  max_residual_in_table = skeltab.elem[0].max_residual;
  for (i = 1; i < skeltab.count; i++){
    if (skeltab.elem[i].max_residual > max_residual_in_table)
      max_residual_in_table = skeltab.elem[i].max_residual;
  }

  res_buf = (float*) Sy_Malloc (skeltab.count*sizeof(float));
  
  for (i = 0, res_buf_count = 0; i < skeltab.count; i++){
    if (skeltab.elem[i].is_foreground)
      res_buf[res_buf_count++] = skeltab.elem[i].max_residual;
  }  

  qsort (res_buf, res_buf_count, sizeof (res_buf[0]), Sy_ResBufCompare);
  median_residual = res_buf[res_buf_count/2];

#ifdef DEBUG
  sprintf (syStrBuf, "Max. residual in table: %f ", max_residual_in_table);
  Sy_Message (SY_FMTNL, syStrBuf);
  sprintf (syStrBuf, "Median value of foreground max. resid.: %f ", 
	   median_residual);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  for (i = 0; i < skeltab.count; i++){
    start_edge = skeltab.elem[i].max_edge;
    maxbirth2  = skeltab.elem[i].is_foreground ? 
      skeltab.elem[i].max_residual : median_residual;
    skelid     = skeltab.elem[i].skelid;
 
    if (rc->is_pyr_trigger_hint){
      Sy_Message (SY_FMTNL, 
		  "Using information from skeleton-space evaluation. ");
      sprintf (syStrBuf, "Descending skeleton %d, thresh.: %4.1f",
	       skelid, rc->pyr_trigger_hint);
      Sy_Message (SY_FMT, syStrBuf);

      if (rc->is_pyr_cluster_hint){
	sprintf (syStrBuf, ", branch clustering set to %4.1f\%", 
		 rc->pyr_cluster_hint);}
      Sy_Message (SY_FMTNL, syStrBuf);
    }
    else{
      sprintf (syStrBuf, 
	 "Descending skeleton %d, thresh.: %4.1f. ",
	       skelid, max(pth*maxbirth2,SY_TPYR_MIN));
      Sy_Message (SY_FMTNL, syStrBuf);
    }

    depth = nrec = 0;

    if (vsk->label.e_[start_edge] == skelid) pyrlab[start_edge] = rank = 1;
    tail_vertex = vsk->edge[start_edge].tail_vertex;
    head_vertex = vsk->edge[start_edge].head_vertex; 
 
    Sy_PyramidGenerate (&start_edge, &tail_vertex, skelid, rank=1, rc); 
    start_edge = skeltab.elem[i].max_edge;
    Sy_PyramidGenerate (&start_edge, &head_vertex, skelid, rank=1, rc);     
  }

  Sy_Free(res_buf);
  Sy_Free(vsk->label.e_); vsk->label.e_ = NULL;
  Sy_Free(vsk->label.v_); vsk->label.v_ = NULL;
}

static void Sy_PyramidTrav(void)
{
#ifdef DEBUG
  sprintf (syStrBuf, "Looking for maxval at skel id = %d... ", syWstate.id);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif
  maxbirth2 = maxbirth2_aux = SY_NEG_INFTY;
  maxedge   = maxedge_aux = 0;
  maxvertex = maxvertex_aux = 0;
  skelid    = syWstate.id;
}


static void Sy_PyramidWalk(void)
{
  float birth2;
  SyIndex next_edge, tail_vertex, head_vertex;

  birth2 = syWstate.vsk->regul[syWstate.current.head_vertex].birth[2]; 

  if (birth2 > maxbirth2_aux){
    maxbirth2_aux = birth2;
    maxvertex_aux = syWstate.current.head_vertex;
    maxedge_aux   = syWstate.current.next_edge;
  }

  if (vsk->is_smcir == SY_TRUE){
    next_edge  = syWstate.current.next_edge;
    tail_vertex = syWstate.vsk->edge[next_edge].tail_vertex;
    head_vertex = syWstate.vsk->edge[next_edge].head_vertex;
    if (vsk->regul_aux[tail_vertex].birth[2] < rctmp->smcir_search_thres ||
	vsk->regul_aux[head_vertex].birth[2] < rctmp->smcir_search_thres){
      return;
    }
  }

  if ((birth2 > maxbirth2) && (birth2 < SY_POS_INFTY)){
    maxbirth2 = birth2;
    maxvertex = syWstate.current.head_vertex;
    maxedge   = syWstate.current.next_edge;
  }
}

static void Sy_PyramidTerm(void)
{
  int i, head_vertex, tail_vertex,
      ref_vertex, ref_edge,
      next_edge, next_edge_buf, p_edge, start_edge, rank;

  float p_birth, *birth_ptr;

  if (maxedge == 0){
    ref_vertex   = maxvertex_aux;
    ref_edge     = maxedge_aux;
    p_birth      = SY_NEG_INFTY;
    p_edge       = maxedge_aux;
    maxbirth2    = maxbirth2_aux;
  }
  else{
    ref_vertex   = maxvertex;
    ref_edge     = maxedge;
    p_birth      = SY_NEG_INFTY;
    p_edge       = maxedge;
  }

  if (vsk->edge[ref_edge].tail_vertex == ref_vertex){
    next_edge = vsk->edge[ref_edge].tail_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].head_vertex].birth[0]);
  }
  else{
    next_edge = vsk->edge[ref_edge].head_edge;
    birth_ptr = &(vsk->regul[vsk->edge[ref_edge].tail_vertex].birth[0]);
  }

  if (birth_ptr[2] > p_birth){ p_birth = birth_ptr[2]; }
  
  while ((next_edge != ref_edge) && (next_edge != 0)){

    if (ref_vertex == vsk->edge[next_edge].tail_vertex){
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].head_vertex].birth[0]);
      next_edge_buf = vsk->edge[next_edge].tail_edge;
    }
    else{
      birth_ptr = &(vsk->regul[vsk->edge[next_edge].tail_vertex].birth[0]);
      next_edge_buf = vsk->edge[next_edge].head_edge;
    }

    if (birth_ptr[2] > p_birth){
      p_birth = birth_ptr[2]; p_edge = next_edge;
    }

    next_edge = next_edge_buf;
  }

#ifdef DEBUG
  sprintf (syStrBuf, "Maxval %f at vertex %d, edge %d, (th: %f) ", 
	   maxbirth2, ref_vertex, p_edge, 
	   (rctmp->is_pyr_trigger_hint) ? 
	   rctmp->pyr_trigger_hint : pth*maxbirth2);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  skeltab.elem[skeltab.count].max_residual  = maxbirth2;
  skeltab.elem[skeltab.count].max_edge      = p_edge;
  skeltab.elem[skeltab.count].skelid        = skelid;
  skeltab.elem[skeltab.count].is_foreground = syWstate.foreground;

  skeltab.count++;
  if (skeltab.count >= skeltab.size){
    skeltab.size += SY_SKEL_TAB_INCR;
    skeltab.elem = (SySkelTableElem*) 
      Sy_Realloc((char*) skeltab.elem, skeltab.size*sizeof(SySkelTableElem));
  }
}

static int Sy_ResCompare(SY_CONST_VOID *ra, SY_CONST_VOID *rb);


/* static counter = 0; */

static void Sy_PyramidGenerate (int *next_edge, int *current_head_vertex, 
				int skelid, int rank, SyRegulControl *rc)
{
  SyResTable *res_tab = NULL;
  int res_tab_dim, tab_ptr, opcount=0, branching_count = 0;

  int current_head_vertex_buff, 
      next_edge_buff, rank_buff, ref_edge, ref_vertex;

  float last_residual, current_residual;

  int tail_vertex, head_vertex, 
      right_point, left_point;

  SyBoolean is_void_nfurcation, is_void_path, 
            is_void_next_call = SY_FALSE;

  float *birth_ptr;

  if (*next_edge == 0) return; 

  if (vlabel[*current_head_vertex] != 0) return;

  current_head_vertex_buff = ref_vertex = *current_head_vertex;
  next_edge_buff = ref_edge             = *next_edge;

  is_void_path = (rank == SY_VOIDLABEL) ? SY_TRUE : SY_FALSE;

  if (vsk->regul[ref_vertex].is_void_nfurcation == SY_TRUE &&
      vsk->regul[ref_vertex].birth[0] > rc->term_threshold){
    if (syIsVerbose){
      sprintf(syStrBuf, "VOID N-FURCATION <%d> ENCOUNTERED ", ref_vertex);
      Sy_Message (SY_FMTNL, syStrBuf);
    }
    is_void_path = SY_FALSE ; 
  }

  if (is_void_path == SY_FALSE && vsk->is_smcir == SY_TRUE){
    tail_vertex = syWstate.vsk->edge[ref_edge].tail_vertex;
    head_vertex = syWstate.vsk->edge[ref_edge].head_vertex;
    if (vsk->regul_aux[tail_vertex].birth[2] < rc->smcir_term_thres ||
	vsk->regul_aux[head_vertex].birth[2] < rc->smcir_term_thres){
      is_void_path = SY_TRUE;}
  }

  rank_buff = is_void_path ? SY_VOIDLABEL : rank;

  if (vsk->edge[*next_edge].tail_vertex == *current_head_vertex){
    next_edge_buff = vsk->edge[*next_edge].tail_edge;
  }
  else{
    next_edge_buff = vsk->edge[*next_edge].head_edge;
  }

  /* build residual table for vertex */
  res_tab_dim = SY_INIT_RES_TAB_SIZE;
  res_tab = (SyResTable*) Sy_Malloc (res_tab_dim*sizeof(SyResTable));
  tab_ptr = 0;

  if (vsk->is_eresid){

    while ((next_edge_buff != ref_edge) && (next_edge_buff != 0)){

      if (vsk->label.e_[next_edge_buff] == skelid){
	if (vsk->edge[next_edge_buff].tail_vertex == ref_vertex){
	  res_tab[tab_ptr].residual = vsk->eresid[next_edge_buff].tail_residual;
	  res_tab[tab_ptr].edge_ref = next_edge_buff;
	}
	else{
	  res_tab[tab_ptr].residual = vsk->eresid[next_edge_buff].head_residual;
	  res_tab[tab_ptr].edge_ref = next_edge_buff;
	}
	tab_ptr++;
	if (tab_ptr >= res_tab_dim){
	  res_tab_dim += SY_RES_TAB_INCR;
	  res_tab = (SyResTable*) Sy_Realloc((char*) res_tab, 
					     res_tab_dim*sizeof(SyResTable));
	}
      }
      if (vsk->edge[next_edge_buff].tail_vertex == ref_vertex){
	next_edge_buff = vsk->edge[next_edge_buff].tail_edge;
      }
      else{
	next_edge_buff = vsk->edge[next_edge_buff].head_edge;
      }      
    }
  }
  else{

    while ((next_edge_buff != ref_edge) && (next_edge_buff != 0)){
      
      if (vsk->label.e_[next_edge_buff] == skelid){
	if (vsk->edge[next_edge_buff].tail_vertex == ref_vertex){
	  res_tab[tab_ptr].residual = 
	    vsk->regul[vsk->edge[next_edge_buff].head_vertex].birth[2];
	  res_tab[tab_ptr].edge_ref = next_edge_buff;
	}
	else{
	  res_tab[tab_ptr].residual = 
	    vsk->regul[vsk->edge[next_edge_buff].tail_vertex].birth[2];
	  res_tab[tab_ptr].edge_ref = next_edge_buff;
	}
	tab_ptr++;
	if (tab_ptr >= res_tab_dim){
	  res_tab_dim += SY_RES_TAB_INCR;
	  res_tab = (SyResTable*) Sy_Realloc((char*) res_tab, 
					     res_tab_dim*sizeof(SyResTable));
	}
      }
      if (vsk->edge[next_edge_buff].tail_vertex == ref_vertex){
	next_edge_buff = vsk->edge[next_edge_buff].tail_edge;
      }
      else{
	next_edge_buff = vsk->edge[next_edge_buff].head_edge;
      }      
    }
  }

  if (tab_ptr == 0){
    Sy_Free(res_tab);
    return;
  }

  qsort (res_tab, tab_ptr, sizeof (SyResTable), Sy_ResCompare);

  vlabel[*current_head_vertex] = skelid;
  last_residual = res_tab[tab_ptr-1].residual;
  
  while (tab_ptr > 0){
    *next_edge       = res_tab[tab_ptr-1].edge_ref; opcount++;
    current_residual = res_tab[tab_ptr-1].residual;

    if (is_void_path == SY_FALSE && vsk->is_smcir == SY_TRUE){
      is_void_next_call = SY_FALSE;
      tail_vertex = syWstate.vsk->edge[*next_edge].tail_vertex;
      head_vertex = syWstate.vsk->edge[*next_edge].head_vertex;
      if (vsk->regul_aux[tail_vertex].birth[2] < rc->smcir_term_thres ||
	  vsk->regul_aux[head_vertex].birth[2] < rc->smcir_term_thres){
#ifdef DEBUG
	Sy_Message(SY_FMTNL, "void branch encountered");
#endif
	is_void_next_call = SY_TRUE;}
    }
    
    if (res_tab[tab_ptr-1].residual < 
	((rc->is_pyr_trigger_hint) ? rc->pyr_trigger_hint : pth*maxbirth2)){
#ifdef DEBUG
      fprintf(stderr, "path continued\n");
#endif
      if (opcount > 1){
#ifdef DEBUG
	fprintf(stderr, "opcount > 1\n");
#endif
	if ((last_residual - current_residual) > last_residual*
	    (rc->is_pyr_cluster_hint ? 
	     ( (last_residual-1.0)*pdelta/(rc->cluster_max-1.0) ) : pdelta)){
#ifdef DEBUG
          fprintf(stderr, "new branch\n");
#endif
	  last_residual = current_residual;
	  if (!is_void_path) { rank_buff++;  branching_count++; }
	}
      }
    }

    if (rc->is_search_threshold && rc->is_term_threshold){
      if ((current_residual < rc->search_threshold) && (opcount > 1)){
#ifdef DEBUG
	sprintf(syStrBuf, "Halting rotation at vertex %d ", ref_vertex);
	Sy_Message (SY_FMTNL, syStrBuf);
#endif
	break;  /* exit from while loop */
      }
    }

    if (vsk->edge[*next_edge].tail_vertex == ref_vertex){
      *current_head_vertex = vsk->edge[*next_edge].head_vertex;
    }
    else{
      *current_head_vertex = vsk->edge[*next_edge].tail_vertex;
    }
    
    if (++depth > max_depth) max_depth = depth; nrec++;
#ifdef DEBUG
    if (syIsVerbose){
      sprintf (syStrBuf, "> depth:%6d, max.depth:%6d, nrec:%6d\r",
	       depth, max_depth, nrec);
      Sy_Message (SY_FMT, syStrBuf);}
#endif
    
    tab_ptr--;
#ifdef DEBUG
    sprintf (syStrBuf, "rank-> %d ", rank_buff);
    Sy_Message (SY_FMTNL, syStrBuf);
#endif
    pyrlab[*next_edge] = is_void_next_call ? SY_VOIDLABEL : rank_buff;
    Sy_PyramidGenerate(next_edge, current_head_vertex, skelid, 
		       is_void_next_call ? SY_VOIDLABEL : rank_buff, rc);
    
    depth--;

#ifdef DEBUG
    sprintf (syStrBuf, "> depth:%6d\r", depth); 
    Sy_Message (SY_FMT, syStrBuf);
#endif
  }

  Sy_Free(res_tab);
}    


static int Sy_ResCompare(SY_CONST_VOID *ra, SY_CONST_VOID *rb)
{
  if ( ((SyResTable*)ra)->residual > ((SyResTable*)rb)->residual) return (1);
  if ( ((SyResTable*)ra)->residual < ((SyResTable*)rb)->residual) return (-1);
  return(0);
}

static int Sy_ResBufCompare(SY_CONST_VOID *a, SY_CONST_VOID *b)
{
  if ( *((float*)a) > *((float*)b) ) return (1);
  if ( *((float*)a) < *((float*)b) ) return (-1);
  return(0);
}




