/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_SmoothPotiId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include <math.h>

#define SY_INIT_MAXCONTOURS 128

enum SyScanDirection {FORWARD, BACKWARD};

extern enum SyResidualCompMode residualCompMode;
extern enum SyResidualBorderMode residualBorderMode;

extern Boolean is_residual_fix, is_maxfix;

static SyHipoContainer *_container;
static double _sy_max_resid = 0.0;

static void Sy_CompGenpoti(SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
			   void (*TraceMethod) (/* ??? */), float *potival);

static void Sy_GenTrace(SyIpoDAC *ipd, SyRegpot *pot, int start_ipo,
			int start_icn, int nr, enum SyScanDirection dir,
			float *potival); 

static void Sy_DiamMap(SySkelDAC *vsk, SyIpoDAC *ipd, float *potival, 
		       float flatten_th);

void Sy_RegFlatten (SySkelDAC *vsk, SyIpoDAC *ipd, 
		    void (*regul_method) (/* ??? */), int is_verbose, 
		    float flatten_th)
{
  SyRegpot pot;
  float    *potival;

  vsk->is_border = 0;

  Sy_AllocRegpot (&pot, ipd->ipodim);

  potival = (float*) Sy_Malloc ((ipd->ipodim+1)*sizeof(float));

  Sy_DiamMap (vsk, ipd, potival, flatten_th);

  if (residualCompMode == RES_FULL){ 
    vsk->eresid    = Sy_AllocEresid (vsk->edim);
    vsk->erdim     = vsk->edim;
    vsk->is_eresid = 1;
  }
  else{ 
    vsk->eresid    = (SyEresid*) NULL;
    vsk->erdim     = 0;
    vsk->is_eresid = 0;
  }

#ifdef DEBUG  
  Sy_Message (SY_FMTNL, "Contour Tracer: ");
#endif

  Sy_CompGenpoti (ipd, &pot, is_verbose, Sy_GenTrace, potival);

  (*regul_method) (vsk, ipd, &pot, is_verbose);
}

static void Sy_CompGenpoti (SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
			    void (*TraceMethod) (/* ??? */), float *potival)
{ 
  void ipo_trace();
  int i, obj_count = 0;

  for (i = 1; i <= ipd->ipodim; i++){
    if (pot->elem_[i].label == 0){
      (*TraceMethod)(ipd, pot, i, ipd->ipl[i].tail_icn, ++obj_count, 
		     FORWARD, potival);
#ifdef DEBUG
      if (syIsVerbose) sprintf (syStrBuf, "contours: %d\r", obj_count);
      Sy_Message (SY_FMT, syStrBuf);
#endif
    }
  }
}

static void Sy_GenTrace (SyIpoDAC *ipd, SyRegpot *pot, int start_ipo,
			 int start_icn, int nr, enum SyScanDirection dir, 
			 float *potival)
{
  int    cur_ipo, cur_icn; 
  int    next_ipo, next_icn;
  float  cur_len, dlen;
  SyIpo  cur_pos, next_pos;
  
  cur_ipo = start_ipo;
  cur_icn = start_icn;
  cur_pos = ipd->ipo[cur_ipo];
  cur_len = 0.0; /* SY_ALV(potival[cur_ipo]); */

  do{
    pot->elem_[cur_ipo].val   = (dir == FORWARD) ? cur_len: -cur_len;
    pot->elem_[cur_ipo].label = nr;

    if (ipd->icn[cur_icn].tail_ipo == cur_ipo){ 
      next_ipo = ipd->icn[cur_icn].head_ipo;}
    else{next_ipo = ipd->icn[cur_icn].tail_ipo;}

    if (ipd->ipl[next_ipo].tail_icn == cur_icn){
      next_icn = ipd->ipl[next_ipo].head_icn;}
    else{next_icn = ipd->ipl[next_ipo].tail_icn;}

    next_pos = ipd->ipo[next_ipo];

    dlen = SY_SQUARE(cur_pos.x-next_pos.x) + SY_SQUARE(cur_pos.y-next_pos.y);
    if (potival[next_ipo] < 20.0) dlen = dlen/400.0;
    cur_len += sqrt (dlen);

    cur_ipo = next_ipo;
    cur_icn = next_icn;
    cur_pos = next_pos;

  } while ((pot->elem_[cur_ipo].label == 0) && (next_ipo != 0));

  pot->obj_count     = nr;

  if (dir == FORWARD){
    pot->obj_tab_[nr].len = cur_len;
    pot->obj_closed_[nr]  = SY_TRUE;
  }
  else{                
    pot->obj_tab_[nr].len = 0.0;
    pot->obj_closed_[nr]  = SY_FALSE;
  }

  if ((dir == FORWARD) && (next_ipo == 0)){
    if (start_icn == ipd->ipl[start_ipo].tail_icn){
      start_icn = ipd->ipl[start_ipo].head_icn;
    }
    else{
      start_icn = ipd->ipl[start_ipo].tail_icn;
    }
    Sy_GenTrace (ipd, pot, start_ipo, start_icn, nr, BACKWARD, potival);
  }
}

static void Sy_Insert(SySkelDAC *vsk, int vindex, float dev, float rad);

void Sy_DiamRegularize (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
			int is_verbose)
{
  int      i;
  SyPoint  lpoint;
  SyVertex tvertex, hvertex;
  double   trad, hrad;
  float    dist;
  SyVector tvec[2], hvec[2];
  int      left_point, right_point;
  int      tail_vertex, head_vertex;
  int      left_label, right_label;

  Sy_Message (SY_FMTNL, "Method: Diameter-Based  ");

  vsk->prune_method = SKE_POT;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	dist  = SY_ALV(pot->elem_[left_point].val - 
		   pot->elem_[right_point].val);
	dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	dist  = SY_ALV(dist);

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist;
	  vsk->eresid[i].head_residual = dist;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	  vsk->eresid[i].same_boundary = 1;
	}

	Sy_Insert (vsk, tail_vertex, dist, trad);
	Sy_Insert (vsk, head_vertex, dist, hrad);
      }
      break;
      case 1:
	left_point  = vsk->edge[i].left_point;
	right_point = vsk->edge[i].right_point;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	tvertex = vsk->vertex[tail_vertex];
      
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }

	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	  dist  = SY_ALV(pot->elem_[left_point].val - 
		      pot->elem_[right_point].val);
	  dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	  dist  = SY_ALV(dist);

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dist;
	    vsk->eresid[i].head_residual = SY_NEG_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  Sy_Insert (vsk, tail_vertex, dist, trad);
	  break;
	}
    }
  }
}

static void Sy_Insert (SySkelDAC *vsk, int vindex, float dev, float rad)
{
  float*  birth;

  vsk->regul[vindex].radius = rad;
  birth = vsk->regul[vindex].birth;

  if (birth[2] < dev){
    birth[0] = birth[1];
    birth[1] = birth[2];
    birth[2] = dev;
    return;
  }
  if (birth[1] < dev){
    birth[0] = birth[1];
    birth[1] = dev;
    return;
  }
  if (birth[0] < dev){
    birth[0] = dev;
  }
}

static void  Sy_DiamMap (SySkelDAC *vsk, SyIpoDAC *ipd, float *potival, 
			 float flatten_th)
{
  float    rad;
  int      conn_flag, conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_map, i;

  for (i = 1; i <= ipd->ipodim; i++) potival[i] = 0.0;
  
  for (i = 1; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;
    
    tail_flag   = vsk->regul[tail_vertex].birth[2] > flatten_th;
    head_flag   = vsk->regul[head_vertex].birth[2] > flatten_th;
    
    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));
    
    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));
    
    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));
    
    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));
    
    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);
    
    is_map = (tail_flag && head_flag && !conn_flag);

    if (is_map){
            rad = min(vsk->regul[tail_vertex].radius, vsk->regul[head_vertex].radius);
      if (potival[left_point] != 0.0){
	if (potival[left_point]  > rad) potival[left_point] = rad;}
      else potival[left_point] = rad;

      if (potival[right_point] != 0.0){
	if (potival[right_point]  > rad) potival[right_point] = rad;}
      else potival[right_point] = rad;
    }
  }
}


static void  Sy_SmoothIpo (SyIpoDAC *ipd, SyIpoFloat *fipo, float sigma_in,
			   SyRegpot *pot)
{
  void Sy_FillContainer(SyIpoDAC *ipd, SyHipoContainer *container);

  SyHipoContainer container;

  int        ipoptr_lo, ipoptr_hi, ptr_lo, ptr_hi, support, 
             k, i, ii, cptr, mul, pdim, sign;
  float      *mask = NULL, scale, x_val, y_val, exp_factor, 
             sigma, first_ellipse_b;
  SyIpo      *ipo, *buf;
  SyIpoFloat *xybuf;
  int        *origin;
  double     orig_len, 
             smooth_len, tmplen;

  SyBoolean  do_fft = SY_FALSE;

  _container = &container;
  sigma = sigma_in; 

#ifdef DEBUG
  sprintf (syStrBuf, "Smoothing with sigma = %f ", sigma);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  Sy_FillContainer (ipd, &container);

  if (ipd->info == NULL) {
    ipd->maxcontourdim = container.maxdim;
    ipd->info = (SyIpoInfo*) Sy_Malloc(sizeof(SyIpoInfo)*
				       (ipd->maxcontourdim+1));
    for (i = 0; i <= ipd->maxcontourdim; i++){
      ipd->info[i].orig_len            = 0.0;
      ipd->info[i].first_ellipse_b_ref = 0.0;
    }
  }

  ipd->contourdim = container.size;

  for (cptr = 0; cptr < container.size; cptr++){

    /* check whether global sigma might be too large for some 
       contours */

    sigma = sigma_in;
    if (ipd->info && (first_ellipse_b = 
	ipd->info[container.info[cptr].contour_id].first_ellipse_b_ref) > 0.0){
      if (first_ellipse_b * 2.0 * M_PI * 0.34 < 0.99*sigma){
	sigma = first_ellipse_b * 2.0 * M_PI * 0.34;
#ifdef DEBUG
	fprintf(stderr, "sigma %f too high, resetting to %f\n",
		sigma_in, sigma);
#endif
      }
    }

    if ((pot->control == NULL) || pot->control->gaussian_smoothing){
      if (sigma > 4.0){
	do_fft = SY_TRUE;
      }
      else { /* construct Gaussian kernel */
	support = max( ((int)(7.0 * sigma))/2, 1);
	exp_factor = 1.0/(2.0 * sigma*sigma);
	mask = (float*) Sy_Malloc (sizeof(float) * (support+1));
      
	scale = 1.0;
	for (i = 1; i <= support; i++){
	  scale += 2.0 * (mask[i] = exp(-exp_factor*(float)(i*i)) );
	}
	scale = 1.0/scale; mask[0] = scale;
	for (i = 1; i <= support; i++) mask[i] *= scale; /* norm */
      }
    } /* if */
    
    sprintf (syStrBuf, "Smoothing contour no %d...\r", cptr);
    Sy_Message (SY_FMT, syStrBuf);

    ipo    = container.ipo[cptr];
    origin = container.origin[cptr];

    if ((pot->control != NULL) && pot->control->oliensis_smoothing){
      mask = Sy_ComputeOlienMask (sigma, orig_len, &support);
    }
    
    xybuf = (SyIpoFloat*) 
      Sy_Malloc (sizeof(SyIpoFloat)*(container.iposize[cptr]+2));

    /* smoothing */    
    if (!do_fft){ /*conventional convolution -- faster for small sigmas */

      mul = max(3,(2*support+container.iposize[cptr])/container.iposize[cptr]);
      if (mul%2 == 0) mul++;

#ifdef DEBUG    
      sprintf(syStrBuf, "mul = %d ", mul);
      Sy_Message (SY_FMTNL, syStrBuf);
#endif

      if (mul <= 3){
	for (i = 1; i <= container.iposize[cptr]; i++){
	  x_val = mask[0] * ipo[i].x; y_val = mask[0] * ipo[i].y;
	  
	  for (k = 1, ipoptr_lo = i-1, ipoptr_hi = i+1; k <= support; k++){
	    if (ipoptr_lo < 1) 
	      ipoptr_lo = -((-ipoptr_lo) % container.iposize[cptr]) 
		+ container.iposize[cptr]; 
	    if (ipoptr_hi > container.iposize[cptr]) 
	      ipoptr_hi =(ipoptr_hi % container.iposize[cptr]);	
	    x_val += (ipo[ipoptr_lo  ].x + ipo[ipoptr_hi  ].x)* mask[k];      
	    y_val += (ipo[ipoptr_lo--].y + ipo[ipoptr_hi++].y)* mask[k];      
	  } 
	  fipo[origin[i]].x = xybuf[i].x = x_val; 
	  fipo[origin[i]].y = xybuf[i].y = y_val;       
	}
      }
      else{      
	buf = (SyIpo*) malloc (sizeof(SyIpo)*mul*(container.iposize[cptr]+1));
	
	for (i = 0; i < mul; i++){
	  memcpy(&buf[1+i*container.iposize[cptr]], &ipo[1], 
		 container.iposize[cptr]*sizeof(SyIpo));
	}
	
	for (i = 1, ii = container.iposize[cptr]*(mul/2)+1; 
	     i <= container.iposize[cptr]; i++,ii++){
	  x_val = mask[0] * buf[ii].x; y_val = mask[0] * buf[ii].y;    
	  for (k = 1, ipoptr_lo = ii-1, ipoptr_hi = ii+1; k <= support; k++){
	    x_val += (buf[ipoptr_lo  ].x + buf[ipoptr_hi  ].x)* mask[k];      
	    y_val += (buf[ipoptr_lo--].y + buf[ipoptr_hi++].y)* mask[k];      
	  } 
	  fipo[origin[i]].x = xybuf[i].x = x_val; 
	  fipo[origin[i]].y = xybuf[i].y = y_val;       
	}    
	free(buf);
      } /* else */
    } /* if (!do_fft) */ 
    else { /* convolution with Gaussian kernel via FFT */
      
      Sy_GaussSmoothFFT (ipo, xybuf, container.iposize[cptr], sigma);

      for (i = 1; i <= container.iposize[cptr]; i++){
	fipo[origin[i]].x = xybuf[i].x; 
	fipo[origin[i]].y = xybuf[i].y;       
      }     
      
    } /* if (!do_fft) */

    if (pot->control != NULL && pot->control->is_concave_point != NULL){
      /* compute curvature info */

      pdim = container.iposize[cptr];      
      for (i = 1; i <= pdim; i++){
	
	ptr_lo = i-2; ptr_hi = i+2;
	
	if (ptr_lo < 1) ptr_lo = -((-ptr_lo) % pdim) + pdim; 
	if (ptr_hi > pdim) ptr_hi =(ptr_hi % pdim);	
      
	sign = Sy_CurvatureSign (xybuf, ptr_lo, i-1, i, i+1, ptr_hi);
	
	pot->control->is_concave_point[i] = (sign < 0);
      }
    } /* if */

    /*  computing attributes  */  

    orig_len = 0.0; smooth_len = 0.0; origin[0] = 0; 
    ipo[0] = ipo[1]; fipo[0] = fipo[origin[1]];

    for (i = 1; i <= container.iposize[cptr]; i++){
      tmplen = SY_SQUARE(ipo[i].x - ipo[i-1].x)+
	SY_SQUARE(ipo[i].y - ipo[i-1].y);
      orig_len += sqrt(tmplen);

      tmplen = SY_SQUARE(fipo[origin[i]].x - fipo[origin[i-1]].x)+
	SY_SQUARE(fipo[origin[i]].y - fipo[origin[i-1]].y);
      smooth_len += sqrt(tmplen);

    }
    tmplen = SY_SQUARE(ipo[container.iposize[cptr]].x - ipo[1].x)+
      SY_SQUARE(ipo[container.iposize[cptr]].y - ipo[1].y);
    orig_len += sqrt(tmplen);

    tmplen = SY_SQUARE(fipo[origin[container.iposize[cptr]]].x - 
		       fipo[origin[1]].x)+
      SY_SQUARE(fipo[origin[container.iposize[cptr]]].y - fipo[origin[1]].y);
    smooth_len += sqrt(tmplen);

    ipd->info[container.info[cptr].contour_id].orig_len   = orig_len;
    ipd->info[container.info[cptr].contour_id].smooth_len = smooth_len;

    Sy_Free(mask);
    Sy_Free(xybuf);
  } 
  /* for all contours */

  Sy_FreeContainer (&container);

#ifdef DEBUG
  Sy_Message (SY_FMTNL, "Done ");
#endif
}

static void Sy_CompSmoothPoti(SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
			      void (*TraceMethod) (/* ??? */), 
			      SyIpoFloat *fipo);

static void Sy_SmoothTrace(SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
			   int start_icn, int nr, enum SyScanDirection dir, 
			   SyIpoFloat *fipo); 
 
static SyObjTab *tab = NULL;

void Sy_ClearObjTabPtr (){ Sy_Free(tab); tab = NULL; }

void Sy_RegSmooth (SySkelDAC *vsk, SyIpoDAC *ipd, 
		   void (*regul_method) (/* ??? */), int is_verbose, 
		   float sigma, int transform, SyRegulControl *cntrl)
{
  SyRegpot   pot;
  SyIpoFloat *fipo = NULL;
  int i;

  Sy_Message (SY_FMTNL, "Pruning... ");

  pot.elem_       = NULL;
  pot.obj_tab_    = NULL;
  pot.obj_closed_ = NULL;

  vsk->is_border = 0;

  Sy_AllocRegpot (&pot, ipd->ipodim);
  vsk->regul = Sy_AllocRegul (vsk->vdim);
  pot.control = cntrl;

  fipo = (SyIpoFloat*) Sy_Malloc ((ipd->ipodim+1)*sizeof(SyIpoFloat));

  Sy_SmoothIpo (ipd, fipo, sigma, &pot);

  if (residualCompMode == RES_FULL){ 
    Sy_Free(vsk->eresid); 
    vsk->eresid    = Sy_AllocEresid (vsk->edim);
    vsk->erdim     = vsk->edim;
    vsk->is_eresid = SY_TRUE;
  }
  else{ 
    vsk->eresid    = (SyEresid*) NULL;
    vsk->erdim     = 0;
    vsk->is_eresid = 0;
  }
  
  Sy_CompSmoothPoti (ipd, &pot, is_verbose, Sy_SmoothTrace, fipo);

  if (pot.control == NULL || !pot.control->no_shrink_estim){
    Sy_Free(tab);
    tab = (SyObjTab*) Sy_Malloc(sizeof(SyObjTab) * pot.obj_dim);
    memcpy (tab, pot.obj_tab_, sizeof(SyObjTab) * pot.obj_dim);
  }
  else if (tab != NULL){
    for (i = 0; i <= pot.obj_count; i++)
      pot.obj_tab_[i].bndry_scale = tab[i].bndry_scale;
  }

  Sy_Message (SY_FMT, "residual = ");
  (*regul_method) (vsk, ipd, &pot, is_verbose, fipo); 

  if (is_residual_fix) Sy_Infix  (vsk, ipd);
  if (is_maxfix)       Sy_Maxfix (vsk, ipd);

  if (ipd->rpe != NULL) Sy_Free(ipd->rpe);
  ipd->rpe = pot.elem_; pot.elem_ = NULL;

  if (transform){
    for (i = 1; i <= ipd->ipodim; i++){
      ipd->ipo[i].x = vsk->point[i].x = (int) fipo[i].x;
      ipd->ipo[i].y = vsk->point[i].y = (int) fipo[i].y;
    }
  }

  Sy_Free (fipo);
  Sy_FreeRegpot(&pot);

  if (pot.control != NULL && pot.control->is_single_shot){
    Sy_Free (tab); tab = NULL;
  }
}

static void Sy_CompSmoothPoti (SyIpoDAC *ipd, SyRegpot *pot, int is_verbose, 
			       void (*TraceMethod) (/* ??? */), 
			       SyIpoFloat *fipo)
{ 
  int i, obj_count = 0;

  for (i = 1; i <= ipd->ipodim; i++){
    if (pot->elem_[i].label == 0){
      (*TraceMethod)(ipd, pot, i, ipd->ipl[i].tail_icn, ++obj_count, 
		     FORWARD, fipo);
    }
  }
  if (syIsVerbose) 
    sprintf (syStrBuf, "\ntracked %d contour[s], ", obj_count);
  Sy_Message (SY_FMT, syStrBuf);
}

static void Sy_SmoothTrace (SyIpoDAC *ipd, SyRegpot *pot, int start_ipo, 
			    int start_icn, int nr, enum SyScanDirection dir, 
			    SyIpoFloat *fipo)
{
  int          cur_ipo, cur_icn; 
  int          next_ipo, next_icn;
  float        cur_len, eff_cur_len, dlen, eff_dlen,
               max_x, min_x, max_y, min_y,
               eff_max_x, eff_min_x, eff_max_y, eff_min_y,
               bndry_scale;
  SyIpoFloat   cur_pos, next_pos;
  SyIpo        eff_cur_pos, eff_next_pos;
  double       tot_len, tot_len_smooth;

  double       af, bf, cf, df, 
               af_smooth, bf_smooth, cf_smooth, df_smooth,
               newarg, oldarg, oldarg_smooth, 
               deltax, deltay, a, b, a_smooth, b_smooth, 
               factor, diffcos, diffsin, 
               teta, sisin, cocos, aa, bb, cc, dd;
  
  cur_ipo = start_ipo;
  cur_icn = start_icn;
  cur_pos = fipo[cur_ipo]; eff_cur_pos = ipd->ipo[cur_ipo];
  cur_len = 0.0; eff_cur_len = 0.0;
  tot_len  = ipd->info[nr].orig_len;
  tot_len_smooth = ipd->info[nr].smooth_len;

  af = bf = cf = df = af_smooth = bf_smooth = cf_smooth = df_smooth =
    oldarg = oldarg_smooth = 0.0;
  min_x = max_x = cur_pos.x; 
  min_y = max_y = cur_pos.y;
  eff_min_x = eff_max_x = eff_cur_pos.x; 
  eff_min_y = eff_max_y = eff_cur_pos.y;
  
  do{
    pot->elem_[cur_ipo].val   = (dir == FORWARD) ? cur_len: -cur_len;
    pot->elem_[cur_ipo].label = nr;

    if (ipd->icn[cur_icn].tail_ipo == cur_ipo){ 
      next_ipo = ipd->icn[cur_icn].head_ipo;}
    else{next_ipo = ipd->icn[cur_icn].tail_ipo;}

    if (ipd->ipl[next_ipo].tail_icn == cur_icn){
      next_icn = ipd->ipl[next_ipo].head_icn;}
    else{next_icn = ipd->ipl[next_ipo].tail_icn;}

    next_pos = fipo[next_ipo]; eff_next_pos = ipd->ipo[next_ipo];

    min_x = min(min_x, next_pos.x); max_x = max(max_x, next_pos.x);
    min_y = min(min_y, next_pos.y); max_y = max(max_y, next_pos.y);
    eff_min_x = min(eff_min_x, eff_next_pos.x); 
    eff_max_x = max(eff_max_x, eff_next_pos.x);
    eff_min_y = min(eff_min_y, eff_next_pos.y); 
    eff_max_y = max(eff_max_y, eff_next_pos.y);

    dlen = SY_SQUARE(cur_pos.x-next_pos.x) + SY_SQUARE(cur_pos.y-next_pos.y);
    cur_len += sqrt (dlen);
    eff_dlen = SY_SQUARE(eff_cur_pos.x-eff_next_pos.x) 
             + SY_SQUARE(eff_cur_pos.y-eff_next_pos.y);
    eff_cur_len += sqrt (eff_dlen);

    /* first harmonic */

    deltax  = eff_cur_pos.x - eff_next_pos.x;
    deltay  = eff_cur_pos.y - eff_next_pos.y;
    newarg  = 2 * M_PI * eff_cur_len/tot_len;
    diffcos = cos (newarg) - cos (oldarg);
    diffsin = sin (newarg) - sin (oldarg);
    af   += deltax * diffcos;
    bf   += deltax * diffsin;
    cf   += deltay * diffcos;
    df   += deltay * diffsin;
    oldarg  = newarg;

    deltax  = cur_pos.x - next_pos.x;
    deltay  = cur_pos.y - next_pos.y;
    newarg  = 2 * M_PI * cur_len/tot_len_smooth;
    diffcos = cos (newarg) - cos (oldarg_smooth);
    diffsin = sin (newarg) - sin (oldarg_smooth);
    af_smooth   += deltax * diffcos;
    bf_smooth   += deltax * diffsin;
    cf_smooth   += deltay * diffcos;
    df_smooth   += deltay * diffsin;
    oldarg_smooth  = newarg;

    cur_ipo = next_ipo;
    cur_icn = next_icn;
    cur_pos = next_pos; eff_cur_pos = eff_next_pos;

  } while ((pot->elem_[cur_ipo].label == 0) && (next_ipo != 0));

  pot->obj_count     = nr;

  if ((pot->control == NULL) || pot->control->gaussian_smoothing){

    if (max_x == min_x){ 
      if (max_y == min_y) bndry_scale = 1.0;
      else bndry_scale = (eff_max_y - eff_min_y)/(max_y - min_y);}
    else{
      if (max_y == min_y)
	bndry_scale = (eff_max_x - eff_min_x)/(max_x - min_x);
      else
	bndry_scale = max ((eff_max_x - eff_min_x)/(max_x - min_x),
			   (eff_max_y - eff_min_y)/(max_y - min_y));
    }
  }
  else{
    bndry_scale = 1.0;
  }

  ipd->max_x = max_x; ipd->min_x = min_x;
  ipd->max_y = max_y; ipd->min_y = min_y;

  if (dir == FORWARD){
    pot->obj_tab_[nr].len     = cur_len;
    pot->obj_tab_[nr].eff_len = eff_cur_len;

    if (pot->control == NULL || !pot->control->no_shrink_estim){
      pot->obj_tab_[nr].bndry_scale = bndry_scale;
    }

    pot->obj_closed_[nr]      = SY_TRUE;

    factor = tot_len / (2.0 * M_PI * M_PI);
    af *= factor; bf *= factor; cf *= factor; df *= factor;

    teta  = 0.5 * atan (2 * (af*bf + cf*df)/(sq(af)+sq(cf)-sq(bf)-sq(df) ));
  
    if ( (sq (af) + sq (cf)) < (sq (bf) + sq(df)) )
      teta -= M_PI / 2.0;

    cocos  = cos (teta);
    sisin  = sin (teta);
    aa     = af; bb = bf; cc = cf; dd = df;
    af   =  aa * cocos + bb * sisin;
    bf   = -aa * sisin + bb * cocos;
    cf   =  cc * cocos + dd * sisin;
    df   = -cc * sisin + dd * cocos;

    a = sqrt(af*af+cf*cf); b = sqrt(bf*bf+df*df);

    factor = tot_len_smooth / (2.0 * M_PI * M_PI);
    af_smooth *= factor; bf_smooth *= factor; 
    cf_smooth *= factor; df_smooth *= factor;

    teta  = 0.5 * atan (2 * (af_smooth*bf_smooth + cf_smooth*df_smooth)/
			(sq(af_smooth)+sq(cf_smooth)
			 -sq(bf_smooth)-sq(df_smooth) ));
  
    if ( (sq(af_smooth) + sq(cf_smooth)) < (sq(bf_smooth) + sq(df_smooth)) )
      teta -= M_PI / 2.0;

    cocos  = cos (teta);
    sisin  = sin (teta);
    aa     = af_smooth; bb = bf_smooth; cc = cf_smooth; dd = df_smooth;
    af_smooth   =  aa * cocos + bb * sisin;
    bf_smooth   = -aa * sisin + bb * cocos;
    cf_smooth   =  cc * cocos + dd * sisin;
    df_smooth   = -cc * sisin + dd * cocos;

    a_smooth = sqrt(af_smooth*af_smooth+cf_smooth*cf_smooth); 
    b_smooth = sqrt(bf_smooth*bf_smooth+df_smooth*df_smooth);

    ipd->info[nr].first_ellipse_len = M_PI*(1.5*(a+b)-sqrt(a*b));
    ipd->info[nr].first_ellipse_len_smooth = 
      M_PI*(1.5*(a_smooth+b_smooth)-sqrt(a_smooth*b_smooth));
    ipd->info[nr].first_ellipse_a        = max(a,b);
    ipd->info[nr].first_ellipse_b        = min(a,b);
    ipd->info[nr].first_ellipse_a_smooth = max(a_smooth,b_smooth);
    ipd->info[nr].first_ellipse_b_smooth = min(a_smooth,b_smooth);

    if (ipd->info[nr].first_ellipse_b_ref <= 0.0){
      ipd->info[nr].first_ellipse_b_ref = 
	ipd->info[nr].first_ellipse_b_smooth;
    }

#ifdef DEBUG
    sprintf (syStrBuf, "cont. len.: %7.3f eff., %7.3f smoothed, \
ratio: %5.2f, BB-ratio: %5.2f, a:%f, b:%f\n\
smoothed: a:%f, b:%f\n", 
             eff_cur_len, cur_len, eff_cur_len/cur_len, 
	     pot->obj_tab_[nr].bndry_scale, a, b, a_smooth, b_smooth);
    Sy_Message (SY_FMTNL, syStrBuf);
    sprintf(syStrBuf, "first ellipse len: %f, smoothed: %f ",
	    ipd->info[nr].first_ellipse_len,
	    ipd->info[nr].first_ellipse_len_smooth);
    Sy_Message (SY_FMTNL, syStrBuf);
#endif
  }
  else{                
    pot->obj_tab_[nr].len = 0.0;
    pot->obj_tab_[nr].bndry_scale = 1.0;
    pot->obj_closed_[nr]  = SY_FALSE;
  }

  if ((dir == FORWARD) && (next_ipo == 0)){
    if (start_icn == ipd->ipl[start_ipo].tail_icn){
      start_icn = ipd->ipl[start_ipo].head_icn;
    }
    else{
      start_icn = ipd->ipl[start_ipo].tail_icn;
    }
    Sy_SmoothTrace (ipd, pot, start_ipo, start_icn, nr, BACKWARD, fipo);
  }
}

static SyWalktask wt;

static void Sy_AddIpo2ContainerTerm(void), 
            Sy_AddIpoStrip2Container(void),
            Sy_Ipo2Container(void);

void Sy_FillContainer (SyIpoDAC *ipd, SyHipoContainer *container)
{
  int i = 0;

  container->maxdim   = SY_INIT_MAXCONTOURS;
  container->size     = 0;
  container->ptr      = -1;
  container->total_iposize = 0;
  container->iposize  = (int*) Sy_Malloc (sizeof(int)*container->maxdim);
  container->ipomaxdim= (int*) Sy_Malloc (sizeof(int)*container->maxdim);
  container->ipo      = (SyIpo**) Sy_Malloc (sizeof(SyIpo*)*container->maxdim);
  container->origin   = (int**) Sy_Malloc (sizeof(int*)*container->maxdim);
  container->info     = (SyIpoInfo*) Sy_Malloc(sizeof(SyIpoInfo)*container->maxdim);
  container->hipodata = (SyHipoData**) NULL;

  for (i = 0; i < container->maxdim; i++){
    container->iposize[i]   = 0;
    container->ipomaxdim[i] = 0; 
    container->ipo[i]       = (SyIpo*) NULL;
    container->origin[i]    = (int*) NULL;
  }

  wt.user_function[WALK] = (void(*)())Sy_Ipo2Container;
  wt.user_function[TRAV] = (void(*)())Sy_AddIpoStrip2Container;
  wt.user_function[TERM] = (void(*)())Sy_AddIpo2ContainerTerm;

  Sy_InitIpoTraverse (ipd, &wt);
  Sy_IpoTraverse (ipd, &wt);  
}

void Sy_FreeContainer (SyHipoContainer *container)
{
  int i = 0;

  Sy_Free(container->iposize);   container->iposize   = NULL;
  Sy_Free(container->ipomaxdim); container->ipomaxdim = NULL;
  Sy_Free(container->info);      container->info      = NULL;

  for (i = 0; i < container->maxdim; i++){
    Sy_Free(container->ipo[i]);
    Sy_Free(container->origin[i]);
  }

  Sy_Free(container->ipo);
  Sy_Free(container->origin);

  container->maxdim        = 0;
  container->size          = 0;
  container->ptr           = -1;
  container->total_iposize = 0;
}

static void Sy_Ipo2Container(void)
{
  int this_ipo;
  
  this_ipo = wi.current.head_ipo;

  _container->iposize[_container->ptr]++;

  (*(_container->ipo+_container->ptr))[_container->iposize[_container->ptr]]
 = wi.ipd->ipo[this_ipo];

  (*(_container->origin+_container->ptr))[_container->iposize[_container->ptr]]
 = this_ipo;

  if (_container->iposize[_container->ptr] >=
      _container->ipomaxdim[_container->ptr]){
    _container->ipo[_container->ptr] = (SyIpo*) 
      Sy_Realloc ((char*) _container->ipo[_container->ptr], 
		  sizeof(SyIpo)*(_container->ipomaxdim[_container->ptr] +
				 SY_IPO_STRIP_INITSIZE));
    _container->origin[_container->ptr] =
      (int*) Sy_Realloc ((char*) _container->origin[_container->ptr],
		      sizeof(int)*(_container->ipomaxdim[_container->ptr] +
				   SY_IPO_STRIP_INITSIZE));
    _container->ipomaxdim[_container->ptr] += SY_IPO_STRIP_INITSIZE;
  }
}

static void Sy_AddIpoStrip2Container(void)
{
  int i;

  _container->ptr++;
  _container->size++;

#ifdef DEBUG
  sprintf (syStrBuf, "Adding contour nr %d ", _container->size);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  if (_container->size >= _container->maxdim){ /* reallocate memory */

    _container->maxdim += SY_INIT_MAXCONTOURS;
    _container->iposize   = (int*) Sy_Realloc(_container->iposize, 
					      sizeof(int)*_container->maxdim);
    _container->ipomaxdim = (int*) Sy_Realloc(_container->ipomaxdim, 
					      sizeof(int)*_container->maxdim);
    _container->ipo       = (SyIpo**) Sy_Realloc(
      _container->ipo, sizeof(SyIpo*)*_container->maxdim);
    _container->origin    = (int**) Sy_Realloc(
      _container->origin, sizeof(int*)*_container->maxdim);
    _container->info      = (SyIpoInfo*) Sy_Realloc(
      _container->info, sizeof(SyIpoInfo)*_container->maxdim);
    
    for (i = _container->size; i < _container->maxdim; i++){
      _container->iposize[i]   = 0;
      _container->ipomaxdim[i] = 0; 
      _container->ipo[i]       = (SyIpo*) NULL;
      _container->origin[i]    = (int*) NULL;
    }    

  }/* if */

  _container->info[_container->ptr].contour_id = wi.contour_id;

  _container->ipo[_container->ptr] = 
    (SyIpo*) Sy_Malloc (sizeof(SyIpo)*(SY_IPO_STRIP_INITSIZE+1)); 
  _container->origin[_container->ptr] =
    (int*) Sy_Malloc (sizeof(int)*(SY_IPO_STRIP_INITSIZE+1));
  _container->ipomaxdim[_container->ptr] = SY_IPO_STRIP_INITSIZE;
}

static void Sy_AddIpo2ContainerTerm(void)
{

#ifdef DEBUG
  sprintf (syStrBuf, "Strip %d with %d points ", 
	   _container->ptr, _container->iposize[_container->ptr]);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  _container->total_iposize += _container->iposize[_container->ptr];
}

void Sy_SmoothRegularize (SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
			  int is_verbose, SyIpoFloat *fipo)
{
  int      i, j, edim;
  SyPoint  lpoint, rpoint;
  SyVertex tvertex, hvertex;
  double   trad, hrad, sec;
  float    dist, dev, tdev, hdev;
  SyVector tvec[2], hvec[2];
  int      left_point, right_point;
  int      tail_vertex, head_vertex;
  int      left_label, right_label;
  int      no_edge_selection = SY_TRUE;

  double   targ, harg, max_resid;
  double   tper, hper, tshrink = 1.0, hshrink = 1.0; 
  int      is_tcorr, is_hcorr, is_mindev, corrcount=0; 

  if (syIsVerbose){
    switch(vsk->prune_method){
    case SKE_SMPOT: Sy_Message (SY_FMTNL, "Smoothed Potentials"); 
      break;
    case SKE_SMCHO: Sy_Message (SY_FMTNL, "Smoothed `Chord'"); 
      break;
    case SKE_SMCIR: Sy_Message (SY_FMTNL, "Smoothed `Circularity R.'");
      break;
    }
  }

  max_resid = SY_NEG_INFTY;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  no_edge_selection = (pot->control == NULL) ||
    (pot->control->is_edge_selection == SY_FALSE);

  if (no_edge_selection) edim = vsk->edim;
  else edim = pot->control->edge_select_dim;
  for (j = 1; j <= edim; j++){
    if (no_edge_selection) i = j;
    else i = pot->control->edge_select[j];
      
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_label  = pot->elem_[left_point].label;
      right_label = pot->elem_[right_point].label;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];

      if (no_edge_selection == 0 && pot->control->is_concave_point != NULL){
	if (pot->control->is_concave_point[left_point] &&
	    pot->control->is_concave_point[right_point]){
	  pot->control->is_concave_vertex[tail_vertex] =
	  pot->control->is_concave_vertex[head_vertex] = SY_TRUE;
	} /* if */
      } /* if */

#define TRUERADIUS
#ifdef TRUERADIUS
      if (vsk->prune_method == SKE_SMCIR){
	tvec[0].dirx = fipo[left_point].x - tvertex.xden/tvertex.num;
	tvec[0].diry = fipo[left_point].y - tvertex.yden/tvertex.num;
	tvec[1].dirx = fipo[right_point].x - tvertex.xden/tvertex.num;
	tvec[1].diry = fipo[right_point].y - tvertex.yden/tvertex.num;
	hvec[0].dirx = fipo[left_point].x - hvertex.xden/hvertex.num;
	hvec[0].diry = fipo[left_point].y - hvertex.yden/hvertex.num;
	hvec[1].dirx = fipo[right_point].x - hvertex.xden/hvertex.num;
	hvec[1].diry = fipo[right_point].y - hvertex.yden/hvertex.num;
      }
      else{
#endif TRUERADIUS
	tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
	tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
	tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
	hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
	hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
	hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
	hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;
#ifdef TRUERADIUS
      }
#endif       
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      if (left_label != right_label){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  vsk->eresid[i].same_boundary = 0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	

	dist  = SY_ALV(pot->elem_[left_point].val - 
		   pot->elem_[right_point].val);
	dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	dist  = SY_ALV(dist);

	switch(vsk->prune_method){
/* --------------------------------------------------------------------*/
  	case SKE_SMCIR: 
	  Sy_CorrResidEdge (vsk, ipd, pot, tvertex, hvertex, 
			    tvec[0], hvec[0], 
			    left_point, right_point, left_point,
			    &is_tcorr, &is_hcorr, &dist, &is_mindev);

	  if (is_mindev){
	    Sy_CorrResidEdge (vsk, ipd, pot, tvertex, hvertex, 
			      tvec[1], hvec[1], 
			      left_point, right_point, right_point,
			      &is_tcorr, &is_hcorr, &dist, &is_mindev);
	  }


	  if (is_tcorr) corrcount++;
	  if (is_hcorr) corrcount++;

/*	  is_tcorr = is_hcorr = SY_FALSE; */

	  targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	    (trad*trad);
	  
	  harg = (hvec[0].dirx*hvec[1].dirx + hvec[0].diry*hvec[1].diry)/
	    (hrad*hrad);
	  
	  if (targ >  1.0) targ =  1.0;
	  if (targ < -1.0) targ = -1.0;
	  if (harg >  1.0) harg =  1.0;
	  if (harg < -1.0) harg = -1.0;

	  if (is_mindev){

	    tdev = (dist - trad*tshrink*acos(targ));
	    hdev = (dist - hrad*hshrink*acos(harg));
	    
	    dev  = min(tdev, hdev);
	    
	    Sy_Insert (vsk, tail_vertex, tdev = dev, trad);
	    Sy_Insert (vsk, head_vertex, hdev = dev, hrad);
	  }
	  else{
	    tper = trad*tshrink*((is_tcorr) ? (2*SY_PI - acos(targ)):acos(targ));
	    hper = hrad*hshrink*((is_hcorr) ? (2*SY_PI - acos(harg)):acos(harg));    
	    Sy_Insert (vsk, tail_vertex, tdev = (dist-tper), trad);
	    Sy_Insert (vsk, head_vertex, hdev = (dist-hper), hrad);  
	  }
	  
	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = tdev;
	    vsk->eresid[i].head_residual = hdev;
	    vsk->eresid[i].tail_corr     = is_tcorr;
	    vsk->eresid[i].head_corr     = is_hcorr;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  break;
/* -------------------------------------------------------------------*/
	case SKE_SMCHO:
	case SKE_SMPOT:

	  sec   = sqrt (SY_SQUARE(fipo[left_point].x - fipo[right_point].x)+
			SY_SQUARE(fipo[left_point].y - fipo[right_point].y));
	  
	  dev = (vsk->prune_method == SKE_SMCHO) ? dist-sec : dist; 
	  dev *= pot->obj_tab_[left_label].bndry_scale; 
	  
	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = dev;
	    vsk->eresid[i].head_residual = dev;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = dist;
	    vsk->eresid[i].same_boundary = 1;
	  }
	  
	  Sy_Insert (vsk, tail_vertex, dev, trad);
	  Sy_Insert (vsk, head_vertex, dev, hrad);

	  if (max_resid < dev) max_resid = dev;

	  break;
	}
      break;
      case 1:
	left_point  = vsk->edge[i].left_point;
	right_point = vsk->edge[i].right_point;
        head_vertex = vsk->edge[i].head_vertex;
	tail_vertex = vsk->edge[i].tail_vertex;
	left_label  = pot->elem_[left_point].label;
	right_label = pot->elem_[right_point].label;
	
	lpoint  = vsk->point[left_point];
	rpoint  = vsk->point[right_point];
	tvertex = vsk->vertex[tail_vertex];
        hvertex = vsk->vertex[head_vertex];

	if (vsk->prune_method == SKE_SMCIR){
	  tvec[0].dirx = fipo[left_point].x - tvertex.xden/tvertex.num;
	  tvec[0].diry = fipo[left_point].y - tvertex.yden/tvertex.num;
	  tvec[1].dirx = fipo[right_point].x - tvertex.xden/tvertex.num;
	  tvec[1].diry = fipo[right_point].y - tvertex.yden/tvertex.num;
	}
	else{
	  tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
	  tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
	  tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
	  tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
	}

	trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

	if (left_label != right_label){

	  if (residualCompMode == RES_FULL){
	    vsk->eresid[i].tail_residual = SY_POS_INFTY;
	    vsk->eresid[i].head_residual = SY_POS_INFTY;
	    vsk->eresid[i].tail_corr     = 0;
	    vsk->eresid[i].head_corr     = 0;
	    vsk->eresid[i].dist          = 0.0;
	    vsk->eresid[i].same_boundary = 0;
	  }

	  Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	}
	else{	
	  dist  = SY_ALV(pot->elem_[left_point].val - 
		      pot->elem_[right_point].val);
	  dist  = min(dist, pot->obj_tab_[left_label].len - dist);
	  dist  = SY_ALV(dist);
	  switch(vsk->prune_method){	  
/* -----------------------------------------------------------------------*/
	  case SKE_SMCIR: 

	    Sy_CorrResidRay (vsk, ipd, pot, tvertex, hvertex, tvec[0], 
			     left_point, right_point, left_point, 
			     &is_tcorr, &dist, &is_mindev);
	  

	    if (is_tcorr) corrcount++;
	    is_tcorr = SY_FALSE;
	    
	    targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	      (trad*trad);

	    if (targ >  1.0) targ =  1.0;
	    if (targ < -1.0) targ = -1.0;

	    tper = trad*tshrink*((is_tcorr) ? (2*SY_PI - acos(targ)) : 
		                              acos(targ));
	    dev  = dist-tper;
	    
	    if (residualCompMode == RES_FULL){
	      vsk->eresid[i].tail_residual = dev;
	      vsk->eresid[i].head_residual = SY_NEG_INFTY;
	      vsk->eresid[i].tail_corr     = is_tcorr;
	      vsk->eresid[i].head_corr     = 0;
	      vsk->eresid[i].dist          = dist;
	      vsk->eresid[i].same_boundary = 1;
	    }

	    Sy_Insert (vsk, tail_vertex, dev, trad);
	    break;

/* ---------------------------------------------------------------------*/
	case SKE_SMCHO:
	case SKE_SMPOT:

	    sec   = sqrt (SY_SQUARE(fipo[left_point].x - fipo[right_point].x)+
			  SY_SQUARE(fipo[left_point].y - fipo[right_point].y));
	
	    dev = (vsk->prune_method == SKE_SMCHO) ? dist-sec : dist; 
	    dev *= pot->obj_tab_[left_label].bndry_scale; 
	    
	    if (residualCompMode == RES_FULL){
	      vsk->eresid[i].tail_residual = dev;
	      vsk->eresid[i].head_residual = SY_NEG_INFTY;
	      vsk->eresid[i].tail_corr     = 0;
	      vsk->eresid[i].head_corr     = 0;
	      vsk->eresid[i].dist          = dist;
	      vsk->eresid[i].same_boundary = 1;
	    }
	    Sy_Insert (vsk, tail_vertex, dev, trad);

	    if (max_resid < dev) max_resid = dev;

	    break;
	  }
	}
      }
    }
  } /* for all edges */

#ifdef DEBUG
  if (vsk->prune_method == SKE_SMCIR && syIsVerbose){
    sprintf (syStrBuf, "%d corrections ", corrcount);
    Sy_Message (SY_FMTNL, syStrBuf);
  }
#endif

  if (pot->control != NULL){
    pot->control->max_resid = max_resid;
  }
  _sy_max_resid  = max_resid;
  vsk->max_resid = max_resid;
}

double Sy_GetMaxResidual(void) { return _sy_max_resid; }

static int Sy_SortBirth3 (SY_CONST_VOID *ra, SY_CONST_VOID *rb)
{
  if ( ((SyRegul*)ra)->birth[0] > ((SyRegul*)rb)->birth[0] ) return (1);
  else
  if ( ((SyRegul*)ra)->birth[0] < ((SyRegul*)rb)->birth[0] ) return (-1);
  else
    return (0);
}
  
void Sy_OutputBirthSetHistogramm (SySkelDAC *vsk, char *filename, float threshold, 
				  float val3d, int is3d)
{
  SyRegul* buf;
  FILE*  fpout;
  int    mem, i, cnt;

  buf = (SyRegul*) Sy_Malloc (mem = (sizeof(SyRegul)*(vsk->vdim+1)) );
  memcpy ((char*) buf, (char*) vsk->regul, mem);
  
  qsort ((char*) (buf+1), vsk->vdim, sizeof(SyRegul), Sy_SortBirth3); 

  fpout = fopen (filename, "w");
  for (i=1, cnt = 0; i <= vsk->vdim; i++){
    if ((buf[i].birth[0] < SY_POS_INFTY) &&
	(buf[i].birth[0] > SY_NEG_INFTY) &&
	(buf[i].birth[0] > threshold) ){
      if (is3d)
	fprintf (fpout, "%8.3f %8.3f %8.3f\n", buf[i].birth[0], val3d, 
		 buf[i].birth[0]);
      else
	fprintf (fpout, "%8.3f %8.3f\n", buf[i].birth[0], buf[i].birth[0]);
    }
  }
  fclose(fpout); free(buf);
}

float Sy_OlienFreq (float ku, float kl, float c, float x)
{
  float buf;
  buf = (fabs(x) >= ku)? 0.0:(fabs(x) <= kl)? 1.0 : 
    1.0/(1.0+exp(c/sq((fabs(x)-ku)) - c/sq(fabs(x)-kl)));
  return buf;
}

float* Sy_ComputeOlienMask (int lowfreq, float bndry_len, int *size)
{
  float *data, *mask;
  float kl, ku, c = 81.0, norm = 0.0;
  int nn, pot, mask_size, i;
  
  kl = lowfreq; ku = kl+18;
  pot = ceil(log(bndry_len)/log(2.0));
  for (i = 1, nn=1; i <= pot; i++, nn*=2);
  
  data = (float*) Sy_Malloc (sizeof(float)*(2*nn));
  
  for (i = 0; i < 2*nn; i++) data[i] = 0.0;
  
  for (i = 0; i < ku; i++) {
    data[2*i] = data[2*(nn-1-i)] = Sy_OlienFreq(ku,kl,c,i);
  }

  four1(data-1, nn, -1); 

  *size = mask_size = nn/10;
  mask = (float*) Sy_Malloc (sizeof(float)*mask_size);
  norm = mask[0] = data[0];
  for (i = 1; i < mask_size; i++){
    norm += 2.0*(mask[i] = data[2*i]);
  }
  norm = 1.0/norm;
  for (i = 0; i < mask_size; i++){ mask[i] *= norm; }

  free(data);
  return mask;
}

double Sy_Curvature (SyIpoFloat *p, int p2, int p1, int i0, int n1, int n2)
{ 
  double d, d_plus, d_minus, db, da, curv, x_s, y_s, 
         x_s_minus, y_s_minus, x_s_plus, y_s_plus, x_ss, y_ss;
  
  d  =  SY_NORM(p[n1].x - p[i0].x, p[n1].y - p[i0].y)
      + SY_NORM(p[i0].x - p[p1].x, p[i0].y - p[p1].y);
    
  x_s = (p[n1].x - p[p1].x)/d;
  y_s = (p[n1].y - p[p1].y)/d;

  d_plus =  SY_NORM(p[n2].x - p[n1].x, p[n2].y - p[n1].y)
    + (da = SY_NORM(p[n1].x - p[i0].x, p[n1].y - p[i0].y) );

  d_minus = (db = SY_NORM(p[i0].x - p[p1].x, p[i0].y - p[p1].y) )
                + SY_NORM(p[p1].x - p[p2].x, p[p1].y - p[p2].y);

  x_s_plus  = (p[n2].x - p[i0].x)/d_plus;
  y_s_plus  = (p[n2].y - p[i0].y)/d_plus;
  
  x_s_minus = (p[i0].x - p[p2].x)/d_minus;
  y_s_minus = (p[i0].y - p[p2].y)/d_minus;
  
  x_ss = (x_s_plus - x_s_minus)/(da + db);
  y_ss = (y_s_plus - y_s_minus)/(da + db);
  
  curv = (x_s * x_s + y_s * y_s); 
  curv = sqrt(1.0/(curv * curv * curv));
  curv = (x_s * y_ss - y_s * x_ss) * curv;

  return curv;
}

int Sy_CurvatureSign (SyIpoFloat *p, int p2, int p1, int i0, 
		      int n1, int n2)
{ 
  double d, d_plus, d_minus, db, da, curv, x_s, y_s, 
         x_s_minus, y_s_minus, x_s_plus, y_s_plus, x_ss, y_ss;
  
  d  =  SY_NORM(p[n1].x - p[i0].x, p[n1].y - p[i0].y)
      + SY_NORM(p[i0].x - p[p1].x, p[i0].y - p[p1].y);
    
  x_s = (p[n1].x - p[p1].x)/d;
  y_s = (p[n1].y - p[p1].y)/d;

  d_plus =  SY_NORM(p[n2].x - p[n1].x, p[n2].y - p[n1].y)
    + (da = SY_NORM(p[n1].x - p[i0].x, p[n1].y - p[i0].y) );

  d_minus = (db = SY_NORM(p[i0].x - p[p1].x, p[i0].y - p[p1].y) )
                + SY_NORM(p[p1].x - p[p2].x, p[p1].y - p[p2].y);

  x_s_plus  = (p[n2].x - p[i0].x)/d_plus;
  y_s_plus  = (p[n2].y - p[i0].y)/d_plus;
  
  x_s_minus = (p[i0].x - p[p2].x)/d_minus;
  y_s_minus = (p[i0].y - p[p2].y)/d_minus;
  
  x_ss = (x_s_plus - x_s_minus)/(da + db);
  y_ss = (y_s_plus - y_s_minus)/(da + db);
  
  
  curv = x_s * y_ss - y_s * x_ss;

  if (SY_ALV(curv) < 1.0e-3) return 0;
  else if (curv > 0.0) return 1;
  else return -1;

}

#undef sq

void Sy_GaussSmoothFFT (SyIpo *ipo, SyIpoFloat *smipo, int pdim, float sigma)
{
  int i, k, n, support, lo;
  float n0, factor, *datax, *datay;
  float exp_factor, mult;

  support = floor(15.0*sigma+0.5);

  n0 = support + pdim;
  n  = floor(exp(ceil(log(n0)/log(2.0))*log(2.0))+0.5);

  factor = 2.0/(float)n;
  exp_factor = -0.5 * M_PI * M_PI * sigma* sigma /((float)(n*n));
  /* note: the proper expression for exp_factor would be 
     exp_factor = -2.0 * M_PI * M_PI * sigma* sigma /((float)(n*n));
     but due to the implementation of `realft' in numerical recipes
     we have to compensate for a reduction of the sampling rate 
     by 2 -> leads to an additional factor of 0.25 */

  /* fprintf (stderr, "support: %d, n: %d \n", support, n); */

  datax = (float*) Sy_Malloc(sizeof(float)*(n+1));
  datay = (float*) Sy_Malloc(sizeof(float)*(n+1));

  lo = support/2;
  
  /* copy data */
  for (i = lo+1, k=1; i <= lo+pdim; i++, k++){ 
    datax[i] = ipo[k].x;
    datay[i] = ipo[k].y;
  }

  /* padding/wraparound */
  for (i = lo+pdim+1, k = 1; i <= n; i++, k++){ 
    datax[i] = ipo[k%pdim].x;
    datay[i] = ipo[k%pdim].y;
  }

  for (i = lo, k = 0; i >= 1; i--, k++){ 
    datax[i] = ipo[pdim-(k%pdim)].x;
    datay[i] = ipo[pdim-(k%pdim)].y;
  }

  realft(datax, n, 1);
  realft(datay, n, 1);
  
  datax[2] = datay[2] = 0.0;
 
  for (i = 1; i <= n; i++){ 
    mult = exp(exp_factor*(double)((i-1)*(i-1)));
    datax[i] *= mult; datay[i] *= mult;
  }
  
  realft(datax, n, -1);
  realft(datay, n, -1);

  for (i = lo+1, k=1; i <= lo+pdim; i++,k++){ 
    smipo[k].x = factor*datax[i];
    smipo[k].y = factor*datay[i];
  }

  Sy_Free(datax);
  Sy_Free(datay);
}
