/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 *
 * Low Level IO for Delaunay triangulation 
 *
 */

static char sy_TrianIOId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Lang.h"
#include "Sy_Yacc.h"

int Sy_Trian_fprintf (FILE *fp, char *indata, unsigned int iomode)
{
  SyTriDAC* tri;
  int	  i, io_result;

  tri = (SyTriDAC*) indata;

  io_result = fprintf (fp, 
     "/* triangulation info: */\npointdim = %d; edgedim = %d; \
reguldim = %d;\nedgerefdim = %d;\n", 
		       (int) tri->pdim, (int) tri->edim, (int) tri->regdim,
		       (int) tri->erefdim);
  if (io_result == EOF) return (io_result); 

  io_result = fprintf (fp, "links = %s;\n", tri->is_linkinfo ? "true":"false");
  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_TRUE);
  fprintf (fp, "\n");

  for (i = 1; i <= tri->pdim; i++){

    io_result =
      fprintf (fp, "%d %d\n", 
	       (int) tri->point[i].x, 
	       (int) tri->point[i].y);
    
    if (io_result == EOF) return (io_result); 
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)tri->point);
    tri->point = (SyPoint*) NULL;
    tri->pdim  = 0;
  }
  
  for (i = 1; i <= tri->edim ; i++){
    
    io_result =
      fprintf (fp, "%6d %6d %6d %6d\n",
	       (int) tri->edge[i].head_edge,   
	       (int) tri->edge[i].tail_edge,
	       (int) tri->edge[i].head_point, 
	       (int) tri->edge[i].tail_point);
    
    if (io_result == EOF) return (io_result); 
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)tri->edge);
    tri->edge = (SyTriEdge*) NULL;
  }

  if (tri->regdim != 0){
    for (i = 1; i <= tri->regdim; i++){
      io_result =
	fprintf (fp, "%f %f %f\n", 
		 tri->regul[i].cont,
		 tri->regul[i].narr,
		 tri->regul[i].prox);
    
      if (io_result == EOF) return (io_result); 
    }
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    Sy_Free ((SY_VOID*)tri->regul);
    tri->regul = (SyTriReg*) NULL;
    tri->regdim  = 0;
  }

  if (tri->is_linkinfo){
    for (i = 1; i <= tri->edim; i++){
      io_result = fprintf (fp, "%d\n", (int)tri->is_hardlink[i]);
      if (io_result == EOF) return (io_result); 
    }
  }

  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    Sy_Free ((SY_VOID*)tri->is_hardlink);
    tri->is_hardlink = (Boolean*) NULL;
    tri->is_linkinfo = 0;
    tri->edim = 0;
  }

  if (tri->erefdim != 0){
    for (i = 1; i <= tri->erefdim; i++){
      io_result = fprintf (fp, "%d\n", tri->eref[i]);
      if (io_result == EOF) return (io_result); 
    }
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
    Sy_Free ((SY_VOID*)tri->eref);
    tri->eref = (SyEdgeRef*) NULL;
    tri->erefdim = 0;
  }

  return (1);
}

int Sy_Trian_fscanf (FILE *fp, SyTriDAC **ret_tri, unsigned int iomode)
{
  SyTriDAC *tri;
  int      i, io_result;
  int      buff [SY_NBUFF];
  double   fbuff[SY_NBUFF];
  int      bytes = 0;

  SyDataObject *dobj;

  tri = (SyTriDAC*) Sy_Malloc (sizeof(SyTriDAC));
  Sy_InitTrian(tri);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO", "Sy_Trian_fscanf","missing pointdim info...\n","");}
  tri->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO", "Sy_Trian_fscanf","missing edgedim info...\n","");}
  tri->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "reguldim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO","Sy_Trian_fscanf","missing reguldim info...\n","");}
  tri->regdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgerefdim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO","Sy_Trian_fscanf",
	     "missing edgerefdim info...\n","");}
  tri->erefdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "links");
  if (dobj == NULL) tri->is_linkinfo = SY_FALSE;
  else tri->is_linkinfo = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  tri->point  = Sy_AllocPoint (tri->pdim);
  tri->edge   = Sy_AllocTriEdge (tri->edim);

  if (tri->regdim != 0)  tri->regul = Sy_AllocTriRegul (tri->regdim);
  if (tri->erefdim != 0) tri->eref  = Sy_AllocEdgeRef (tri->erefdim);
  if (tri->is_linkinfo)  tri->is_hardlink = Sy_AllocBoolean (tri->edim);

  for (i = 1; i <= tri->pdim; i++){
    
    io_result = fscanf (fp, "%d %d", &buff[0], &buff[1]);  
    tri->point[i].x = buff[0]; 
    tri->point[i].y = buff[1];
    
    if (io_result != 2) return (EOF); 
  }

  for (i = 1; i <= tri->edim; i++){

    io_result =
      fscanf (fp, "%d %d %d %d", &buff[0], &buff[1], &buff[2], &buff[3]);
    
    tri->edge[i].head_edge   = buff[0];   
    tri->edge[i].tail_edge   = buff[1];
    tri->edge[i].head_point  = buff[2];  
    tri->edge[i].tail_point  = buff[3];

    if (io_result != 4) return (EOF); 
  }
  
  if (tri->regdim != 0){
    for (i = 1; i <= tri->regdim; i++) {
      io_result =
	fscanf (fp, "%lf %lf %lf", &fbuff[0], &fbuff[1], &fbuff[2]); 
      tri->regul[i].cont = fbuff[0];
      tri->regul[i].narr = fbuff[1];
      tri->regul[i].prox = fbuff[2];

      if (io_result != 3) return (EOF); 
    }
  }

  if (tri->is_linkinfo){
    for (i = 1; i <= tri->edim; i++) {
      io_result = fscanf (fp, "%d", &buff[0]); 
      tri->is_hardlink[i] = buff[0];
      
      if (io_result != 1) return (EOF); 
    }    
  }

  if (tri->erefdim != 0){
    for (i = 1; i <= tri->erefdim; i++) {
      io_result = fscanf (fp, "%d", &buff[0]); 
      tri->eref[i] = buff[0];

      if (io_result != 1) return (EOF); 
    }
  }

  *ret_tri = tri; 
  return (bytes);
}

int Sy_Trian_fwrite (FILE *fp, char *indata, unsigned int iomode)
{
  SyTriDAC* tri;
  int	    i, io_result;

  tri = (SyTriDAC*) indata;

  io_result = fprintf (fp, 
     "/* triangulation info: */\npointdim = %d; edgedim = %d; \
reguldim = %d;\nedgerefdim = %d;\n", 
		       (int) tri->pdim, (int) tri->edim, (int) tri->regdim,
		       (int) tri->erefdim);
  if (io_result == EOF) return (io_result); 

  io_result = fprintf (fp, "links = %s;\n", tri->is_linkinfo ? "true":"false");
  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_FALSE);

  io_result = fwrite ((SY_VOID*)tri->point, sizeof(SyPoint), tri->pdim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)tri->point);
    tri->point = (SyPoint*) NULL;
    tri->pdim  = 0;
  }
  
  io_result = fwrite ((SY_VOID*)tri->edge, sizeof(SyTriEdge), tri->edim+1, fp);  
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)tri->edge);
    tri->edge = (SyTriEdge*) NULL;
  }

  if (tri->regdim != 0){
    io_result = fwrite ((SY_VOID*)tri->regul, sizeof(SyTriReg), tri->regdim+1, fp);
    if (io_result == 0) return (EOF);
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)tri->regul);
      tri->regul  = (SyTriReg*) NULL;
      tri->regdim = 0;
    }
  }  

  if (tri->is_linkinfo){
    io_result = fwrite ((SY_VOID*)tri->is_hardlink, sizeof(Boolean), 
			tri->edim+1, fp);
    if (io_result == 0) return (EOF);
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)tri->is_hardlink);
      tri->is_hardlink = (Boolean*) NULL;
      tri->is_linkinfo = 0;
      tri->edim = 0;
    }
  }

  if (tri->erefdim != 0){
    io_result = fwrite ((SY_VOID*)tri->eref, sizeof(SyEdgeRef), tri->erefdim+1, fp);
    if (io_result == 0) return (EOF);
    if (SY_IS_FREE_AFTER_WRITE(iomode)) {
      Sy_Free ((SY_VOID*)tri->eref);
      tri->eref    = (SyEdgeRef*) NULL;
      tri->erefdim = 0;
    }
  }  
  return (1);
}

int Sy_Trian_fread (FILE *fp, SyTriDAC **ret_tri, unsigned int iomode)
{
  SyTriDAC  *tri;
  int     io_result;
  int     bytes = 0;
  SyDataObject *dobj;

  tri = (SyTriDAC*) Sy_Malloc (sizeof(SyTriDAC));
  Sy_InitTrian(tri);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO", "Sy_Trian_fread","missing pointdim info...\n","");}
  tri->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO", "Sy_Trian_fread","missing edgedim info...\n","");}
  tri->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "reguldim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO","Sy_Trian_fread","missing reguldim info...\n","");}
  tri->regdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgerefdim");
  if (dobj == NULL){
    Sy_Error("Sy_TrianIO","Sy_Trian_fread",
	     "missing edgerefdim info...\n","");}
  tri->erefdim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "links");
  if (dobj == NULL) tri->is_linkinfo = SY_FALSE;
  else tri->is_linkinfo = strcmp(dobj->t.s_val, "true") ? SY_FALSE : SY_TRUE; 

  tri->point  = Sy_AllocPoint (tri->pdim);
  tri->edge   = Sy_AllocTriEdge (tri->edim);

  if (tri->regdim != 0)  tri->regul = Sy_AllocTriRegul (tri->regdim);
  if (tri->erefdim != 0) tri->eref  = Sy_AllocEdgeRef (tri->erefdim);
  if (tri->is_linkinfo)  tri->is_hardlink = Sy_AllocBoolean (tri->edim);

  io_result = fread ((SY_VOID*)tri->point, sizeof(SyPoint), tri->pdim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)tri->edge, sizeof(SyTriEdge), tri->edim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  if (tri->regdim != 0){
    io_result = fread ((SY_VOID*)tri->regul, sizeof(SyTriReg), tri->regdim+1, fp);
    if (io_result == 0) return (EOF);
  }

  if (tri->is_linkinfo){
    io_result = fread ((SY_VOID*)tri->is_hardlink, sizeof(Boolean), 
		       tri->edim+1, fp);
    if (io_result == 0) return (EOF);
  }

  if (tri->erefdim != 0){
    io_result = fread ((SY_VOID*)tri->eref, sizeof(SyEdgeRef), tri->erefdim+1, fp);
    if (io_result == 0) return (EOF);
  }

  *ret_tri = tri;
  return (bytes);
}






