/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Voronoi to Triangulation Transform
 *
 */

static char sy_Voro2TrianId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

typedef struct _SyRotInfo{
  int last_edge;
} SyRotInfo;

static enum SyTrianMode _SKE_trian_mode = TRIAN_SHORT;
static int _SKE_is_edgeref = 1;

static SyTriDAC*  tri;
static SyVoroDAC* vd;

static int ecount    = 0;
static int ref_edge  = 0;
static int last_edge = 0;

void Sy_Vor2TriInit(void);
void Sy_Vor2TriTrav(void);
void Sy_Vor2TriRotate(void);
void Sy_Vor2TriTerm(void);
void Sy_Vor2TriExit(void);

static void Sy_SimpleTrianRotate (SyTriEdge *edge, int ref_edge, int ref_point, SyRotInfo *info);

void Sy_Voronoi2Trian (SyVoroDAC *vd_in, SyTriDAC *ret_tri)
{     
  SyWalktask wt;
  int i;
  int next_edge, current_head_vertex;

  int left_point, right_point,
      tail_vertex, head_vertex,
      head_edge, tail_edge,
      tail_point, head_point,
      ref_edge, ref_point;

  Boolean *plabel;

  SyRotInfo info;
  
  Sy_ResetWalktask (&wt);

  syWstate.vd = vd_in;
  tri   = ret_tri;
  vd    = vd_in;

  tri->pdim    = vd_in->pdim;
  tri->edim    = vd_in->edim;
  tri->regdim  = 0;
  tri->erefdim = vd_in->edim;

  tri->is_linkinfo = 0;

  tri->point   = Sy_AllocPoint(tri->pdim);
  tri->edge    = Sy_AllocTriEdge(tri->edim);

  tri->label.v_= Sy_AllocVertexLabel(vd->vdim);
  tri->eref    = Sy_AllocEdgeRef (tri->edim);

  memcpy ((char*)tri->point, (char*)vd->point, (vd->pdim+1)*sizeof(SyPoint)); 
  
  wt.user_function[(int)INIT] = (void(*)())Sy_Vor2TriInit;
  wt.user_function[(int)TRAV] = (void(*)())Sy_Vor2TriTrav;
  wt.user_function[(int)TERM] = (void(*)())Sy_Vor2TriTerm;
  wt.user_function[(int)ROT ] = (void(*)())Sy_Vor2TriRotate;
  wt.user_function[(int)FEXIT] = (void(*)())Sy_Vor2TriExit;

  Sy_InitVoronoiTraverse(vd, &wt);    
  
  Sy_Vor2TriInit();
  
  /* Pass 1 */

  for (i = 1; i <= vd->edim; i++){
    tail_vertex = vd->edge[i].tail_vertex;
    head_vertex = vd->edge[i].head_vertex;

    if (vd->edge[i].rank < 2) tri->label.v_[head_vertex] = 1;

    if (tri->label.v_[tail_vertex] == 0){
      next_edge            = vd->edge[i].tail_edge;
      current_head_vertex  = tail_vertex;
      syWstate.start.this_edge   = i;
      syWstate.start.next_edge   = next_edge;
      syWstate.start.head_vertex = tail_vertex;

      Sy_VoronoiRotate (vd, &next_edge, &current_head_vertex);
    }

    if (tri->label.v_[head_vertex] == 0){
      next_edge            = vd->edge[i].head_edge;
      current_head_vertex  = head_vertex;
      syWstate.start.this_edge   = i;
      syWstate.start.next_edge   = next_edge;
      syWstate.start.head_vertex = head_vertex;

      Sy_VoronoiRotate (vd, &next_edge, &current_head_vertex);
    }
  } 
  
  /* Pass 2 : Convex Hull References */

  plabel = Sy_AllocBoolean (tri->pdim);

  for (i = 1; i <= vd->edim; i++){
    if (vd->edge[i].rank >= 2) continue;

    ref_edge = tri->eref[i];
    tail_point = tri->edge[ref_edge].tail_point;
    head_point = tri->edge[ref_edge].head_point;
    tail_edge  = tri->edge[ref_edge].tail_edge;
    head_edge  = tri->edge[ref_edge].head_edge;
    if ((plabel[tail_point] == 0) && (tail_edge != 0)){
      ref_point = tail_point;
    }
    else if ((plabel[head_point] == 0) && (head_edge != 0)){
      ref_point = head_point;
    }
    else continue;

    Sy_SimpleTrianRotate (tri->edge, ref_edge, ref_point, &info);
    
    if (tri->edge[info.last_edge].tail_point == ref_point){
      tri->edge[info.last_edge].tail_edge = ref_edge;
    }
    else{
      tri->edge[info.last_edge].head_edge = ref_edge;
    }

    plabel[ref_point] = 1;
  }
  
  Sy_Vor2TriExit();

}

void Sy_Vor2TriInit (void)
{
  ecount = 0;
}

void Sy_Vor2TriExit (void)
{
  tri->edim = ecount;

  free ((char*)tri->label.v_);
  if (_SKE_is_edgeref == 0){
    free ((char*)tri->eref);
    tri->erefdim = 0;
  }
}

void Sy_Vor2TriTrav (void)
{
  ref_edge  = syWstate.current.next_edge;
  last_edge = 0;
}


void Sy_Vor2TriRotate(void)
{
  int this_edge;
  int this_triedge,
      last_triedge;

  this_edge = syWstate.current.next_edge;

  if (tri->eref[this_edge] == 0){
    ecount++;
    tri->edge[ecount].tail_point  = vd->edge[this_edge].left_point;
    tri->edge[ecount].head_point  = vd->edge[this_edge].right_point;
    tri->eref[this_edge] = ecount;
  }   
  if (last_edge != 0){
    last_triedge = tri->eref[last_edge];
    this_triedge = tri->eref[this_edge];
    if ((tri->edge[last_triedge].tail_point == tri->edge[this_triedge].tail_point)||
	(tri->edge[last_triedge].tail_point == tri->edge[this_triedge].head_point)){
      tri->edge[last_triedge].tail_edge = this_triedge;
    }
    else{
      tri->edge[last_triedge].head_edge = this_triedge;
    }
  }
  last_edge = this_edge;
}

void Sy_Vor2TriTerm (void)
{
  int last_triedge,
      ref_triedge;

  if (last_edge != 0){
    last_triedge = tri->eref[last_edge];
    ref_triedge  = tri->eref[ref_edge];
    if ((tri->edge[last_triedge].tail_point == tri->edge[ref_triedge].tail_point)||
	(tri->edge[last_triedge].tail_point == tri->edge[ref_triedge].head_point)){
      tri->edge[last_triedge].tail_edge = ref_triedge;
    }
    else{
      tri->edge[last_triedge].head_edge = ref_triedge;
    }
  }
}

void Sy_AddHardlink (SyIpoDAC *ipd, SyTriDAC *tri)
{
  int i;

  tri->is_linkinfo = 1;
  tri->is_hardlink = Sy_AllocBoolean (tri->edim);
  
  for (i = 1; i <= tri->edim; i++){
    tri->is_hardlink[i] = Sy_IpoConnected (ipd, tri->edge[i].tail_point,
					     tri->edge[i].head_point);
  }
}
    

void Sy_SetEdgeref (int val)
{
  _SKE_is_edgeref = val;
}

void Sy_SetTrianMode (enum SyTrianMode mode)
{
  _SKE_trian_mode = mode;
}


static void Sy_SimpleTrianRotate (SyTriEdge *edge, int ref_edge, int ref_point, SyRotInfo *info)
{
  int next_edge, last_edge;

  next_edge   = (edge[ref_edge].tail_point == ref_point) ? edge[ref_edge].tail_edge:
                                                           edge[ref_edge].head_edge;

  if (next_edge == 0){
    info->last_edge = ref_edge;
    return;
  }

  while ((next_edge != ref_edge) && (next_edge != 0)){
    last_edge = next_edge;
    if (ref_point == edge[next_edge].tail_point){
      next_edge = edge[next_edge].tail_edge;
    }
    else{
      next_edge = edge[next_edge].head_edge;
    }
  }

  info->last_edge = last_edge;
}





