
/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Low Level IO for VD
 *  
 */

static char sy_VoroIOId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Lang.h"
#include "Sy_Yacc.h"

/*********************************************************************
 * function int Sy_Voro_fprintf (fp, vd)
 * 
 * purpose: sends Voronoi-diagram data to a file referenced through fp
 *          (ASCII format)
 *
 * param:   [input] SyVoroDAC* vd   -  Voronoi-diagram data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Voro_fprintf (FILE *fp, char *indata, unsigned int iomode)
{
  SyVoroDAC* vd;
  int	   i, io_result;

  vd = (SyVoroDAC*) indata;
  
  io_result = fprintf (fp, 
     "/* contour info: */\npointdim = %d; edgedim = %d; vertexdim = %d;\n", 
		       (int) vd->pdim, (int) vd->edim, (int) vd->vdim);

  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_TRUE);
  fprintf (fp, "\n");

  for (i = 1; i <= vd->pdim; i++){
    
    io_result = fprintf (fp, "%d %d\n", 
			 (int) vd->point[i].x, (int) vd->point[i].y);
		
    if (io_result == EOF) return (io_result); 
  }
  if (SY_IS_FREE_AFTER_WRITE(iomode)){ 
    Sy_Free ((SY_VOID*)vd->point);
    vd->point = (SyPoint*) NULL;
    vd->pdim  = 0;
  }

  for (i = 1; i <= vd->edim; i++){

    io_result =
      fprintf (fp, "%6d %6d %6d %6d %6d %6d %2d %6d %6d %6d %6d\n",
	       (int) vd->edge[i].head_edge,   
	       (int) vd->edge[i].tail_edge,
	       (int) vd->edge[i].head_vertex, 
	       (int) vd->edge[i].tail_vertex,
	       (int) vd->edge[i].left_point,  
	       (int) vd->edge[i].right_point,
	       (int) vd->edge[i].rank,
	       (int) vd->edge[i].dirx, 
	       (int) vd->edge[i].diry,
	       (int) vd->edge[i].xoff, 
	       (int) vd->edge[i].yoff);

    if (io_result == EOF) return (io_result); 
  }
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vd->edge);
    vd->edge = (SyEdge*) NULL;
    vd->edim = 0;
  }

  for (i = 1; i <= vd->vdim; i++) {
    io_result =
      fprintf (fp, "%20.0lf %20.0lf %20.0lf\n", 
	       vd->vertex[i].xden, 
	       vd->vertex[i].yden,
	       vd->vertex[i].num);
    
    if (io_result == EOF) return (io_result); 
  }
  
  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
   Sy_Free ((SY_VOID*)vd->vertex);
   vd->vertex = (SyVertex*) NULL;
   vd->vdim   = 0;
 }

  return (1);
}


/**********************************************************************
 * function int Sy_Voro_fscanf (fp, vd)
 * 
 * purpose: gets Voronoi-diagram data from a file referenced through fp
 *          (ASCII format)
 *
 * param:   [inout] SyVoroDAC* vd   -  Voronoi-diagram data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Voro_fscanf (FILE *fp, SyVoroDAC **ret_vd, unsigned int iomode)
{
  SyVoroDAC* vd;
  int	   i, io_result;
  int      buff [SY_NBUFF];
  double   fbuff[SY_NBUFF];
  int      bytes = 0;
  SyDataObject *dobj;

  vd = (SyVoroDAC*) Sy_Malloc (sizeof(SyVoroDAC));
  Sy_InitVoronoi(vd);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fscanf", "missing pointdim info...\n","");}
  vd->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fscanf", "missing edgedim info...\n","");}
  vd->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "vertexdim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fscanf", "missing vertexdim info...\n","");}
  vd->vdim = dobj->t.i_val; 

  vd->point  = Sy_AllocPoint (vd->pdim);
  vd->edge   = Sy_AllocEdge  (vd->edim);
  vd->vertex = Sy_AllocVertex(vd->vdim);
  
  for (i = 1; i <= vd->pdim; i++){

    io_result =  fscanf (fp, "%d %d", &buff[0], &buff[1]);  
    vd->point[i].x = buff[0]; 
    vd->point[i].y = buff[1];
    
    if (io_result != 2) return (EOF); 
  }

  for (i = 1; i <= vd->edim; i++){
	  
    io_result =
      fscanf (fp, "%d %d %d %d %d %d %d %d %d %d %d",
	      &buff[0], &buff[1], &buff[2], &buff[3],
	      &buff[4], &buff[5], &buff[6], &buff[7], 
	      &buff[8], &buff[9], &buff[10]);
    
    vd->edge[i].head_edge    = buff[0];   
    vd->edge[i].tail_edge    = buff[1];
    vd->edge[i].head_vertex  = buff[2]; 
    vd->edge[i].tail_vertex  = buff[3];
    vd->edge[i].left_point   = buff[4];  
    vd->edge[i].right_point  = buff[5];
    vd->edge[i].rank         = buff[6];
    vd->edge[i].dirx         = buff[7]; 
    vd->edge[i].diry         = buff[8];
    vd->edge[i].xoff         = buff[9]; 
    vd->edge[i].yoff         = buff[10];
    
    if (io_result != 11) return (EOF); 
  }
  
  for (i = 1; i <= vd->vdim; i++) {
    
    io_result = fscanf (fp, "%lf %lf %lf", &fbuff[0], &fbuff[1], &fbuff[2]); 
    vd->vertex[i].xden = fbuff[0];
    vd->vertex[i].yden = fbuff[1];
    vd->vertex[i].num  = fbuff[2];
    
    if (io_result != 3) return (EOF); 
  }
  *ret_vd = vd;
  return (bytes);
}


/*********************************************************************
 * function int Sy_Voro_fwrite (fp, vd)
 * 
 * purpose: sends Voronoi-diagram data to a file referenced through fp
 *          (BINARY format)
 *
 * param:   [input] SyVoroDAC* vd   -  Voronoi-diagram data
 *          [input] FILE*   fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during write 
 *********************************************************************/

int Sy_Voro_fwrite (FILE *fp, char *indata, unsigned int iomode)
{
  SyVoroDAC* vd;
  int	io_result;

  vd = (SyVoroDAC*) indata;
 
  io_result = fprintf (fp, 
     "/* contour info: */\npointdim = %d; edgedim = %d; vertexdim = %d;\n", 
		       (int) vd->pdim, (int) vd->edim, (int) vd->vdim);

  if (io_result == EOF) return (io_result); 

  Sy_Wrbegload (fp, "Sy_GenLoader", SY_FALSE, SY_FALSE);
  
  io_result = fwrite ((SY_VOID*)vd->point, sizeof(SyPoint), vd->pdim+1, fp);
  if (io_result == 0) return (EOF); 
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vd->point);
    vd->point = (SyPoint*) NULL;
    vd->pdim  = 0;
  }

  io_result = fwrite ((SY_VOID*)vd->edge, sizeof(SyEdge), vd->edim+1, fp);
  if (io_result == 0) return (EOF);   
  if (SY_IS_FREE_AFTER_WRITE(iomode)){
    Sy_Free ((SY_VOID*)vd->edge);
    vd->edge = (SyEdge*) NULL;
    vd->edim = 0;
  }

  io_result = fwrite ((SY_VOID*)vd->vertex, sizeof(SyVertex), vd->vdim+1, fp);
  if (io_result == 0) return (EOF); 

  if (SY_IS_FREE_AFTER_WRITE(iomode)) {
   Sy_Free ((SY_VOID*)vd->vertex);
   vd->vertex = (SyVertex*) NULL;
   vd->vdim   = 0;
 }  
  return (1);
}

/**********************************************************************
 * function int Sy_Voro_fread (fp, vd)
 * 
 * purpose: gets Voronoi-diagram data from a file referenced through fp
 *          (BINARY format)
 *
 * param:   [inout] SyVoroDAC* vd   -  Voronoi-diagram data
 *          [input] FILE*    fp   -  file pointer
 *
 * retval:  EOF (-1) indicates an error during read 
 *********************************************************************/

int Sy_Voro_fread (FILE *fp, SyVoroDAC **ret_vd, unsigned int iomode)
{
  SyVoroDAC* vd;
  int	   io_result, bytes = 0;
  SyDataObject *dobj;
  
  vd = (SyVoroDAC*) Sy_Malloc (sizeof(SyVoroDAC));  
  Sy_InitVoronoi(vd);

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "pointdim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fread", "missing pointdim info...\n","");}
  vd->pdim = dobj->t.i_val; 
  
  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "edgedim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fread", "missing edgedim info...\n","");}
  vd->edim = dobj->t.i_val; 

  dobj = Sy_GetDataSameObject (&sy_yacc_tree, "vertexdim");
  if (dobj == NULL){
    Sy_Error ("Sy_VoroIO", "Sy_Voro_fread", "missing vertexdim info...\n","");}
  vd->vdim = dobj->t.i_val; 

  vd->point  = Sy_AllocPoint (vd->pdim);
  vd->edge   = Sy_AllocEdge  (vd->edim);
  vd->vertex = Sy_AllocVertex(vd->vdim);
  
  io_result = fread ((SY_VOID*)vd->point, sizeof(SyPoint), vd->pdim+1, fp);

  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)vd->edge, sizeof(SyEdge), vd->edim+1, fp);
  
  if (io_result == 0) return (EOF); 
  
  io_result = fread ((SY_VOID*)vd->vertex, sizeof(SyVertex), vd->vdim+1, fp);
  
  if (io_result == 0) return (EOF);

  *ret_vd = vd; 
  return (bytes);
}


