/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * computation of the Voronoi Diagram
 *
 */

static char sy_VoronoiId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

#ifndef FASTVORO
static char version[] = "VD Version 10 (exp., precision)";
#else
static char version[] = "VD Version 10 (exp., fast)";
#endif

static void Sy_InitVoro(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
			SyHull *hull, SyComb *comb);

static void Sy_EndVoro(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
		       SyVertex *vertex, SyVoroStack *stack);
 
static void Sy_ComputeSupport(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
			      SyComb *comb, SyHull *hull, SySupport *support);
     
static void Sy_PrepComb(SyVoroDAC *vd, SyPoint *point, SyComb *comb);
 
static void Sy_LinkVoro(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
			SyVertex *vertex, SyComb *comb, SyHull *hull, 
			SySupport *support, SyVoroStack *stack);
     
static int  Sy_GetScanline(SyVoroDAC *vd, SyPoint *point, SyComb *comb);
     
static void Sy_PrepComb(SyVoroDAC *vd, SyPoint *point, SyComb *comb);

static void Sy_InitPrescan(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
			   SyPredat *predat);

static void Sy_VoroPrescan(SyVoroDAC *vd, SyPoint *point, SyEdge *edge, 
			   SyVertex *vertex, SyComb *comb, SyHull *hull, 
			   SyPredat *predat);

static int Sy_Intersection(SyEdge *e1, SyEdge *e2, SyVertex *v);

static void swap(SyEdge *e);

/**********************************************************************
 * function void voronoi (vd, is_verbose)
 * 
 * purpose: computes the complete Voronoi-diagram (VD)
 *
 * param:   vd (SyVoroDAC*)  - pointer to VD data
 *          is_verbose
 *********************************************************************/
void Sy_Voronoi (SyVoroDAC *vd, int is_verbose)
{
	SyComb            comb;
	SyHull            hull;
	SySupport         supp, supp_new;
	SyVoroStack       stack;

	SyGenInfo         info;
	SyPredat          predat;

	int             line_nr = 1, diff;
	static char     rotor[] = "|/-\\";
	
        sprintf (syStrBuf, "%s ", version);
	Sy_Message (SY_FMTNL, syStrBuf);

	Sy_AllocVoroEnvironment (vd, &comb, &hull, &supp, &stack, &info);


	if (Sy_GetScanline (vd, vd->point, &comb)){
	  if (comb.n > comb.max_edges) Sy_ReallocVoroComb (&comb, &info);

	  if (comb.n > 1){
	    Sy_Message (SY_FMTNL, "mode: STANDARD (N) ");

	    Sy_InitVoro (vd, vd->point, vd->edge, &hull, &comb);
	    predat.vorop = VORO_SCAN;
	  }
	  else{
	    Sy_Message (SY_FMTNL, "mode: PRESCAN (N) ");

	    Sy_InitPrescan (vd, vd->point, vd->edge, &predat);
	    predat.vorop = VORO_PRESCAN;
	  }
	}
	
	while (predat.vorop == VORO_PRESCAN){

	  if (Sy_GetScanline(vd, vd->point, &comb)){
	    if (comb.n > comb.max_edges) Sy_ReallocVoroComb (&comb, &info);
	    Sy_VoroPrescan (vd, vd->point, vd->edge, vd->vertex, &comb, &hull, &predat);
	  }
	  else predat.vorop = VORO_NOP;
	}

	while (Sy_GetScanline (vd, vd->point, &comb)){

		if (comb.n > comb.max_edges)
		       Sy_ReallocVoroComb (&comb, &info);

		if (line_nr > info.max_lines)
		       Sy_ReallocVoroHullStack (&hull, &stack, &info);

		Sy_ComputeSupport (vd, vd->point, vd->edge, &comb, &hull, &supp);

		Sy_PrepComb (vd, vd->point, &comb);

		Sy_LinkVoro (vd, vd->point, vd->edge, vd->vertex, 
                           &comb, &hull, &supp, &stack);

		line_nr++;

               /*
		if (stack.e_ptr > 0){
                  sprintf (syStrBuf, "%%%% VORONOI::INCONSISTENCY DETECTED");
                  sprintf (syStrBuf, ", e_stack pointer = %d  ", stack.e_ptr);
		  Sy_SimpleError(SY_FMTNL, syStrBuf);
		}
		*/

		if (is_verbose){
                  sprintf (syStrBuf, 
			   "%c nr:%d y:%d points:%d edges:%d vertices:%d\r",
                           (char)rotor[(comb.phigh/SY_ROTORSPEED)%SY_RODIM],
			   line_nr, vd->point[comb.phigh].y, comb.phigh,
			   vd->edim, vd->vdim);
		  Sy_Message (SY_FMT, syStrBuf);
		}
        }                
	Sy_EndVoro (vd, vd->point, vd->edge, vd->vertex, &stack);
	Sy_FreeVoroEnvironment (vd, &comb, &hull, &supp, &stack, &info);
	Sy_Message(SY_FMTNL, " done.");
}

static void Sy_Bisector(SyPoint *p1, SyPoint *p2, SyEdge *e);
static void  Sy_SetEdge(SyEdge *e, SyEdge *e_base, int he, int te, int hv, 
			int tv, int lp, int rp, int rank);

/**********************************************************************
 * function void Sy_InitVoro (vd, point, edge, hull, comb)
 * 
 * purpose: initialization of 'vd', 'edge', 'hull'
 *
 * param:   vd      - pointer to Voronoi-diagram data
 *          point   - generating point array
 *          comb    - pointer to first scan-line data
 *          edge    - edge array
 *          hull    - pointer to initialized convex hull
 *********************************************************************/
static void Sy_InitVoro (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyHull *hull, SyComb *comb)
{
	int	    i;
	SyEdge	    chain;

	for (i = 1;i <= vd->pdim;i++){
	  vd->point[i].x <<= 1; vd->point[i].y <<= 1;
	}

	for (i = 1; i < comb->n; i++) {
		Sy_Bisector(&point[i], &point[i + 1], &chain);
		Sy_SetEdge(&edge[i], &chain, i-1, i-1, 0, 0, i, i+1, 0);
 	}
	edge[1].tail_edge = edge[1].head_edge = comb->n - 1;
	
	hull->left_count = hull->right_count = 0;
	hull->left_index = hull->right_index = 0;

	if (comb->n > 1){
	  hull->left_index  =
	  hull->right_index = hull->dim/(2*sizeof(int));
	  hull->left_count++; 
          hull->right_count++;
	  hull->left[hull->left_count]   = 
	    hull->right[hull->right_count] = 1;
	  hull->elem[hull->left_index] =
	    hull->elem[hull->right_index] = 1;
	}

	if (comb->n > 2) {
	  hull->right_index++;
	  hull->right[hull->right_count] = comb->n - 1;
	  hull->elem [hull->right_index] = comb->n - 1;
	}
	vd->edim = comb->n - 1;
	vd->vdim = 0; 
}

static void Sy_InitPrescan (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyPredat *predat)
{
  int i;

  for (i = 1;i <= vd->pdim;i++){
    vd->point[i].x <<= 1; vd->point[i].y <<= 1;
  }

  predat->line_nr      = 1;
  predat->stripcount   = 1;
  predat->pstart       = point[1];
  predat->ref_vec.dirx = 0;
  predat->ref_vec.diry = 0;
  vd->edim = vd->vdim  = 0;
}

enum SyPreScanMode {LSke_PRESCAN_DOWN, LSke_PRESCAN_UP, LSke_PRESCAN_HALT};

static void Sy_VoroPrescan (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyVertex *vertex, SyComb *comb, SyHull *hull, SyPredat *predat)
{
  SyEdge    chain;
  SyVertex  vcomb, vhull, vlast;

  SyIntVector tvec;

  double  a, b;

  enum SyPreScanMode mode;

  int     oper, i, mem_count = 0;

  int     last_hindex  = 0,
          last_eindex  = 0,
          last_eindex2 = 0,
          last_vertex  = 0,

          eindex,
          eindex2,
          hindex,
          vindex,
          combindex,
          first_hindex;

  int     left_point,     right_point, term_point,
          is_point_left,  is_point_right,
          is_left_nondeg, is_right_nondeg;

  int     comb_intersect, hull_intersect, rot, hull_done = 0;

  predat->line_nr++;

  if (comb->n == 1){
    if (predat->stripcount == 1){
      predat->ref_vec.dirx = point[2].x - point[1].x;
      predat->ref_vec.diry = point[2].y - point[1].y;
      Sy_Bisector(&point[1], &point[2], &chain);
      Sy_SetEdge(&edge[1], &chain, 0, 0, 0, 0, 1, 2, 0);
      predat->stripcount++;
      vd->edim++;
    }
    else{
      tvec.dirx = point[comb->phigh].x - point[comb->phigh-1].x;
      tvec.diry = point[comb->phigh].y - point[comb->phigh-1].y;
      
      if (predat->ref_vec.dirx * tvec.diry == predat->ref_vec.diry * tvec.dirx){
	Sy_Bisector(&point[comb->phigh-1], &point[comb->phigh], &chain);
	Sy_SetEdge(&edge[comb->phigh-1], &chain, 0, 0, 0, 0, comb->phigh-1, comb->phigh, 0);
	predat->stripcount++;
	vd->edim++;
      }
      else
	predat->vorop = VORO_SCAN;
    }
  }else{
    if (predat->stripcount == 1){
      predat->ref_vec.dirx = point[2].x - point[1].x;
      predat->ref_vec.diry = point[2].y - point[1].y;
    }
    predat->vorop = VORO_SCAN;
  }	
  
  if (predat->vorop == VORO_PRESCAN) return;

  /* build first instance of nontrivial VD */

  Sy_PrepComb (vd, point, comb);  

  vlast.xden = vlast.yden = vlast.num = 0.0;

  left_point = comb->phigh;
  is_point_left = ((point[left_point].x - predat->pstart.x) * predat->ref_vec.diry) >
                  ((point[left_point].y - predat->pstart.y) * predat->ref_vec.dirx);

  is_point_right= ((point[comb->plow].x - predat->pstart.x) * predat->ref_vec.diry) <
                  ((point[comb->plow].y - predat->pstart.y) * predat->ref_vec.dirx);

  is_left_nondeg  = is_point_left;
  is_right_nondeg = is_point_right;

  combindex       = comb->n-1;
  hull_done       = 0;

  if (is_point_left){
    mode = LSke_PRESCAN_DOWN; right_point = 1;
    hindex    = (predat->stripcount > 1)? 1 : 0;
    hull_done = (predat->stripcount > 1)? 0 : 1; 
    for (i = 1; i <= predat->stripcount-1; i++){
      swap (&edge[i]);
    }

  }
  else{
    hindex = predat->stripcount - 1;
    mode = LSke_PRESCAN_UP; right_point = predat->stripcount;
  }

  term_point = is_point_right? 1 : predat->stripcount;


  while (mode != LSke_PRESCAN_HALT){
    if ((left_point == comb->plow) && (right_point == term_point)){
      mode = LSke_PRESCAN_HALT;
      break;
    }
    hull_intersect = comb_intersect = 0;

    Sy_Bisector(&point[left_point], &point[right_point], &chain);

    comb_intersect = ((combindex > 0) &&
		      Sy_Intersection(&chain, &comb->elem[combindex], &vcomb));

    if (!hull_done){
      is_point_left = ((point[left_point].x - predat->pstart.x)*predat->ref_vec.diry)>
                      ((point[left_point].y - predat->pstart.y)*predat->ref_vec.dirx);
      if (is_point_left && (mode != LSke_PRESCAN_DOWN)){
	hull_intersect = 0;
      }else{
        hull_intersect = Sy_Intersection(&chain, &edge[hindex], &vhull);
      }
    }

    rot =((vlast.xden * vhull.num)==(vlast.num * vhull.xden))
         && (mem_count != 0);
 
    a = vhull.xden * vcomb.num;
    b = vhull.num  * vcomb.xden;
    
    if ( ((a > b) && hull_intersect && comb_intersect)
	|| (hull_intersect && !comb_intersect))         oper = 1;
    if ( ((a < b) && hull_intersect && comb_intersect)
	|| (comb_intersect && !hull_intersect))         oper = 2;
    if ( (oper == 1) && rot )                           oper = 3;
    if ( (a == b) && hull_intersect && comb_intersect)  oper = 4;

    switch (oper) {
    case 1:
      eindex = ++(vd->edim); vindex = ++(vd->vdim); mem_count++;

      vertex[vindex] = vhull;
      
      switch (edge[hindex].rank) {
      case 0:
	edge[hindex].tail_vertex = vindex;
	edge[hindex].tail_edge = eindex;
	break;
      case 1:
	edge[hindex].head_vertex = vindex;
	edge[hindex].head_edge = eindex;
	break;
      }	
      Sy_SetEdge (&edge[eindex], &chain, hindex, last_eindex, vindex, last_vertex, 
		left_point, right_point, 2);
      
      switch (edge[last_eindex2].rank) {
      case 2: 
	edge[last_eindex2].head_edge = eindex;
	break;
      default: 
	edge[last_eindex2].tail_edge = eindex;
	break;
      }
      if (right_point == edge[hindex].left_point) 
	right_point = edge[hindex].right_point;
      else
	right_point = edge[hindex].left_point;
    
      last_eindex  = hindex;
      last_eindex2 = eindex;
      last_vertex  = vindex;
      vlast        = vhull;
      edge[hindex].rank++;
      break;
    case 2:
      eindex = ++(vd->edim); eindex2 = ++(vd->edim); vindex = ++(vd->vdim);
      mem_count++;

      vertex[vindex] = vcomb;
      
      Sy_SetEdge (&edge[eindex], &chain, eindex2, last_eindex, vindex, last_vertex,
		left_point, right_point, 2);
      
      switch (edge[last_eindex2].rank) {
      case 2: 
	edge[last_eindex2].head_edge = eindex;
	break;
      default: 
	edge[last_eindex2].tail_edge = eindex;
	break;
      }
      Sy_SetEdge (&edge[eindex2], &comb->elem[combindex], 0, eindex, 0, vindex,
		(int)comb->elem[combindex].left_point,
		(int)comb->elem[combindex].right_point, 1);
      
      left_point   = comb->elem[combindex].right_point;
      last_vertex  = vindex;
      last_eindex  = eindex;
      last_eindex2 = eindex2;
      vlast        = vcomb;
      combindex--; 
      break;
    case 3:				
      switch (edge[last_eindex].rank) {
      case 1: 
	vindex = edge[last_eindex].tail_vertex;
	break;
      case 2:
	vindex = edge[last_eindex].head_vertex;
	break;
      }
      switch (edge[hindex].rank) {
      case 0:
	edge[hindex].tail_vertex = vindex;
	edge[hindex].tail_edge   = last_eindex;
	break;
      case 1:
	edge[hindex].head_vertex = vindex;
	edge[hindex].head_edge   = last_eindex;
	break;
      }	
      if (right_point == edge[hindex].left_point) 
	right_point = edge[hindex].right_point;
      else
	right_point = edge[hindex].left_point;
      
      last_eindex = hindex; 
      vlast       = vhull;
      edge[hindex].rank++;
      break;
    case 4:	
      eindex  = ++(vd->edim); eindex2 = ++(vd->edim); vindex  = ++(vd->vdim);
      mem_count++;

      vertex[vindex] = vcomb;
      
      switch (edge[hindex].rank){
      case 0:
	edge[hindex].tail_vertex = vindex;
	edge[hindex].tail_edge = eindex;
	break;
      case 1:
	edge[hindex].head_vertex = vindex;
	edge[hindex].head_edge = eindex;
	break;
      }				
      Sy_SetEdge (&edge[eindex], &chain, eindex2, last_eindex, vindex, last_vertex,
		left_point, right_point, 2);

      switch (edge[last_eindex2].rank){
      case 2: 
	edge[last_eindex2].head_edge = eindex;
	break;
      default: 
	edge[last_eindex2].tail_edge = eindex;
	break;
      }
      Sy_SetEdge (&edge[eindex2], &comb->elem[combindex], 0, hindex, 0, vindex,
		(int)comb->elem[combindex].left_point,
		(int)comb->elem[combindex].right_point,1);
      
      if (right_point == edge[hindex].left_point) 
	right_point = edge[hindex].right_point;
      else
	right_point = edge[hindex].left_point;

      left_point = comb->elem[combindex].right_point;
      
      last_vertex  = vindex; 
      last_eindex  = hindex;
      last_eindex2 = eindex2;
      vlast        = vhull;
      combindex--;
      edge[hindex].rank++;
      break;
    }

    if (oper != 2){
      last_hindex = hindex;
      hindex      = (mode == LSke_PRESCAN_DOWN)? hindex + 1 : hindex - 1;
      if (hindex >= predat->stripcount){
        hindex = predat->stripcount - 1;
        mode   = LSke_PRESCAN_UP;
      }
    }
    
  } /*  while */
  
  eindex = ++(vd->edim);
  
  Sy_Bisector(&point[left_point], &point[right_point], &chain);
  
  Sy_SetEdge (&edge[eindex], &chain, 0, last_eindex, 0, last_vertex,
	    left_point, right_point, 1);
  
  switch (edge[last_eindex2].rank) {
  case 2:
    edge[last_eindex2].head_edge = eindex;
    break;
  default: 
    edge[last_eindex2].tail_edge = eindex;
    break;
  }

  first_hindex = predat->stripcount;
  swap (&edge[first_hindex]); edge[first_hindex].rank--;

  hull->left_index  = hull->right_index  = hull->dim/(2*sizeof(int));

  hull->elem[hull->left_index] = eindex;
  
  if (!is_right_nondeg){
    for (i = predat->stripcount-1; i >= 1; i--){
      hull->right_index++;
      hull->elem[hull->right_index] = i;
    }
  }

  if (!is_left_nondeg){
    for (i = 1; i <= predat->stripcount-1; i++){
      hull->right_index++;
      hull->elem[hull->right_index] = i;
    }
  }
  
  hull->right_index++; 
  hull->elem[hull->right_index] = first_hindex;
}

/**********************************************************************
 * function void Sy_EndVoro (vd, point, edge, vertex, stack)
 * 
 * purpose: final adjustments of 'vd' 
 *
 * param:   vd     - pointer to Voronoi-diagram data
 *          point  - generating point array
 *          edge   - edge array
 *          vertex - vertex array
 *          stack  - VD-stack
 *********************************************************************/
static void Sy_EndVoro (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyVertex *vertex, SyVoroStack *stack)
{
	int	i, vindex;
	SyVertex	vtemp;

	for (i = 1; i <= vd->edim; i++) {
		
	  switch (edge[i].rank) {
	  case 1:
	          vtemp = vertex[edge[i].tail_vertex];
		  if (stack->v_ptr) 
                    vindex = stack->v_[--stack->v_ptr];
		  else 
                    vindex = ++vd->vdim;
		  edge[i].head_vertex = vindex;
                  edge[i].head_edge   = 0;
		  vtemp.xden = edge[i].dirx * vtemp.num + vtemp.xden;
		  vtemp.yden = edge[i].diry * vtemp.num + vtemp.yden;
		  vertex[vindex] = vtemp;
		  break;
	  case 0:
		  if (stack->v_ptr) 
                    vindex = stack->v_[--stack->v_ptr];
		  else 
                    vindex = ++vd->vdim;
		  edge[i].tail_vertex = vindex;
                  edge[i].head_edge   = 0;
                  edge[i].tail_edge   = 0;
		  vertex[vindex].xden = edge[i].xoff;
		  vertex[vindex].yden = edge[i].yoff;
		  vertex[vindex].num  = 1;
		  if (stack->v_ptr) 
                    vindex = stack->v_[--stack->v_ptr];
		  else 
                    vindex = ++vd->vdim;
		  edge[i].head_vertex = vindex;
		  vertex[vindex].xden = edge[i].xoff + edge[i].dirx;
		  vertex[vindex].yden = edge[i].yoff + edge[i].diry;
		  vertex[vindex].num  = 1;
		  break;
	  }
	
	}
	for (i = 1; i <= vd->vdim; i++){
	  vertex[i].num *= 2; }

	for (i = 1; i <= vd->pdim; i++){
	  point[i].x >>= 1; point[i].y >>= 1;
	}
}


/**********************************************************************
 * function void Sy_PrepComb (vd, point, comb)
 * 
 * purpose: initializes 'comb'-structure of a single scan-line
 *
 * param:   vd    - pointer to Voronoi-diagram
 *          point - generating point array
 *          comb  - initialized 'comb'-structure 
 *********************************************************************/
static void Sy_PrepComb(SyVoroDAC *vd, SyPoint *point, SyComb *comb)
{
	int  i, j = 1; 

	for (i = comb->plow; i < comb->phigh; i++) {
		Sy_Bisector(&point[i+1], &point[i], &comb->elem[j]);
		comb->elem[j].left_point = i + 1;
		comb->elem[j].right_point = i;
		comb->elem[j].rank = 0;
		j++;
	}
	comb->count = comb->n - 1;
}

#if defined (sun4)
#define SY_FLOOR floor
#else 
#define SY_FLOOR floor
#endif

#ifndef FASTVORO 
#define SY_TRFAC 1000.0
#define SY_VEQUALX(va,vb) ( (SY_FLOOR((SY_TRFAC*(va).xden)/(va).num) != \
                          SY_FLOOR((SY_TRFAC*(vb).xden)/(vb).num)) ? \
			0 : Sy_Vequalx ((va),(vb)))
#else
#define SY_VEQUALX(va,vb) ( ((va).xden * (vb).num)==((va).num * (vb).xden) )
#endif

static void Sy_DcelDown(SyEdge *edge, int *init_eindex, int *init_vref, int *poly_halt);

static void Sy_Discard(SyEdge *chain, SyEdge *edge, SyVertex *vertex, SyVoroStack *stack, 
		       int *start_eindex, int *init_vref, int *is_member); 

static void Sy_SetEdge(SyEdge *e, SyEdge *e_base, int he, int te, int hv, int tv, 
		       int lp, int rp, int rank);

static void Sy_Bisector(SyPoint *p1, SyPoint *p2, SyEdge *e);

static int member(SyEdge *chain, SyEdge *e, SyVertex *vertex);
static int right(SyEdge *chain, SyEdge *e, SyVertex *vertex, int *vright);
static int Sy_Vequalx(SyVertex a, SyVertex b);

/**********************************************************************
 * function void Sy_LinkVoro (vd, point, edge, vertex, comb, hull, 
 *                          support, stack)
 * 
 * purpose: merges `comb`-VD and regular VD 
 *
 * param:   vd      - pointer to Voronoi-diagram 
 *          point   - generating point array
 *          edge    - edge array
 *          vertex  - vertex array
 *          comb    - pointer to `comb`-structure
 *          hull    - convex hull 
 *          support - pointer to `support`-structure
 *          stack   - specific VD-routine stack
 *********************************************************************/
static void Sy_LinkVoro (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyVertex *vertex, SyComb *comb, SyHull *hull, SySupport *support, SyVoroStack *stack)
{
	SyEdge	        chain;
	SyVertex          vhull, vcomb, vlast;

	double          a, b;
	int		mem_count = 0, hcut_count = 0;

        int      pleft, pright,

	         hindex,
	         eindex, eindex2, 
	         vindex,
	         combindex,

	         vref,
	         vright,
	         
                 last_hindex  = 0,
	         last_eindex  = 0,
	         last_eindex2 = 0,
	         last_vertex  = 0,
              
	         start_eindex,
	         rank,
	         first_hindex,
		 
		 comb_intersect,
		 hull_intersect,

	         rot, rot0,
                 poly_halt = 0,
	         dis       = 0,
		 hull_done = 0, 
	         scan_done = 0,
	         is_member = 0, oper;

        vcomb.xden = vcomb.yden = 0; 
        vcomb.num = 1.0;
        
	pright       = support->suphigh; 
	pleft        = comb->phigh; 
	hindex       = support->rayhigh; 
	combindex    = comb->n - 1;
	vlast.xden   = 0;
        vlast.num    = 1;
	vref         = edge[hindex].tail_vertex;
	rank         = edge[hindex].rank;
	hull_done    = 0; 
	stack->h_ptr = 0;

	Sy_DcelDown (edge, &hindex, &vref, &poly_halt);

	do {
                hull_intersect = 0;
                comb_intersect = 0;
  		if ((pleft == comb->plow)&&(pright == support->suplow)) break;
              
		Sy_Bisector(&point[pleft], &point[pright], &chain);

		comb_intersect = 
		     ( (combindex > 0) && 
		        Sy_Intersection(&chain, &comb->elem[combindex], &vcomb) );

		if (!hull_done){

		  if (!poly_halt){
		    Sy_Discard (&chain, edge, vertex, stack, &hindex, &vref, 
			     &is_member);
		  }

		  if (!is_member || poly_halt) 
		    is_member = member(&chain, &edge[hindex], vertex);

		  if (is_member
		      && Sy_Intersection(&chain, &edge[hindex], &vhull)){

		    rot =(SY_VEQUALX (vlast, vhull) && (mem_count != 0));

                    if (dis || !((last_hindex == hindex) && rot)){
		      hull_intersect = 1;
		    }
		  }
		}

		a = vhull.xden * vcomb.num;
		b = vhull.num  * vcomb.xden;
		
		if ( ((a > b) && hull_intersect && comb_intersect) 
                   || (hull_intersect && !comb_intersect))          oper = 1;
		if ( ((a < b) && hull_intersect && comb_intersect)
                    || (comb_intersect && !hull_intersect))         oper = 2;
                if ( (oper == 1) && rot )                           oper = 3;
                if ( (a == b) && hull_intersect && comb_intersect)  oper = 4; 

		if (oper != 2){

		  switch(edge[hindex].rank){
		  case 1:
		    if ((edge[hindex].right_point == support->suplow) &&
			!dis  && (stack->h_ptr == 0) && (hcut_count > 0)){
		      hull_done = 1;
		    }
		    break;
		  case 0:
		    
		    if ((edge[hindex].left_point == support->suplow) && 
			!dis && (stack->h_ptr == 0) && (hcut_count > 0)){
		      hull_done = 1;
		    }
		    break;
		  }

		  if (poly_halt) poly_halt = 0;

		  if (dis && (stack->h_ptr == 0)) {
		    dis = 0;
		    poly_halt = edge[hindex].rank < 1;
		  }
		}		  
	      
		switch (oper) {
		case 1:
				
			if (stack->e_ptr) eindex = stack->e_[--stack->e_ptr];
			else eindex = ++(vd->edim);
		        if (stack->v_ptr) vindex = stack->v_[--stack->v_ptr];
			else vindex = ++(vd->vdim);
			
			vertex[vindex] = vhull;
			
			switch (edge[hindex].rank) {
			case 0:
				edge[hindex].tail_vertex = vindex;
				edge[hindex].tail_edge = eindex;
				break;
			case 1:
				edge[hindex].head_vertex = vindex;
				edge[hindex].head_edge = eindex;
				break;
			}	
			Sy_SetEdge (&edge[eindex], &chain, hindex, last_eindex,
			        vindex, last_vertex, pleft, pright, 2);

                        switch (edge[last_eindex2].rank) {
			case 2: 
				edge[last_eindex2].head_edge = eindex;
                        	break;
		       	default: 
                        	edge[last_eindex2].tail_edge = eindex;
                        	break;
			}
			if (pright == edge[hindex].left_point) {
				pright = edge[hindex].right_point;
			} else {
				pright = edge[hindex].left_point;
			}

			if (!mem_count){
			  mem_count++; first_hindex = eindex;
			  swap (&edge[first_hindex]); edge[first_hindex].rank--;
			}

                        last_eindex = hindex;
                        last_eindex2 = eindex;
			last_vertex = vindex;
                        vlast = vhull;
			edge[hindex].rank++;
			hcut_count++;
                        break;
 		case 2:
			if (stack->e_ptr) eindex = stack->e_[--stack->e_ptr];
			else eindex = ++(vd->edim);
			if (stack->e_ptr) eindex2 = stack->e_[--stack->e_ptr];
			else eindex2 = ++(vd->edim);
		        if (stack->v_ptr) vindex = stack->v_[--stack->v_ptr];
			else vindex = ++(vd->vdim);
			
			vertex[vindex] = vcomb;

                        Sy_SetEdge (&edge[eindex],&chain, eindex2,
                           last_eindex, vindex, last_vertex,
                           pleft, pright, 2);

                        switch (edge[last_eindex2].rank) {
			case 2: 
			       edge[last_eindex2].head_edge = eindex;
                               break;
		        default: 
                               edge[last_eindex2].tail_edge = eindex;
                               break;
			}
			Sy_SetEdge (&edge[eindex2],&comb->elem[combindex],
				0,eindex,0,vindex,
                           	(int)comb->elem[combindex].left_point,
				(int)comb->elem[combindex].right_point,1);
                        
			if (!mem_count){
			  mem_count++; first_hindex = eindex;
			  swap (&edge[first_hindex]); edge[first_hindex].rank--;
			}

                        pleft = comb->elem[combindex].right_point;
			last_vertex = vindex;
			last_eindex = eindex;
			last_eindex2 = eindex2;
                        combindex--; vlast = vcomb;
                        break;
		case 3:				
			switch (edge[last_eindex].rank) {
		        case 1: 
		        	vindex = edge[last_eindex].tail_vertex;
		          	break;
		        case 2:
		          	vindex = edge[last_eindex].head_vertex;
		          	break;
		        }
			switch (edge[hindex].rank) {
			case 0:
				edge[hindex].tail_vertex = vindex;
				edge[hindex].tail_edge = last_eindex;
				break;
			case 1:
				edge[hindex].head_vertex = vindex;
				edge[hindex].head_edge = last_eindex;
				break;
			}	
		      	if (pright == edge[hindex].left_point) 
				pright = edge[hindex].right_point;
			else
				pright = edge[hindex].left_point;

 			last_eindex = hindex; vlast = vhull;
			edge[hindex].rank++;
			hcut_count++;
                        break;
		case 4:	
			if (stack->e_ptr) eindex = stack->e_[--stack->e_ptr];
			else eindex = ++(vd->edim);
			if (stack->e_ptr) eindex2 = stack->e_[--stack->e_ptr];
			else eindex2 = ++(vd->edim);
		        if (stack->v_ptr) vindex = stack->v_[--stack->v_ptr];
			else vindex = ++(vd->vdim);

			vertex[vindex] = vcomb;

			switch (edge[hindex].rank){
			case 0:
				edge[hindex].tail_vertex = vindex;
				edge[hindex].tail_edge = eindex;
				break;
			case 1:
				edge[hindex].head_vertex = vindex;
				edge[hindex].head_edge = eindex;
				break;
			}				
                        Sy_SetEdge (&edge[eindex],&chain,eindex2,
                           last_eindex, vindex, last_vertex,
                           pleft, pright, 2);

                        switch (edge[last_eindex2].rank){
			case 2: 
			       edge[last_eindex2].head_edge = eindex;
                               break;
		        default: 
                               edge[last_eindex2].tail_edge = eindex;
                               break;
			}
			Sy_SetEdge (&edge[eindex2],&comb->elem[combindex],
			   0,hindex,0,vindex,
                           (int)comb->elem[combindex].left_point,
			   (int)comb->elem[combindex].right_point,1);
			   
		      	if (pright == edge[hindex].left_point) 
				pright = edge[hindex].right_point;
			else
				pright = edge[hindex].left_point;

                        pleft = comb->elem[combindex].right_point;

			if (!mem_count){
			  mem_count++; first_hindex = eindex;
			  swap (&edge[first_hindex]); edge[first_hindex].rank--;
			}

			last_vertex = vindex; vlast = vhull;
			last_eindex = hindex;
                        last_eindex2 = eindex2;
                        combindex--;
			edge[hindex].rank++;
			hcut_count++;
                        break;
		}
		if ((oper != 2)){
		  last_hindex = hindex;
		  dis = (stack->h_ptr != 0);

		  if (dis){
		    hindex = stack->h_[--stack->h_ptr];
		    vref   = edge[hindex].tail_vertex; 
		    rank   = edge[hindex].rank;
		  }
		  else if (!poly_halt){
		    vref   = edge[hindex].tail_vertex; 
		    Sy_DcelDown (edge, &hindex, &vref, &poly_halt);		    
		  }
		}

	} while (1);
	
	if (stack->e_ptr) eindex = stack->e_[--stack->e_ptr];
	  else eindex = ++(vd->edim);
	
	Sy_Bisector(&point[pleft], &point[pright], &chain);

        Sy_SetEdge (&edge[eindex], &chain, 0, last_eindex, 0, last_vertex,
                   pleft, pright, 1);

        switch (edge[last_eindex2].rank) {
	case 2:
		edge[last_eindex2].head_edge = eindex;
                break;
	default: 
                edge[last_eindex2].tail_edge = eindex;
                break;
	}

        hull->left_count++;
	hull->right_count++;

        hull->right_index++; 

	hull->left[hull->left_count]   =
	hull->right[hull->right_count] = first_hindex;

	hull->elem[hull->right_index] = first_hindex;

        if (mem_count > 0) {
	  hull->left[hull->left_count] = eindex;

	  hull->left_index--;
	  hull->elem[hull->left_index] = eindex;
	}
}

/**********************************************************************
 * function void Sy_DcelDown (edge, start_eindex, vright)
 * 
 * purpose: 
 *
 * param:   edge           - edge array 
 *          start_eindex   - first edge of the `Sy_DcelDown` operation
 *          vright         - reference vertex
 *********************************************************************/
static void Sy_DcelDown (SyEdge *edge, int *init_eindex, int *init_vref, int *poly_halt)
{
	int curedge, vref;
	
	if (*init_vref == 0) return;
	curedge = *init_eindex;

	if ((edge[curedge].rank < 2) && (edge[curedge].head_edge != 0)){
	  curedge = edge[*init_eindex].head_edge;
	  edge[*init_eindex].head_edge = 0;
	  if (curedge == *init_eindex) *poly_halt = 1;
	  *init_eindex = curedge;
          *init_vref   = 0;
	  return;
	}

	vref    = *init_vref;
	
	do {
		if (edge[curedge].tail_vertex == vref) 
			curedge = edge[curedge].tail_edge;
		else
			curedge = edge[curedge].head_edge;
		if (edge[curedge].tail_vertex == vref)
			vref = edge[curedge].head_vertex;
		else
			vref = edge[curedge].tail_vertex;
			
	} while (edge[curedge].rank >= 2);
	*init_eindex = curedge;
	*init_vref   = edge[curedge].tail_vertex;
}


/**********************************************************************
 * function void Sy_Discard (chain, edge, vertex, stack, start_eindex,
 *                        endhindex, mode)
 * 
 * purpose: dicard edges with respect to the dividing chain 
 *
 * param:   chain         - current member of the dividing chain
 *          edge          - edge array
 *          vertex        - vertex array
 *          stack         - VD-stack
 *          start_eindex  - edge where `Sy_Discard` starts
 *          endhindex     - edge where `Sy_Discard` stops
 *          mode          - mode of operation
 *********************************************************************/
static void Sy_Discard (SyEdge *chain, SyEdge *edge, SyVertex *vertex, SyVoroStack *stack, int *start_eindex, int *init_vref, int *is_member)
{
        int next, vref, next2, next3, curedge, vright;
	*is_member = 0;

	if (!right(chain, &edge[*start_eindex], vertex, &vright)) return;
	
	*is_member = 1;
	curedge = *start_eindex;
	vref = edge[curedge].tail_vertex;
	next = edge[curedge].tail_edge;

	do {
		do {
			stack->h_[stack->h_ptr++] = next;
			if (edge[next].tail_vertex == vref) {
				next2 = edge[next].tail_edge;
				swap (&edge[next]);
				edge[next].rank--;}
			else {
				next2 = edge[next].head_edge;
				edge[next].rank--;
			}
			if (edge[next2].tail_vertex == vref)
				next3 = edge[next2].tail_edge;
			else 
				next3 = edge[next2].head_edge;
			next = next2;
		} while (next3 != curedge);
		
		stack->e_[stack->e_ptr++] = curedge;
		stack->v_[stack->v_ptr++] = vref;
		curedge = next;
		
		if (edge[curedge].rank < 2) break;
		if (!right(chain, &edge[curedge], vertex, &vright)) break;
		
		if (edge[curedge].tail_vertex == vref) {
			vref = edge[curedge].head_vertex;
			next = edge[curedge].head_edge;}
		else {
			vref = edge[curedge].tail_vertex;
			next = edge[curedge].tail_edge;
		}
	} while (1);
	
	if ( (edge[curedge].tail_vertex == vref) && 
	     (edge[curedge].rank >= 2) ) {  
		swap (&edge[curedge]);
		edge[curedge].rank--;
	}
	else 
		edge[curedge].rank--;
		
	*start_eindex = curedge;
	*init_vref    = vref;
}		


/**********************************************************************
 * function void Sy_Bisector(p1, p2, e)
 * 
 * purpose: computes Sy_Bisector of two points (rotation + SY_PI/2)
 *
 * param:   p1, p2  - points 
 *          e       - resulting Sy_Bisector
 *********************************************************************/
static void Sy_Bisector (SyPoint *p1, SyPoint *p2, SyEdge *e)
{
	(*e).dirx = (*p2).y - (*p1).y;
	(*e).diry = (*p1).x - (*p2).x;
       	(*e).xoff  = ( (*p1).x + (*p2).x ) >> 1;  /*  >> 1  <==>  / 2   */ 
	(*e).yoff  = ( (*p1).y + (*p2).y ) >> 1;
}


/**********************************************************************
 * function int Sy_Intersection (e1, e2, v)
 * 
 * purpose: computes the Sy_Intersection of two lines, returns 0 in
 *          the case of parallel lines
 *
 * param:   e1, e2 - lines (infinite edges)
 *          v      - resulting vertex
 *********************************************************************/
#if defined (sun4) || defined (DOUBLE_BEST) 
static int Sy_Intersection (SyEdge *e1, SyEdge *e2, SyVertex *v)
{	
  double  tden;

  (*v).num = (double)(*e2).dirx * (double)(*e1).diry  
           - (double)(*e2).diry * (double)(*e1).dirx;
  if ( (*v).num == 0 ) return (0);
      	
  if ( (*v).num < 0 ) {
    (*v).num = - (*v).num;
    tden =  (double)(*e2).diry * (double)( (*e2).xoff - (*e1).xoff )
          + (double)(*e2).dirx * (double)( (*e1).yoff - (*e2).yoff );
  }
  else
    tden =  (double)(*e2).diry * (double)( (*e1).xoff - (*e2).xoff )
          + (double)(*e2).dirx * (double)( (*e2).yoff - (*e1).yoff );
  
  (*v).xden =  (*e1).dirx * tden + (*e1).xoff * (*v).num;
  (*v).yden =  (*e1).diry * tden + (*e1).yoff * (*v).num;
  
  return (1);
}
#else
static int Sy_Intersection (SyEdge *e1, SyEdge *e2, SyVertex *v)
{	
	double  	tden;

	(*v).num = (*e2).dirx * (*e1).diry  - (*e2).diry * (*e1).dirx;
      	if ( (*v).num == 0 ) return (0);
      	
      	if ( (*v).num < 0 ) {
      		(*v).num = - (*v).num;
      		tden =  (*e2).diry * ( (*e2).xoff - (*e1).xoff )
	              + (*e2).dirx * ( (*e1).yoff - (*e2).yoff );
	}
	else
		tden =  (*e2).diry * ( (*e1).xoff - (*e2).xoff )
	              + (*e2).dirx * ( (*e2).yoff - (*e1).yoff );
  
	(*v).xden =  (*e1).dirx * tden + (*e1).xoff * (*v).num;
	(*v).yden =  (*e1).diry * tden + (*e1).yoff * (*v).num;
	
	return (1);
}
#endif

/**********************************************************************
 * function int member (chain, e, vertex)
 * 
 * purpose: tests whether the vertex resulting from the Sy_Intersection
 *          of the dividing chain with an edge is a member of that edge,
 *          returns 1 if positive 
 *
 * param:   chain  - dividing chain
 *          e      - edge
 *          vertex - vertex array
 *********************************************************************/
static int member (SyEdge *chain, SyEdge *e, SyVertex *vertex)
{	
        double tmp1, tmp2;
	SyVertex 	vt, vh;
	
	switch ((*e).rank) {
	case 2:
		vt   = vertex[(*e).tail_vertex];
		vh   = vertex[(*e).head_vertex];
		tmp1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
		     - ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                if (tmp1 == 0.0) return (1);  
		tmp2 = ((*chain).dirx * ( vh.yden - (*chain).yoff * vh.num ))
		     - ((*chain).diry * ( vh.xden - (*chain).xoff * vh.num ));
                if (tmp2 == 0.0) return (1);
		return ((tmp1 < 0.0) != (tmp2 < 0.0));
		break;
	case 1:
		vt   = vertex[(*e).tail_vertex];
		tmp1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
		     - ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                if (tmp1 == 0.0) return (1);
		tmp2 = ((*chain).diry * (*e).dirx)
		     - ((*chain).dirx * (*e).diry);
		if (tmp2 == 0.0) return (0);
		return ((tmp1 < 0.0) == (tmp2 < 0.0)); 
		break;
	case 0:
		return (1); break;
	}
}

/**********************************************************************
 * function void Sy_ComputeSupport (vd, point, edge, comb, hull, support)
 * 
 * purpose: computes the support for the next Voronoi-diagram merge step
 *
 * param:   vd      - pointer to Voronoi-diagram data
 *          point   - generating point array
 *          edge    - edge array
 *          comb    - comb interval of the new scan-line
 *          hull    - convex hull
 *          support - `support`-structure
 *********************************************************************/
static void Sy_ComputeSupport (SyVoroDAC *vd, SyPoint *point, SyEdge *edge, SyComb *comb, SyHull *hull, SySupport *support)
{
	SyIntVector   vright, vleft, vmax, vmin;
	SyPoint     plo, phi;
	int	  max_i, min_i, rindex, lindex, i;

	plo      = point[comb->plow]; 
        phi      = point[comb->phigh];
	rindex   = edge[hull->elem[hull->right_index]].right_point;
	lindex   = edge[hull->elem[hull->left_index ]].left_point;
	max_i    = hull->right_index;
        min_i    = hull->left_index;
        vmax.dirx = point[rindex].x - phi.x;
        vmax.diry = point[rindex].y - phi.y;
        vmin.dirx = point[lindex].x - plo.x;
        vmin.diry = point[lindex].y - plo.y;
       
	for (i = hull->right_index-1; i >= hull->left_index; i--) {

		rindex = edge[hull->elem[i]].right_point;
		vright.dirx = point[rindex].x - phi.x;
		vright.diry = point[rindex].y - phi.y;

		if ((vright.dirx * vmax.diry) < (vright.diry * vmax.dirx)){
		  max_i    = i;
		  vmax      = vright;
		}
		else break;
		
	}

	for (i = hull->left_index+1; i <= hull->right_index; i++) {

	        lindex     = edge[hull->elem[i]].left_point;
		vleft.dirx = point[lindex].x - plo.x;
		vleft.diry = point[lindex].y - plo.y;

		if ((vleft.dirx * vmin.diry) > (vleft.diry * vmin.dirx)){
		  min_i    = i;
		  vmin      = vleft;
		}
		else break;
			
	}

	if (max_i+1 < min_i){
	  Sy_SimpleError (SY_FMTNL, 
			  "INTERNAL ERROR IN CONVEX HULL ALGORITHM ");
	} 

	hull->right_index = max_i;
	hull->left_index  = min_i;
	support->suplow   = edge[hull->elem[min_i]].left_point;
	support->raylow   = hull->elem[min_i]; 
	support->suphigh  = edge[hull->elem[max_i]].right_point;
	support->rayhigh  = hull->elem[max_i];
}

/**********************************************************************
 * function void Sy_SetEdge (e, e_base, he, te, hv, tv, lp, rp, rank)
 * 
 * purpose: updates the parameters of an edge
 *
 * param:   e       - edge to be updated 
 *          e_base  - edge to copy data from
 *          he      - head_edge
 *          te      - tail_edge
 *          hv      - head_vertex
 *          tv      - tail_vertex
 *          lp      - left_point
 *          rp      - right_point
 *          rank    - rank of e 
 *********************************************************************/
static void Sy_SetEdge (SyEdge *e, SyEdge *e_base, int he, int te, int hv, int tv, int lp, int rp, int rank)
{
	*e               = *e_base;
	(*e).head_edge   = he;
	(*e).tail_edge   = te;
	(*e).head_vertex = hv;
	(*e).tail_vertex = tv;
	(*e).left_point  = lp;
	(*e).right_point = rp;
	(*e).rank        = rank;
}


/**********************************************************************
 * function int Sy_GetScanline (vd, point, comb)
 * 
 * purpose: get a scan-line of points from `vd`
 *
 * param:   vd    - pointer to Voronoi-diagram data
 *          point - generating point array
 *          comb  - pointer to current comb interval: 
 *                   comb->plow, comb->phigh, comb->n 
 *********************************************************************/
static int Sy_GetScanline (SyVoroDAC *vd, SyPoint *point, SyComb *comb)
{
  int    count, oldy, i = 0;

  count = comb->plow = comb->phigh + 1;
  if (count > vd->pdim) return (comb->n = i);

  i = 1; count++;
  oldy = point[comb->plow].y;

  while (count <= vd->pdim) {
    if (oldy == point[count].y){
      count++; i++;
    }
    else break;
  }
  comb->phigh = count - 1;
  return (comb->n = i);
} 


/**********************************************************************
 * function void swap (e)
 * 
 * purpose: reverts the direction of an edge 
 *
 * param:   e  - edge  
 *********************************************************************/
static void swap (SyEdge *e)
{
#ifdef SY_EXCH
#undef SY_EXCH
#endif
#define SY_EXCH(x,y) temp = x; x = y; y = temp

	int	temp; /* for 'SY_EXCH' macro */
	
	switch((*e).rank){
	case 2:
        case 0:
	  SY_EXCH ((*e).tail_vertex, (*e).head_vertex);
	  SY_EXCH ((*e).tail_edge, (*e).head_edge);
   	  SY_EXCH ((*e).left_point, (*e).right_point);
   	  (*e).dirx = - (*e).dirx;
   	  (*e).diry = - (*e).diry;
           break;
	}
	
}


/**********************************************************************
 * function int right (chain, e, vertex, vright)
 * 
 * purpose: tests whether edge `e` is on the right side with respect
 *          to `chain`, an element of the dividing chain
 *          returns 1 if positive
 *
 * input:   chain  - element of the dividing chain 
 *          e      - edge to be tested
 *          vertex - vertex array
 *          vright - vertex located on the right side of `chain`
 *********************************************************************/
static int right (SyEdge *chain, SyEdge *e, SyVertex *vertex, int *vright)
{	
	int    res1, res2;
	SyVertex 	vt, vh;
	
	switch ((*e).rank) {
	case 2:
		vt = vertex[(*e).tail_vertex];
		vh = vertex[(*e).head_vertex];
		res1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
		    <= ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
		res2 = ((*chain).dirx * ( vh.yden - (*chain).yoff * vh.num ))
		    <= ((*chain).diry * ( vh.xden - (*chain).xoff * vh.num ));
		if (res1) *vright = (*e).tail_vertex;
		if (res2) *vright = (*e).head_vertex;
		return (res1 && res2);
		break;
	case 1:
		vt = vertex[(*e).tail_vertex];
		res1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
		    <= ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
		*vright = (*e).tail_vertex;
		return (res1); 
		break;
	case 0:
		*vright = 0; return (0); break;
	}
}

static int Sy_Vequalx (SyVertex a, SyVertex b)
{
  double xdiv, ndiv, xmod, nmod, h;
  int    abuf, bbuf;

  abuf = SY_SGN(a.xden);
  bbuf = SY_SGN(b.xden);

  if (abuf != bbuf) return (0);
  if ((abuf == 0) && (bbuf == 0)) return(1);
  
  if (abuf < 0){
    a.xden = SY_ALV(a.xden);
    b.xden = SY_ALV(b.xden);
  }
   
  do{
    if (a.xden < b.xden){
      if (a.num > b.num) return(0);
      h = a.xden; a.xden = b.xden; b.xden = h;
      h = a.num;  a.num  = b.num;  b.num  = h;
    }
  
    xdiv = floor(a.xden/b.xden) - 1.0;
    xmod = a.xden - xdiv * b.xden;
    if (xmod >= b.xden){xdiv += 1.0; xmod -= b.xden;}
    
    ndiv = floor(a.num/b.num) - 1.0;
    nmod = a.num - ndiv * b.num;
    if (nmod >= b.num){ndiv += 1.0; nmod -= b.num;}
    
    if (xdiv != ndiv) return (0);

    a.xden = xmod; a.num = b.xden; b.xden = nmod;

  } while((xmod != 0) && (nmod != 0));

  return (xmod == nmod);
}
